// ====================================================================
// acl.cc -- A class for Access Control Lists
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.3 $ $Date: 2006/12/05 06:44:27 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include "acl.h"
#include "global.h"
#include "syslog.h"
#include "utility.h"

#include <iostream>

// ====================================================================
// Validate a credential by calling the supplied callback.  The
// callback is only called once, assuming that it's value is 
// not going to change.  
// ====================================================================

bool Cred::Validate(void)
{
   if (!callback) return false;
   bool result = callback->ValidateCred(this);
   callback = NULL;
   return result;
}

// ====================================================================
// Check a list of credentials against this ACL to see if any
// of them are allowed access.
// ====================================================================

bool ACE::Allow(CredList *list)
{
//   cerr << "ACE::Allow(CredList) called" << endl;
   for (Cred *c = list->First(); c; c = list->Next())
   {
      if (c->Type() == Type())
      {
         if (! c->Valid() && ! c->Validate()) continue;
         if (Allow(c)) return true;
      }
   }
   return false;
}

// ====================================================================
// Check if the host part of the entry is allowed.
// ====================================================================

bool ACE::HostAllowed(string h)
{
//   cerr << "ACE::HostAllowed(" << h << ") called" << endl;
   if (host == "" || host == "*") return true;
   
   return HostCompare(h.c_str(), host.c_str());
}

// ====================================================================
// Access control entries for unauthenticated host access
// ====================================================================

bool ACE_Host::Allow(Cred *c)
{
//   cerr << "ACE_HOST::Allow(Cred(" 
//        << c->StringType() 
//        << ","
//        << c->Host()
//        << ","
//        << c->Token()
//        << ") FOR("
//        << token
//        << "@"
//        << host
//        << ")) called" << endl;
   if (c->Type() != HOST) return false;
   if (! HostAllowed(c->Host())) return false;
   
   if (token == "" || token == "*" || token == c->Token()) return true;
   return false;
}

// ====================================================================
// Access control entries for GSSAPI 
// ====================================================================

bool ACE_GSSAPI::Allow(Cred *c)
{
//   cerr << "ACE_GSSAPI::Allow(Cred(" 
//        << c->StringType() 
//        << ","
//        << c->Host()
//        << ","
//        << c->Token()
//        << ") FOR("
//        << token
//        << "@"
//        << host
//        << ")) called" << endl;
   if (c->Type() != GSSAPI) return false;
   if (! HostAllowed(c->Host())) return false;

   if (c->Token() == token) return true;
   
   return false;
}

// ====================================================================
// Access control entries for Certificate
// ====================================================================

bool ACE_CERT::Allow(Cred *c)
{
//   cerr << "ACE_CERT::Allow(Cred(" 
//        << c->StringType() 
//        << ","
//        << c->Host()
//        << ","
//        << c->Token()
//        << ") FOR("
//        << token
//        << "@"
//        << host
//        << ")) called" << endl;
   if (c->Type() != CERT) return false;
   if (! HostAllowed(c->Host())) return false;
   
   return false;
}

// ====================================================================
// Access control entries for an ACL sublist
// ====================================================================

bool ACE_ACL::Allow(CredList *credlist)
{
//   cerr << "ACE_ACL::Allow(CredList)) called" << endl;
   return subacl->Allow(credlist);
}

bool ACE_ACL::Allow(Cred *cred)
{
//   cerr << "ACE_ACL::Allow(Cred)) called" << endl;
   // Must be defined, but is not used.
}

// ====================================================================
// Tracing
// ====================================================================

void ACE::Trace(void)
{
   if (GLOBAL.Debug())
   {
      SYSLOG << Severity(Syslog::debug)
             << "Access allowed by " << StringType()
             << " rule for " 
             << Token()
             << " at "
             << Host()
             << endl;
   }
}

// ====================================================================
// Constructor for ACLs.
// ====================================================================

ACL::ACL(void)
{
}

// ====================================================================
// Destructor for ACLs
// ====================================================================

ACL::~ACL(void)
{
}

// ====================================================================
// Add an access control entry to an access control list.
// ====================================================================

void ACL::Add(ACE *newace)
{
   list.push_back(newace);
}

// ====================================================================
// Determine if a user from a host is allowed access by checking the
// entire list of ACLs until one returns true.
// ====================================================================

bool ACL::Allow(CredList *credlist)
{
   vector<ACE *>::iterator i = list.begin();
   while (i != list.end())
   {
      if ((*i)->Allow(credlist)) 
      {
         (*i)->Trace();
         return true;
      }
      i++;
   }

   return false;
}
