// ====================================================================
// acl.h -- A class for Access Control Lists
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.3 $ $Date: 2006/12/05 06:44:27 $ $Author: boeheim $
// ******************************************************************** //

#ifndef acl_hh
#define acl_hh

#include <string>
#include <vector>

class Cred;
class CredList;

// --------------------------------------------------------------------
// Access Control Entry - virtual base class.
// --------------------------------------------------------------------
class ACE
{
protected:
   string          token;       // Security token to check
   string          host;        // Hosts it is valid from.
   
public:
   typedef enum { NONE, HOST, GSSAPI, CERT, SUBACL } ACE_TYPE;
   ACE(void)                 { token = ""; host = ""; }; 
   ACE(string t, string h)   { token = t;  host = h;  };
   void SetToken(string t)   { token = t; }; 
   void SetHost(string h)    { host = h;  };
   string Token(void)        { return (token == "") ? "ANYONE" : token; }
   string Host(void)         { return (host == "" || host == "*") ? "ANYWHERE" : host; }

   virtual bool Allow(CredList *list);
   virtual bool Allow(Cred *c) = 0;
   bool HostAllowed(string host);
   virtual ACE_TYPE Type(void) = 0;
   virtual string StringType(void) = 0;
   virtual void Trace(void);
};

// --------------------------------------------------------------------
// Access Control Entry - unauthenticated
// --------------------------------------------------------------------
class ACE_Host : public ACE
{
public:
   ACE_Host(string t, string h) : ACE(t,h) {  };
   bool Allow(Cred *c);
   ACE_TYPE Type(void)     { return HOST; };
   string StringType(void) { return "Host"; };
};

// --------------------------------------------------------------------
// Access Control Entry - GSSAPI
// --------------------------------------------------------------------
class ACE_GSSAPI : public ACE
{
public:
   ACE_GSSAPI(string t, string h) : ACE(t,h) {  };
   bool Allow(Cred *c);
   ACE_TYPE Type(void)     { return GSSAPI; };
   string StringType(void) { return "GSSAPI"; };
};

// --------------------------------------------------------------------
// Access Control Entry - Certificate
// --------------------------------------------------------------------
class ACE_CERT : public ACE
{
public:
   ACE_CERT(string t, string h) : ACE(t,h) {  };
   bool Allow(Cred *c);
   ACE_TYPE Type(void)     { return CERT; };
   string StringType(void) { return "CERT"; };
};

// --------------------------------------------------------------------
// Access Control Entry - Sub access list
// --------------------------------------------------------------------
class ACL;
class ACE_ACL : public ACE
{
private:
   ACL *subacl;
public:
   ACE_ACL(ACL *newacl) { subacl = newacl; } ;
   bool Allow(CredList *credlist);
   bool Allow(Cred *cred);
   ACE_TYPE Type(void)     { return SUBACL; };
   string StringType(void) { return "ACL"; };
   void Trace(void) {};
};

// ====================================================================
// Access Control List
// ====================================================================
class ACL
{
private:
   vector<ACE *> list;
   
public:
   ACL(void);                   // constructor
   
   ~ACL(void);                  // Destructor

   void Add(ACE *newace);

   bool Allow(CredList *credlist);
};

// ====================================================================
// Credential Entry -- virtual base class
// ====================================================================

   // Callback object
   class CECallback
   {
     public:
      CECallback(void) {};
      virtual bool ValidateCred(Cred *) = 0;
   };

class Cred
{
protected:
   string          token;      // Security token
   string          host;       // Host received from
   bool            valid;      // Mark entry validated.
   CECallback      *callback;  // Callback object
   
public:
   Cred(string t = "",
        string h = "",
        bool v = false)         { token = t;  host = h; valid = v; };
   void SetCallback(CECallback *cb) { callback = cb; };
   void SetToken(string s)      { token = s; } ;
   string Token(void)           { return token; };
   string Host(void)            { return host; };
   void SetValid(bool v = true) { valid = v; };
   bool Valid(void)             { return valid; };
   bool Validate(void);

   virtual ACE::ACE_TYPE Type(void) {return(ACE::NONE);};
   virtual string StringType(void) {return("");};
};

// --------------------------------------------------------------------
// Host credential
// --------------------------------------------------------------------
class Cred_HOST : public Cred
{
public:
   Cred_HOST(string t, string h, bool v) : Cred(t,h,v) { };
   ACE::ACE_TYPE Type(void) { return ACE::HOST; };
   string StringType(void)  { return "HOST"; };
};


// --------------------------------------------------------------------
// GSSAPI credential
// --------------------------------------------------------------------
class Cred_GSSAPI : public Cred
{
public:
   Cred_GSSAPI(string t, string h, bool v) : Cred(t,h,v) { };
   ACE::ACE_TYPE Type(void) { return ACE::GSSAPI; };
   string StringType(void)  { return "GSSAPI"; };
};

// --------------------------------------------------------------------
// Certificate credential
// --------------------------------------------------------------------
class Cred_CERT : public Cred
{
public:
   Cred_CERT(string t, string h, bool v) : Cred(t,h,v) { };
   ACE::ACE_TYPE Type(void) { return ACE::CERT; };
   string StringType(void)  { return "CERT"; };
};

// --------------------------------------------------------------------
// List of credentials
// --------------------------------------------------------------------
class CredList
{
  protected:
   vector<Cred *>  list;
   vector<Cred *>::iterator pos;
  public:
   CredList(void)    { };
   void Add(Cred *c) { list.push_back(c); }
   Cred *First()     { pos = list.begin(); return *pos; };
   Cred *Next()      { pos++; return( (pos == list.end()) ? NULL : *pos); };
};

#endif
