// ====================================================================
// buffer.hh -- Buffer for I/O data.
// Class to implement a Buffer with a single write Cursor
// and read Cursor.  Buffers may have a prefix area reserved.
// The area may be filled with a timestamp plus constant information.
// It also supports characters to filter from the stream, and 
// a trigger character that causes a callback to write the contents
// of the buffer.  For example, a buffer may trigger on every newline
// character written into the buffer.
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.11 $ $Date: 2006/07/03 12:49:48 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include <ctype.h>
#include <stdexcept>
#include <time.h>

#include "buffer.h"
#include "slock.h"

// ====================================================================
// Initialize a Buffer object.
// ====================================================================

// Specifying the capacity...

Buffer::Buffer(int newcapacity)
      : capacity(newcapacity),
        bufferStart(new unsigned char[newcapacity])
{
   bufferAlloc  = true;
   prefix       = 0;
   rpos         = bufferStart;
   wpos         = bufferStart + prefix;
   bufferEnd    = bufferStart + capacity;
   overflow     = 0;
   filterParity = 0;
   filterMode   = Raw;
   filter       = 0;
   trigger      = 0;
   callback     = 0;
   state        = 0;
   tag          = ' ';
}

// Specifying another buffer...

Buffer::Buffer(const Buffer &rhs)
      : capacity(rhs.capacity),
        bufferStart(new unsigned char[rhs.capacity])
{
   int rhslen = rhs.wpos - rhs.bufferStart;
   
   bufferEnd = bufferStart + capacity;

   // Copy any contents from old buffer to new.

   if (rhslen > 0)
   {
      memmove(bufferStart, rhs.bufferStart, rhslen);
   }

   wpos = bufferStart + rhslen;
   rpos = bufferStart + (rhs.rpos - rhs.bufferStart);
   
   bufferAlloc  = true;
   prefix       = rhs.prefix;
   filterMode   = rhs.filterMode;
   filterParity = rhs.filterParity;
   overflow     = 0;
   state        = filterMode;
   filter       = rhs.filter;
   trigger      = rhs.trigger;
   callback     = rhs.callback;
   tag          = rhs.tag;
}

// Specifying a fixed area to use.

Buffer::Buffer(unsigned char *start, int len)
      : capacity(len),
        bufferStart(start)
{
   bufferEnd = bufferStart + capacity;

   bufferAlloc  = false;
   prefix       = 0;
   rpos         = bufferStart;
   wpos         = bufferStart + prefix;
   bufferEnd    = bufferStart + capacity;
   overflow     = 0;
   filterParity = 0;
   filterMode   = Raw;
   trigger      = 0;
   callback     = 0;
   state        = 0;
   tag          = ' ';
}

// ====================================================================
// Destroy a Buffer object.
// ====================================================================

Buffer::~Buffer(void)
{
   if (bufferAlloc) delete[] bufferStart;
}

// ====================================================================
// Set a trigger condition and object
// ====================================================================

void Buffer::Trigger(TriggerCallback &cb, unsigned char trig, unsigned char t)
{
   callback = &cb;
   trigger  = trig;
   tag      = t;
   if (filterMode == Raw) filterMode = Chars;
   state    = filterMode;
}

// ====================================================================
// Set filtering of ANSI escape sequences.
// ====================================================================

void Buffer::FilterMode(int mode, unsigned char cc)
{
   if (mode == Parity) filterParity = true;
   else                filterMode   = mode;

   if (mode == Chars)  filter       = cc;
   state = filterMode;
}

// ====================================================================
// Set a prefix area and contents.
// ====================================================================

int Buffer::Prefix(string newdata)
{
   return(Prefix(newdata.size(), newdata));
}

int Buffer::Prefix(int newpref, string newdata)
{
   int oldpref = prefix;
   int amt;
   
   prefix = newpref;
   amt = newdata.size();
   if (prefix < amt) amt = prefix;
   
   Clear();
   
   memmove(bufferStart, newdata.c_str(), amt);

   return oldpref;
}

// ====================================================================
// Assign various data types to a buffer object
// ====================================================================

// Assign another buffer...

Buffer &Buffer::operator=(const Buffer &rhs)
{
   if (this != &rhs)
   {
      int rhslen = rhs.wpos - rhs.bufferStart;
      if (rhslen > capacity) rhslen = capacity;
      if (rhslen > 0)
         memmove(bufferStart, rhs.bufferStart, rhslen);

      wpos = bufferStart + rhslen;         
      rpos = bufferStart + (rhs.bufferStart - rhs.rpos);
      if (rpos > wpos) rpos = wpos;
      
      prefix = rhs.prefix;
      state  = filterMode;
   }

   return *this;
}

// Assign a string...

Buffer &Buffer::operator=(const string &rhs)
{
   int rhslen = (int) rhs.size();
   
   if (rhslen > capacity) rhslen = capacity;
   
   if (rhslen > 0)
      memmove(bufferStart+prefix, rhs.c_str(), rhslen);
   wpos = bufferStart + prefix + rhslen;         
   rpos = bufferStart;
   state  = filterMode;

   return *this;
}

// Assign a C string...

Buffer &Buffer::operator=(const char *rhs) { return *this = (unsigned char *) rhs; }

Buffer &Buffer::operator=(const unsigned char *rhs)
{
   int rhslen = strlen((char *)rhs);
   
   if (rhslen > capacity) rhslen = capacity;
   
   if (rhslen > 0)
      memmove(bufferStart+prefix, rhs, rhslen);

   wpos = bufferStart + prefix + rhslen;
   rpos = bufferStart;
   state  = filterMode;

   return *this;
}

// Assign a single character...

Buffer &Buffer::operator=(char c)
{
   bufferStart[prefix] = c;
   wpos = bufferStart + prefix + 1;
   rpos = bufferStart;
   state  = filterMode;
   
   return *this;
}

// ====================================================================
// Write some data at the end of the buffer.
// ====================================================================

int Buffer::Append(unsigned char c)   
{ 
   int ret = 1;
   bool calltrigger = false;
   
   // Filter high bit if requested
   if (filterParity) c &= 0x7f;

   // Recognize ANSI escape sequences if enabled.
   // This recognizes all legal escape sequences, but will also
   // match illegal sequences.  We shouldn't see those in practice,
   // so this shouldn't hurt.
   switch(state)
   {
       case Raw:  // No filtering
          break;
          
       case Literal:            // Saw a \ escape, take this char literally.
          state = filterMode;
          break;

       case Ansi: // ANSI filtering state, pass all chars but switch state on escape.
          if (c == '\033') 
          {
             state = Escape;
             spos  = wpos;
          }
          // Fall into next case.

       case Chars: // Filter specified char only.
          if (c == '\\') { state = Literal; return 0; }
          if (c == filter) return 0;
          // If trigger character, trigger the callback.
          if (trigger && (c == trigger)) calltrigger = true;
          // If within a few chars of filling the buffer, trigger the callback
          if (wpos > (bufferEnd - TRIGGERZONE)) calltrigger = true;
          break;

       case Escape: // Escape state, look for CSI
          switch(c)
          {
              case '[':  // CSI
                 state = Parameter;
                 break;
                 
              case ']':  // OSC: let it through for now.
                 state = Ansi;
                 break;
                 
              default:   // Eat any other char
                 wpos = spos;
                 state = Ansi;
                 return 0;
          }
          break;
          
       case Parameter: // Parameter state: look for end of sequence.
          if (isdigit(c) || c == ';' || c == '?') break; 
          wpos = spos;
          state = Ansi;
          return 0;
   }

   // Add to buffer unless filtered.
   if (wpos >= bufferEnd) { overflow++; ret = 0;}  
   else *wpos++ = c;
   
   if (calltrigger)
   {
      if (*(wpos-1) != '\n') *wpos++ = '\n'; // Insure record terminated.
      callback->WriteTriggered(*this, tag);
      Clear();
   }

   return ret; 
};   

int Buffer::Append(int i)   
{ 
   unsigned char temp[17];
   unsigned char *p = temp+16;
   int sign = i;
   i = abs(i);
   
   *p = '\0';
   
   while (i > 0 && p > temp)
   {
      *--p = i % 10 + '0';
      i    = i / 10;
   }

   if (*p == '\0') *--p = '0';
   if (sign < 0) *--p = '-';
      
   return Append(p, strlen((char *)p));
};   

int Buffer::Append(const void *data, const int length)
{
   unsigned char *p = (unsigned char *) data;
   int  i  = length;
   int  dlen;
   
   while (i--)
   {
      dlen += Append(*p++);
   }

   return dlen;
}

int Buffer::Append(const string &s, const int len = -1)
{
   unsigned char *p = (unsigned char *) s.c_str();
   unsigned char *e = p + s.size();
   int i = (len > 0) ? len : s.size();
   int dlen;
   
   while (i--)
   {
      if (p < e) dlen += Append(*p++);
      else       dlen += Append((unsigned char) ' ');
   }
   
   return dlen;
}


// ====================================================================
// Resize the buffer 
// ====================================================================

void Buffer::Resize(const int newcapacity)
{
   unsigned char *temp = new unsigned char[newcapacity];
   
   int size = wpos - bufferStart;

   if (size > newcapacity)
   {
      overflow += size - newcapacity;
      size = newcapacity;
   }

   memmove(temp, bufferStart, size);
   
   rpos = temp + (rpos - bufferStart);
   
   if (bufferAlloc) delete[] bufferStart;
   
   bufferAlloc = true;
   bufferStart = temp;
   bufferEnd = bufferStart + newcapacity;
   capacity = newcapacity;
   wpos = bufferStart + size;
   state = filterMode;
}

// ====================================================================
// Move the read and write pointers
// ====================================================================

void Buffer::RSeek(int how, int pos)
{
   if (how == Beginning)     rpos = bufferStart + pos;
   else if (how == Current)  rpos += pos;
   else if (how == End)      rpos = wpos - pos;
   else                      rpos = bufferStart + prefix;
   
   if (rpos > wpos)        rpos = wpos;
   if (rpos < bufferStart) rpos = bufferStart;
}

void Buffer::WSeek(int how, int pos)
{
   if (how == Beginning)     wpos = bufferStart + pos;
   else if (how == Current)  wpos += pos;
   else if (how == End)      wpos = wpos - pos;
   else                      wpos = bufferStart + prefix;
   
   if (wpos < bufferStart) wpos = bufferStart;
   if (wpos > bufferEnd)   wpos = bufferEnd;
   if (rpos > wpos)        rpos = wpos;
   state = filterMode;   
}      

// ====================================================================
// Find a string in a buffer.  Return offset of first character of
// match, or null if not found or if search string is null.
// ====================================================================

unsigned char *Buffer::Find(const char *search, unsigned char *start)
{
   if (start < rpos) start = rpos;
   
   unsigned char *p   = start;  // Current buffer position.
   const    char *q   = search; // Match pointer.
   unsigned char *pos = 0;      // Beginning of tentative match.

   // Scan buffer trying to match string.

   while (p < wpos)
   {
      // If current char matched, bump to next pos.  If
      // no previous matches, mark beginning of match.
      if (*p == *q) 
      { 
         if (!pos) pos = p;
         q++;
         // If search string exhausted, return pos.  
         if (*q == '\0') return pos;
      }
      // Else char doesn't match, forget any tentative match.
      else
      {
         pos = 0;
         q   = search;
      }
      p++;
   }

   // Entire search string not matched, return null.
   return 0;
}

// ====================================================================
// Delete chars from buffer.
// ====================================================================

void Buffer::Delete(unsigned char *first, unsigned char *last)
{
   // Make sure pointers within bounds.
   if (first >= wpos) return;
   if (first < rpos) first = rpos;
   if (last <= rpos) return;
   if (last > wpos) last = wpos;
   
   // If deleting everything, clear the buffer.
   if (first == rpos && last == wpos) { Clear(); return; }
   
   // If deleting tail, just move the write pointer.
   if (last >= wpos) { wpos = first; return; }
   
   // If deleting head, just move the read pointer.
   if (first <= rpos) { rpos = last; return; }
   
   // Partial deletion, adjust the buffer.

   memmove(first, last, wpos-last);
   wpos -= wpos - last;
}

// ====================================================================
// Set the timestamp format for the buffer.
// ====================================================================

void Buffer::TimeStampFormat(string fmt)
{
   char sample[256];
   struct tm nowtm;

   // Store the time format for this buffer.
   tsformat = fmt;

   // Get a sample timestamp to set the prefix area size.
   time(&lastts);
   localtime_r(&lastts, &nowtm);
   int cnt = strftime(sample, sizeof(sample), tsformat.c_str(), &nowtm);
   Prefix(cnt, sample);
}
   
// ====================================================================
// Insert the current time in buffer, leaving remainder undisturbed.
// Because consoles can produce many hundreds of lines per second
// and the strftime operation is expensive, only do it if the
// value is changes.  No more than one thread may timestamp the
// same buffer at the same time, so the caller must lock the buffer
// if multiple threads may do this; this should be quite rare.
// ====================================================================

void Buffer::TimeStamp(void)
{
   time_t now;
   struct tm nowtm;
   char timestamp[256];
   
   time(&now);
   if (now != lastts)
   {
      localtime_r(&now, &nowtm);
      int cnt = strftime(timestamp, sizeof(timestamp), tsformat.c_str(), &nowtm);
      if (prefix < cnt) cnt = prefix;
      memmove(bufferStart, timestamp, cnt);
      lastts = now;
   }
   
   return;
}

// ====================================================================
// Return a pointer to the first character, after locating the first
// newline in the buffer and replacing it with null, and positioning
// rpos immediately after it.  Has the effect of returning a pointer
// to a c string and deleting it from the buffer.
// ====================================================================

unsigned char *Buffer::GetLine(void)
{
   unsigned char *start = rpos;
   
   for ( ; rpos < wpos && *rpos != '\n'; rpos++) ;
   
   *rpos++ = '\0';
   
   return start;
}

// ====================================================================
// Return true if the current buffer contains a newline.
// ====================================================================

bool Buffer::Haveline(void)
{
   unsigned char *p = rpos;

   for ( ; p < wpos; p++) if (*p == '\n') return true;
   
   return false;
}
