// ====================================================================
// buffer.hh -- Buffer for I/O data.
// Class to implement a Buffer with a single write Cursor
// and read Cursor.  Buffers may have a prefix area reserved.
// The area may be filled with a timestamp plus constant information.
// It also supports characters to filter from the stream, and 
// a trigger character that causes a callback to write the contents
// of the buffer.  For example, a buffer may trigger on every newline
// character written into the buffer.
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.2 $ $Date: 2006/07/03 12:49:51 $ $Author: boeheim $
// ******************************************************************** //

#ifndef buffer_h
#define buffer_h

#include <string>
#include <time.h>

class File;
class Buffer;

#define TRIGGERZONE 20

// ====================================================================
// Class that a method using callbacks should inherit from.
// Override the WriteTriggered method to dispose of the buffer.
// ====================================================================

class TriggerCallback
{
public:
   virtual void WriteTriggered(Buffer &record, char tag) {};
};

// ====================================================================
// Class for a buffer.
//====================================================================

class Buffer
{
protected:
   int capacity;                // Size of allocated Buffer.
   bool bufferAlloc;            // True if we allocated buffer.
   unsigned char *bufferStart;  // Pointer to first byte.
   unsigned char *bufferEnd;    // Pointer to next byte beyond end.
   unsigned char *rpos;         // Read Cursor within Buffer.
   unsigned char *wpos;         // Write Cursor within Buffer.
   unsigned char *spos;         // Save wpos for restoration.
   int overflow;                // Quantity of data lost.
   int prefix;                  // Length of prefix area.
   time_t lastts;               // Time last timestamp put in prefix area.
   string tsformat;             // Format for timestamp in strftime format.

   int filterMode;              // Type of filtering to do.
   int state;                   // State in character recognition.
   bool filterParity;           // True if filtering high bit.
   unsigned char filter;        // Character to remove from buffer.
   unsigned char trigger;       // Character that causes file write.
   unsigned char tag;           // Tag to pass to callback.
   TriggerCallback *callback;   // Object to handle trigger.

public:
   Buffer(int s=1024);          // Constructor.
   Buffer(const Buffer &);
   Buffer(unsigned char *start, int len);

   ~Buffer(void);               // Destructor.
   void Resize(const int);      // Adjust size of buffer.

   int  Prefix(void) { return prefix; };
   int  Prefix(int newpref, string newdata = "");
   int  Prefix(string newdata);

   enum {Beginning, End, Current, Data};
   void RSeek(int how, int pos = 0); // Move rpos in buffer.
   void WSeek(int how, int pos = 0); // Move wpos in buffer.

   void Trigger(TriggerCallback &cb, unsigned char trig = 0, unsigned char tag = ' ');

   enum {Raw, Chars, Literal, Ansi, Parity, Escape, Parameter};
   void FilterMode(int mode = Raw, unsigned char cc = '\r');

   // assignment operators

   Buffer &operator=(const Buffer &);
   Buffer &operator=(const string &);
   Buffer &operator=(const unsigned char *);
   Buffer &operator=(const char *p);
   Buffer &operator=(char);

   // Append operators and function

   int Append(const string &s, const int len);
   int Append(const void *d, const int l);
   int Append(unsigned char c);
   int Append(int i);

   Buffer &operator+=(const Buffer &rhs)  { Append(rhs.RPos(), rhs.Size());  return *this; };
   Buffer &operator+=(const string &rhs)  { Append(rhs.c_str(), rhs.size()); return *this; };
   Buffer &operator+=(const char *rhs)    { Append(rhs, strlen(rhs));        return *this; };
   Buffer &operator+=(unsigned char c)    { Append(c);                       return *this; };
   Buffer &operator+=(char c)             { Append((unsigned char) c);       return *this; };
   Buffer &operator+=(int i)              { Append(i);                       return *this; };

   // Access functions
   
   char operator[](int i) { return bufferStart[i]; };
   
   int   Capacity  (void) const { return capacity; };
   int   Size      (void) const { return wpos - rpos; };
   int   Dirty     (void) const { return wpos > bufferStart + prefix && rpos < wpos; };
   int   Space     (void) const { return bufferEnd - wpos;};
   int   Overflowed(void)       { int t = overflow; overflow = 0; return t; };
   // Non-destructive (do not remove chars from buffer).
   unsigned char *WPos(void) const { return wpos; };
   unsigned char *RPos(void) const { return rpos; };
   unsigned char *Cstr(void) const { *wpos = '\0'; return rpos; }
   unsigned char *Find(const char *search, unsigned char *start=0);
   bool  Haveline(void);
   // Destructive (remove chars from buffer by moving rpos and/or wpos).
   void  Clear  (void) { rpos = bufferStart; wpos = bufferStart + prefix; };
   void  Delete(unsigned char *first, unsigned char *last);
   unsigned char  GetChar(void) { return (rpos < wpos) ? *rpos++ : 0; };
   unsigned char *GetLine(void);

   // Put current date and/or time in prefix area.
   void TimeStampFormat(string fmt);
   void TimeStamp(void);
};

// ====================================================================
// Class for a small buffer allocated inline.
// ====================================================================

class SmallBuffer : public Buffer
{
private:
   unsigned char data[128];
public:
   SmallBuffer(void) 
         : Buffer(data, 128) {};
};

#endif
