// ====================================================================
// client.cc - Base class for clients connections to console server.
//             A client is a Socket and is a Reader
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.27 $ $Date: 2007/04/16 01:57:07 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include "acl.h"
#include "client.h"
#include "consoled.h"
#include "dispatcher.h"
#include "entry.h"
#include "file.h"
#include "global.h"
#ifdef HAVE_LIBGSSAPI
#include "gssapi.h"
#endif
#include "host.h"
#include "psstream.h"
#include "syslog.h"
#include "utility.h"

#include <algorithm>
#include <arpa/inet.h>
#include <ctype.h>
#include <errno.h>
#include <iostream>
#include <netdb.h>
#include <queue>
#ifdef HAVE_SECURITY_PAM_APPL_H
extern "C" 
{
#include <security/pam_appl.h>
}
#endif
#ifdef HAVE_PAM_PAM_APPL_H
extern "C" 
{
#include <pam/pam_appl.h>
}
#endif
#include <sys/socket.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#ifndef SHUT_WR
#define SHUT_WR 2
#endif

string FmtCtl(string);

// ====================================================================
// Storage for global function tables and flag to initialize them on
// first use.
// ====================================================================

  static void (Client::*fsm_table[Client::FSM_SIZE])(char);

  static void (Client::*cmd_table[256])(char);

  static bool initialized = false;

// ====================================================================
// Constructor -- Create a new client.  The socket is assumed to be 
//                already open and connected to the client, so
//                we can write messages to it.
//                Also used to initialize the state tables for later.
// ====================================================================

Client::Client(void)
{
   wantWrite = canWrite = false; // No writing until attach
   terse         = false;       // Normal output.
   lastActivity  = time(NULL);  // Last activity when created.
   idleForce     = DAEMON.IdleForce();  // Time until disconnect
   idleWarn      = DAEMON.IdleWarn();   // Time until warning
   escape        = DAEMON.Escape(); // Set default escape
   console       = NULL;        // Initially, no console
   state         = S_USER1;     // Begin by waiting for userid.
   input.Clear();               // Initialize buffers.

   // Just once, initialize the static state tables.

   if (!initialized)
   {
      fsm_table[S_CATTN ] = &Client::fsm_cattn ; 
      fsm_table[S_CESC  ] = &Client::fsm_cesc  ;  
      fsm_table[S_CMD   ] = &Client::fsm_cmd   ;   
      fsm_table[S_CONN  ] = &Client::fsm_conn  ;
      fsm_table[S_ESC1  ] = &Client::fsm_esc1  ;  
      fsm_table[S_HALT1 ] = &Client::fsm_halt1 ; 
      fsm_table[S_QUIT  ] = &Client::fsm_quit  ;
      fsm_table[S_QUOTE ] = &Client::fsm_quote ;
      fsm_table[S_USER1 ] = &Client::fsm_user  ;
      fsm_table[S_USER2 ] = &Client::fsm_user  ;
      fsm_table[S_USER3 ] = &Client::fsm_user  ;
      fsm_table[S_NORMAL] = 0;          // Handle inline.
      fsm_table[S_CLOSE ] = 0;          // Handle inline.

      for (int i=0; i<256; i++) cmd_table[i] = &Client::cmd_unknown;  
                                                           
      // The upper-case varient sets the 'terse' flag which causes
      // some routines to produce terse, parsable output.

      cmd_table[(int) 'a'] = cmd_table[(int) 'A']  = &Client::cmd_attach;     
      cmd_table[(int) 'b'] = cmd_table[(int) 'B']  = &Client::cmd_hb;
      cmd_table[(int) 'c'] = cmd_table[(int) 'C']  = &Client::cmd_connect;
      cmd_table[(int) 'd'] = cmd_table[(int) 'D']  = &Client::cmd_down;       
      cmd_table[(int) 'e'] = cmd_table[(int) 'E']  = &Client::cmd_descape;    
      cmd_table[(int) 'f'] = cmd_table[(int) 'F']  = &Client::cmd_attach;      
      cmd_table[(int) 'g'] = cmd_table[(int) 'G']  = &Client::cmd_locate;
      cmd_table[(int) 'h'] = cmd_table[(int) 'H']  = &Client::cmd_help;       
      cmd_table[(int) 'i'] = cmd_table[(int) 'I']  = &Client::cmd_info;      
      cmd_table[(int) 'k'] = cmd_table[(int) 'K']  = &Client::cmd_timeout;      
      cmd_table[(int) 'l'] = cmd_table[(int) 'L']  = &Client::cmd_halt1;      
      cmd_table[(int) 'm'] = cmd_table[(int) 'M']  = &Client::cmd_macro;
      cmd_table[(int) 'o'] = cmd_table[(int) 'O']  = &Client::cmd_open;       
      cmd_table[(int) 'p'] = cmd_table[(int) 'P']  = &Client::cmd_speed;       
      cmd_table[(int) 'q'] = cmd_table[(int) 'Q']  = &Client::cmd_quit;        
      cmd_table[(int) 'r'] = cmd_table[(int) 'R']  = &Client::cmd_replay;    
      cmd_table[(int) 's'] = cmd_table[(int) 'S']  = &Client::cmd_attach;
      cmd_table[(int) 't'] = cmd_table[(int) 'T']  = &Client::cmd_flow;       
      cmd_table[(int) 'u'] = cmd_table[(int) 'U']  = &Client::cmd_status;      
      cmd_table[(int) 'v'] = cmd_table[(int) 'V']  = &Client::cmd_version;    
      cmd_table[(int) 'w'] = cmd_table[(int) 'W']  = &Client::cmd_who;        
      cmd_table[(int) 'x'] = cmd_table[(int) 'X']  = &Client::cmd_examine;
      cmd_table[(int) 'y'] = cmd_table[(int) 'Y']  = &Client::cmd_control;
      cmd_table[(int) 'z'] = cmd_table[(int) 'Z']  = &Client::cmd_health;
      cmd_table[(int) '%']                   = &Client::cmd_servers;       
      cmd_table[(int) '@']                   = &Client::cmd_reload;       
      cmd_table[(int) '?']                   = &Client::cmd_help;       
      cmd_table[(int) '.']                   = &Client::cmd_disconnect; 
      cmd_table[(int) ' '] = cmd_table[(int) '\n'] = &Client::cmd_abort;     
      cmd_table[(int) ' '] = cmd_table[(int) '\r'] = &Client::cmd_abort;     
      cmd_table[(int) '\\']                  = &Client::cmd_quote;     

      initialized = true;
   }
}

// ====================================================================
// Loop reading the client socket and passing through most characters.
// On recognition of an escape sequeence, execute a command.
// ====================================================================

void Client::ReaderProc(void)
{
   try {                        // Catch all exceptions

   // ----------------------------------------------------------------
   // Loop until shut down or error on socket.
   // ----------------------------------------------------------------

   for (;;)
   {
      // --------------------------------------------------------------
      // Read socket.  If I/O error (including unexpected close from
      // other end), return, which will end this thread.
      // --------------------------------------------------------------

      if (state == S_CLOSE) break;

      try
      {
        input.Clear();
	Receive(input);
      }
      catch (File::IOErr(e))
      {
  	 if (e.reason == EAGAIN) { cout << "EAGAIN signaled\n"; continue; }
         
         SYSLOG << Severity(Syslog::error)
                << e.routine 
                << ": " << e.filename 
                << ": " << e.message 
                << endl;

         break;
      }

      lastActivity = time(NULL);    // Record that we heard from him.

      if (input.Size() == 0) { state = S_CLOSE; break; }

      // --------------------------------------------------------------
      // Process all of the characters through the finite state machine.
      // --------------------------------------------------------------         
      while (unsigned char cc = input.GetChar())
      {
         cc &= 127;             // always clear parity from the network 

         // ----------------------------------------------------------
         // In-line handling of standard and most frequent case, a
         // normal character when not processing an escape sequence.
         // ----------------------------------------------------------
         
         if (state == S_NORMAL)
	 {
	    // --------------------------------------------------------
            // If it is an escape sequence shift states
            // -------------------------------------------------------- 
            
             if (cc == escape[0])
             {
                state = S_ESC1;
                continue;
             }

            // -------------------------------------------------------
            // If we can write, write to the console.
            // -------------------------------------------------------

             if (console && canWrite && console->IsUp())
             {
                console->Write(cc);
                continue;
             }

             // -------------------------------------------------------
             // If input from client on read-only line, give them
             // feedback on actions.
             // -------------------------------------------------------

             if (cc == '\r' || cc == '\n')
             {
                string temp = "[read-only -- use " +
                   FmtCtl(escape) +
                   "? for help]\r\n";
                Send(temp);
             }
             continue;
         }
      

	 // ------------------------------------------------------------
         // If the state changed to CLOSE, exit the loop.
         // ------------------------------------------------------------

         if (state == S_CLOSE) 
         {
            break;
         }

         // ------------------------------------------------------------
         // Otherwise, call the routine to handle this state.
         // ------------------------------------------------------------

         (this->*fsm_table[state])(cc);

      }
   }

   } // end of try
   
   catch (File::IOErr(e))
   {
      SYSLOG << Severity(Syslog::error)
             << "Thread ending because of uncaught I/O error "
             << e.routine 
             << ": " << e.filename 
             << ": " << e.message 
             << endl;
      try 
      {
         Send("[I/O error: ");
         Send(e.filename);
         Send(": ");
         Send(e.message);
         Send("]\r\n");
      }
      catch(...)
      {
      }
   }
   catch(string s)
   {
      SYSLOG << Severity(Syslog::error)
             << "Thread ending because of uncaught exception "
             << s
             <<endl;
      try 
      {
         Send("[");
         Send(s);
         Send("]\r\r");
      }
      catch(...)
      {
      }
   }
   catch(...)
   {
      SYSLOG << Severity(Syslog::error)
             << "Thread ending because of uncaught unknown exception"
             << endl;
      try 
      {
         Send("[abort]\r\n");
      }
      catch(...)
      {
      }
   }
   
   // -----------------------------------------------------------------
   // Drop the connection to the console.  
   // -----------------------------------------------------------------

   DropConsole();
   Close();
   
   // -----------------------------------------------------------------
   // Upon return from this routine, the thread will finish.
   // The next time the dispatcher drives the idlecheck routine
   // it will return true because the thread is no longer running,
   // which will cause the dispatcher to delete this object.
   // -----------------------------------------------------------------
}

// ====================================================================
// Get the next word from the client.  No state changes allowed
// until the whitespace at the end of the word.  The first 
// whitespace character is read from the stream and discarded by
// this routine.
// ====================================================================

int Client::GetNum(bool echo)
{
   string word;
   
   word = GetWord(echo, true);
   if (word.size() == 0) return -1;
   return StrToInteger(word);
}

string Client::GetWord(bool echo, bool num)
{
   unsigned char cc;
   string word;

   for (;;)
   {
      cc = input.GetChar();
      
      if (cc == 0) 
      {
         input.Clear();
         Receive(input);
         if (input.Size() == 0) throw string("Unexpected close of client socket");
         continue;
      }
      
      if (isspace(cc)) break;

      if (num && ! isdigit(cc)) { Send("\007"); continue; }

      if (echo) Send(cc);
      word += cc;
   }

   return word;
}

// ====================================================================
// Authenticate the user by checking tickets, passwords, etc.
// ====================================================================

bool Client::ValidateCred(Cred *cred)
{
   if (cred->Type() == ACE::GSSAPI)
   {
#ifdef HAVE_LIBGSSAPI
      Send("{g}");              // Signal client to start handshake
      GSSAPI g(this);
      if (g.Accept())
      {
         cred->SetValid();
         cred->SetToken(g.ClientName());
      }
#else
      SYSLOG << Severity(Syslog::error)
             << "Client requested GSSAPI authentication but server does not support it"
             << endl;
#endif
   }
   else if (cred->Type() == ACE::CERT)
   {
   }
   return cred->Valid();
}


// bool Client::Authenticate(void)
// {
//    string password;
// 
//    if (authenticated) return true;
// 
// #ifdef HAVE_LIBPAM
//    pam_handle_t *pamh = NULL;
//    int retval;
//    extern int pcstub(int, const struct pam_message **, 
//                           struct pam_response **, void *);
//    struct pam_conv conv = { pcstub, this };
//    
//    retval = pam_start("console", userid.c_str(), &conv, &pamh);
//    
//    pam_set_item(pamh, PAM_RHOST, Peername().c_str());
//    pam_set_item(pamh, PAM_RUSER, Userid().c_str());
//    
//    if (retval == PAM_SUCCESS) 
//       retval = pam_authenticate(pamh, 0);
//    
//    if (retval == PAM_SUCCESS)
//       retval = pam_acct_mgmt(pamh, 0);
//    
//    authenticated = (retval == PAM_SUCCESS);
//    
//    pam_end(pamh, retval);
//    
//    return authenticated;
//    
// #else
//    Send("pass]\r\n");
//    password = GetWord();
//    Send("[");
// 
//    // Check the password
// 
//    if (GLOBAL.Verbose())
//    {
//       SYSLOG << Severity(Syslog::debug)
//              << "User " << UserAtHost()
//              << " has authenticated"
//              << endl;
//    }
// 
// #endif
//    authenticated = true;
// 
//    return true;
// }
// 
// #ifdef HAVE_LIBPAM
//    int pcstub(int num_msg, 
//               const struct pam_message **prompt, 
//               struct pam_response **response, 
//               void *ptr)
//    {
//       return ((Client *)ptr)->PamConversation(num_msg, prompt, response);
//    }
// 
// int Client::PamConversation(int num_msg, 
//                             const struct pam_message **prompt,
//                             struct pam_response **response)
// {
//    string temp;
//    
//    Send("Authenticate]\r\n[");
//    Send(prompt[0]->msg);
//    temp = GetWord();
//    Send(" - ");
// 
//    *response = (struct pam_response *)malloc(sizeof(pam_response));
// 
//    (*response)->resp = strdup(temp.c_str());
//    (*response)->resp_retcode = 0;
// 
//    return 0;
// }
// #endif

// ====================================================================
// Offer write access to this client.
// ====================================================================

bool Client::OfferWrite(void)
{
   if (! wantWrite) return false;
   
   Send("[Attached readwrite]\r\n");

   Buffer info;

   info.TimeStampFormat(console->TSFormat());
   
   info += "[Attached readwrite by ";
   info += UserAtHost();
   info += "]\n";
   
   info.TimeStamp();
   console->WriteLog(info);
   
   SYSLOG << Severity(Syslog::info)
          << "Console "
          << console->Name()
          << ": Attached readwrite by "
          << UserAtHost()
          << endl;
   
   canWrite = true;
   
   return true;
}

// ====================================================================
// Drop a console
// ====================================================================

void Client::DropConsole(void)
{
   if (!console) return;

   Buffer info;
   info.TimeStampFormat(console->TSFormat());
   
   info += "[Detached by ";
   info += UserAtHost();
   info += "]\n";
   
   info.TimeStamp();
   console->WriteLog(info);

   SYSLOG << Severity(Syslog::info)
          << "Console "
          << console->Name()
          << ": Detached by "
          << UserAtHost()
          << endl;
      
   console->DropClient(this);   // Remove from console's list.
   
   if (canWrite)
   {
      wantWrite = false;     // We're not writing any more.
      canWrite  = false;
      
      console->OfferWrite(); // See if anyone else wants to.
   }

   console = NULL;
}
   
// ********************************************************************
// Finite State Machine routines, in alphabetic order.
// ********************************************************************

// ====================================================================
// CATTN state: Redefine the escape sequence.
// ====================================================================

void Client::fsm_cattn(char cc)
{
   escape = cc;

   Send(FmtCtl(string(1,cc)));

   state = S_CESC;
   return;
}

// ====================================================================
// CESC state: Redefine the escape sequence, part 2.
// ====================================================================

void Client::fsm_cesc(char cc)
{
   escape += cc;

   state = S_NORMAL;

   Send(FmtCtl(string(1,cc)) + " ok]\r\n");
   return;
}

// ====================================================================
// CONN state: get char for type of connection.
// ====================================================================

void Client::fsm_conn(char cc)
{
   cc = tolower(cc);

   Send("\b\b\b\b\b\b\b\b");
   
   if (cc == 'a')
   {
      Send("read/write: ");
      cmd_attach(cc);
   }
   else if (cc == 'f')
   {
      Send("force read/write: ");
      cmd_attach(cc);
   }
   else if (cc == 's')
   {
      Send("readonly: ");
      cmd_attach(cc);
   }
   else
      Send("aborted]\r\n");
   
   state = S_NORMAL;

   return;
}

// ====================================================================
// CMD state: Execute the appropriate routine for this
//            (single-character) command.
// ====================================================================

void Client::fsm_cmd(char cc)
{
   state = S_NORMAL;
   terse = (isupper(cc)) ? true : false;
   (this->*cmd_table[(int) cc])(cc);
   return;
}

// ====================================================================
// ESC1 state: Recognizing the escape sequence
//             In this state we have seen the first of the two
//             character sequence and are looking for the second.
// ====================================================================

void Client::fsm_esc1(char cc)
{
   // -----------------------------------------------------------------
   // If received escape character 2, look for the following
   // command sequence next.
   // -----------------------------------------------------------------
   
   if (cc == escape[1])
   {
      Send("\r\n[");
      state = S_CMD;
      return;
   }

   // -----------------------------------------------------------------
   // Received escape character1 again, pass the first one and
   // keep looking for the second.
   // -----------------------------------------------------------------

   if (cc == escape[0])
   {
      if (console && canWrite && console->IsUp())  // Only if we're the writer! 
      {
         console->Write(escape, 1);
      }
      return;
   }
   
   // -----------------------------------------------------------------
   // Received any other character.  Pass both characters and
   // go back to normal state.
   // -----------------------------------------------------------------
   
   state = S_NORMAL;
   if (console && canWrite && console->IsUp())
   {
      console->Write(escape, 1);
      console->Write(&cc, 1);
   }
   return;
}


// ====================================================================
// HALT state: try to send a halt (break) to the console.
// ====================================================================

void Client::fsm_halt1(char cc)
{
   state = S_NORMAL;
   
   if (!console)
   {
      Send("not connected]\r\n");
      return;
   }

   if (cc != '1')
   {
      Send("aborted]\r\n");
      return;
   }
   
   Send("- ");

   try{
      console->SendBreak();
      Send("sent]\r\n");
   }
   catch (IOErr(e)){
      Send("failed]\r\n");
   }
   
   return;
}

// ====================================================================
// QUIT state: waiting for confirmation of quit
// ====================================================================

void Client::fsm_quit(char cc)
{
   state = S_NORMAL;

   if (tolower(cc) != 'y')
   {
      Send("aborted]\r\n");
      return;
   }

   bool authorized = false;   
   ACL *quitacl = DAEMON.FindACL("shutdown");

   if (quitacl)
   {
      authorized = quitacl->Allow(&credlist);
   }
   if (!authorized)
   {
      Send("not authorized]\r\n");
      return;
   }
   
   Send("Accepted]\r\n");

   SYSLOG << Severity(Syslog::info)
          << "Quit command sent by "
          << UserAtHost()
          << endl;
   
   DISPATCHER.Stop("Quit command sent by " + UserAtHost());
   
   return;
}

// ====================================================================
// QUOTE state: collect 3 octal digits and send the corresponding
//              character to the console.
// ====================================================================

void Client::fsm_quote(char cc)
{
   // If a digit, accumulate it, but not more than 3.
   
   if (isdigit(cc))
   {
      temp += cc;

      if (temp.Size() < 3) return;
   }
   
   // Convert octal digits to character.
   
   char newc = '\0';
   for (int i = temp.Size() - 1; i >= 0; i--)
   {
      newc *= 8;
      newc += temp[i];
   }
   
   // Echo to client and write to console.

   temp += "]\r\n";
   
   Send(temp);

   if (console && canWrite && console->IsUp())
   {
      console->Write(newc);
   
      // If terminated by non-digit, make sure to write that out.
   
      if (!isdigit(cc)) console->Write(&cc, 1);
   }
   
   state = S_NORMAL;

   return;
}

// ====================================================================
// USER state: looking for the userid leadin characters followed by
//             the userid string.  The only entry to this state is
//             the initial state on the creation of a client connection.
//             There is no way to change userids once the connection
//             is established.  There is no console selected yet.
// ====================================================================

void Client::fsm_user(char cc)
{
   temp.Clear();
   
   try
   {
      if (state == S_USER1)
      {
         if (cc != escape[0]) throw state;

         state = S_USER2;
      }
      
      else if (state == S_USER2)
      {
         if (cc != escape[1]) throw state;

         state = S_USER3;
      }

      else  // state == S_USER3
      {
         if (cc != ';') throw state;

         state = S_USER4;
         
         userid = GetWord();

         credlist.Add(new Cred_HOST(userid, Peername(), true));
#ifdef HAVE_LIBGSSAPI
         Cred_GSSAPI *c = new Cred_GSSAPI(userid, Peername(), false);
         c->SetCallback(this);
         credlist.Add(c);
#endif
         
         // Got connected, switch to normal state to await commands.

         Send("[ok]\n");       // Acknowledge
         state = S_NORMAL;      // Command state next.
         return;
      }
   }
   
   catch (int s)
   {
      if (s != S_USER4) Send("[Protocol error]\n");
      else Send("[Access denied]\n");
  
      state = S_CLOSE;
   }
   
}

             
// ********************************************************************
// Command routines to perform a command following an escape sequence.
// Called from fsm_cmd above after recognizing an escape + character.
// ********************************************************************

// ====================================================================
// Abort escape sequence.
// ====================================================================

void Client::cmd_abort(char cc)
{ 
   Send("ignored]\r\n");
   return;
}


// ====================================================================
// Attempt to attach a console read/write.
// Entered on one of the following conditions:
// 
//                            Have Console  Have Console
//                No Console  Change State Change to new console
//   Request r/w   (esc-a)      (esc-a)       (esc-ca)
//   Force   r/w   (esc-f)      (esc-f)       (esc-cf)
//   Request r/o   (esc-s)      (esc-s)       (esc-cs)
// ====================================================================

void Client::cmd_attach(char cc)
{
   Client *writer;
   string consname;
   bool addclient = false;
      
   state = S_NORMAL;            // After attaching, back to normal state.

   // Acquire a new console if necessary

   if (!console)
   {
      addclient = true;
      
      consname = GetWord(false);
     
      if (GLOBAL.Verbose())
      {
         SYSLOG << Severity(Syslog::debug)
                << "User " << UserAtHost()
                << " requested console " 
                << consname
                << endl;
      }

      // Try to locate the named console.
   
      console = DISPATCHER.FindConsole(consname);
   
      // If not local, see if it is a known remote console.

      if (!console)
      {
         BConsole *rconsole;
         
         if ((rconsole = DISPATCHER.FindRConsole(consname)))
         {
            string temp = consname + ": redirect: " + rconsole->Server() + "]\r\n";
            Send(temp);

            // Client should close connection and contact new server.
            
            state = S_CLOSE;
            
            return;
         }
      }
      
      if (!console)
      {
         Send(consname);
         Send(": unknown]\r\n");
         
         return;
      }
   }

   // Check all credentials that the user has presented us against the
   // access control list for the requested console.

   if (! console->Authorized(&credlist))
   {
      Send(consname);
      Send(": not allowed]\r\n");

      SYSLOG << Severity(Syslog::error)
             << "User " 
             << UserAtHost()
             << " denied access to console "
             << console->Name()
             << endl;

      console = NULL;
      
      return;
   }

   // -----------------------------------------------------------------
   // We are now attached.  All responses before this indicated not
   // attached, all messages after this indicate attached.
   // -----------------------------------------------------------------

   Send(consname);
   Send(": ");
   
   if (addclient) console->AddClient(this); // Add to the end
      
   wantWrite = (tolower(cc) == 'a' || tolower(cc) == 'f') ? true : false;

   try
   {
      // Check that the console is up.

      if (!console->IsUp())
      {
         throw string("Attached, line down on " + console->Server());
      }

      // Check that the console can be written to.
      
      if (console->ReadOnly())
      {
         throw string("Attached readonly on " + console->Server());
      }

      // If there is no current writer to the console and we want to write,
      // we've got it now.

      if (!(writer = console->Writer()) && wantWrite)
      {
         canWrite = true;          // Do it if no one else is attached. 
         throw string("Attached readwrite on " + console->Server());
      }
   
      // If we were already the writer, decide if we want it

      if (writer == this)
      {
         if (wantWrite)
         {
            canWrite = true;
            throw string("ok");
         }
         else
         {
            canWrite = false;
            console->OfferWrite();
            throw true;
         }
      }
   
      // If only wanted readonly, we're done now.
      
      if (tolower(cc) == 's') throw true;

      // If someone else is the writer, bump if a force was requested.
      
      if (tolower(cc) == 'f')
      {
         writer->canWrite  = false;
         canWrite = true;
      
         writer->Send("\r\n[Forced readonly by " + userid + "]\r\n");

         throw string("Attached readwrite on " + console->Server() + ", bumped " + writer->UserAtHost());
      }

      // Otherwise, mark client as wanting write, but deny it. 

      else
      {
         throw string("Attached readonly on " + console->Server() + ", " + writer->UserAtHost() + " is readwrite");
      }
   }
   catch(bool)
   {
      return;
   }
   catch(string s)
   {
      Send(s + "]\r\n");

      SYSLOG << Severity(Syslog::info)
             << "Console " << console->Name()
             << ": " << s << " by " << UserAtHost()
             << endl;
      
      Buffer info;
      info.TimeStampFormat(console->TSFormat());
      
      info += '[';
      info += s;
      info += " by ";
      info += UserAtHost();
      info += "]\n";
      
      info.TimeStamp();
      console->WriteLog(info);
   }

   return;
}

// ====================================================================
// Connect to a new console.
// ====================================================================

void Client::cmd_connect(char cc)
{ 
   if (!terse) Send("Connect (a|f|s):");
   DropConsole();
   state = S_CONN;              // Looking for attach char now.
   return;
}

// ====================================================================
// Display and manipulate control signals on the line.
// ====================================================================

void Client::cmd_control(char cc)
{
   if (!console)
   {
      Send("not attached]\r\n");
      return;
   }
   
   Buffer buffer;
   vector<bool> sigState;
   
   sigState = console->GetSignals();
   
   buffer += (sigState[TTYFile::SigRTS]) ? "+RTS" : "-RTS";
   buffer += (sigState[TTYFile::SigCTS]) ? "+CTS" : "-CTS";
   buffer += (sigState[TTYFile::SigDSR]) ? "+DSR" : "-DSR";
   buffer += (sigState[TTYFile::SigCAR]) ? "+DCD" : "-DCD";
   buffer += (sigState[TTYFile::SigDTR]) ? "+DTR" : "-DTR";
   buffer += (sigState[TTYFile::SigRI ]) ? "+RI"  : "-RI";
   
   Send("line status ");
   Send(buffer);
   Send("]\r\n");

   Send("[Toggle (Rts|Cts|Dcd|dSr|dTr|rI|None): ");

   while ((cc = input.GetChar()) == 0)
   {
      input.Clear();
      Receive(input);
   }
   
   switch(cc)
   {
       case 'r':
       case 'R':
          console->SetRTS(not sigState[TTYFile::SigRTS]);
          Send("RTS]\r\n");
          break;
          
       case 'c':
       case 'C':
          console->SetCTS(not sigState[TTYFile::SigCTS]);
          Send("CTS]\r\n");
          break;

       case 'd':
       case 'D':
          console->SetCAR(not sigState[TTYFile::SigCAR]);
          Send("DCD]\r\n");
          break;
          
       case 's':
       case 'S':
          console->SetDSR(not sigState[TTYFile::SigDSR]);
          Send("DSR]\r\n");
          break;
          
       case 't':
       case 'T':
          console->SetDTR(not sigState[TTYFile::SigDTR]);
          Send("DTR]\r\n");
          break;
          
       case 'i':
       case 'I':
          console->SetRI(not sigState[TTYFile::SigRI]);
          Send("RI]\r\n");
          break;
          
       default:
          Send("None]\r\n");
          break;
   }
   
   sigState = console->GetSignals();
   
   buffer = "";
   buffer += (sigState[TTYFile::SigRTS]) ? "+RTS" : "-RTS";
   buffer += (sigState[TTYFile::SigCTS]) ? "+CTS" : "-CTS";
   buffer += (sigState[TTYFile::SigDSR]) ? "+DSR" : "-DSR";
   buffer += (sigState[TTYFile::SigCAR]) ? "+DCD" : "-DCD";
   buffer += (sigState[TTYFile::SigDTR]) ? "+DTR" : "-DTR";
   buffer += (sigState[TTYFile::SigRI ]) ? "+RI"  : "-RI";
   
   Send("[new status  ");
   Send(buffer);
   Send("]\r\n");
}

// ====================================================================
// Redefine escape characters.
// ====================================================================

void Client::cmd_descape(char cc)
{ 
   // Shift FSM into state to look for new chars.
   state = S_CATTN;
   Send("redef: ");
   return;
}

// ====================================================================
// Disconnect the client.
// ====================================================================

void Client::cmd_disconnect(char cc)
{
   Send("disconnect]\r\n");

   if (console && canWrite && console->IsUp()) console->FlowOn();
   
   Shutdown(SHUT_WR);
   state = S_CLOSE;
   return;
}

// ====================================================================
// Down a console.
// ====================================================================

void Client::cmd_down(char cc)
{ 
   if (!console)
   {
      // Can't if no console yet.
      Send("not attached]\r\n");
      return;
   }

   if (!canWrite && !console->ReadOnly())
   {
      // Can't if it's read-only. 
      Send("no -- readonly]\r\n");
      return;
   }

   if (!console->IsUp())
   {
      // Just say ok if it's already down. 
      Send("ok]\r\n");
      return;
   }

   // -----------------------------------------------------------------
   // Mark it unwritable and take the line down.
   // -----------------------------------------------------------------
   
   canWrite = 0;
   console->Close();
   Send("line down]\r\n");
   
   // -----------------------------------------------------------------
   // Tell all who closed it
   // -----------------------------------------------------------------

   SYSLOG << Severity(Syslog::error)
          << "Line "
          << console->DevName()
          << " for console "
          << console->Name()
          << " closed down by "
          << userid 
          << endl;
   
   string msg = "[line down by " + userid + "]\r\n";
   
   console->LockClientlist(MLock::Shared);

   Console::clientlist::iterator pos = console->FirstClient();
   
   for  (; pos != console->LastClient(); pos++)
   {
      if ((*pos) == this) continue;
      Send(msg);
   }

   console->UnlockClientlist(MLock::Shared);
   
   return;
}

// ===================================================================
// Display detailed console status
// ===================================================================

void Client::cmd_examine(char cc)
{
   Buffer info(1024);
   
   if (!console)
   {
      Send("not attached]\r\n");
      return;
   }
   
   Send ("Examine console]\r\n\n");

   console->ConsoleDetail(info);
   
   Send(info);
   
   return;
}

// ====================================================================
// Toggle flow control.
// ====================================================================

void Client::cmd_flow(char cc)
{
   if (!console)
   {
      Send("not attached]\r\n");
      return;
   }
   
   switch(console->FlowToggle())
   {
     case -1:
      Send("failed]\r\n");
      break;
      
     case 0:
      Send("flow OFF]\r\n");
      break;
      
     case 1:
      Send("flow ON]\r\n");
      break;
   }
   
   return;
}

// ====================================================================
// Halt character 1
// ====================================================================

void Client::cmd_halt1(char cc)
{
   if (!console)
   {
      Send("not attached]\r\n");
      return;
   }
   
   if (console->ReadOnly())
   {
      // Can't, console is read-only. 
      Send("can\'t halt read-only console]\r\n");
      return;
   }
   
   if (!canWrite)
   {
      // Can't, not the one writing to the console 
      Send("attach to halt]\r\n");
      return;
   }

   // -----------------------------------------------------------------
   // Need one final character to halt (for safety), so set the
   // fsm to enter fsm_halt1() when received.  The real work
   // of halting the console is done there.
   // -----------------------------------------------------------------
   
   state = S_HALT1;
   Send("halt ");
   return;
}

// ====================================================================
// Give heartbeat of consoles on this server.
// ====================================================================

void Client::cmd_hb(char cc)
{
   Buffer buffer;
   
   list<Console*>::iterator  pos = DISPATCHER.FirstConsole();
   list<Console*>::iterator epos = DISPATCHER.LastConsole();
   
   Send("Console heartbeat]\r\n\n");

   for  (; pos != epos ; pos ++)
   {
      (*pos)->ConsoleHB(buffer);
      Send(buffer);
      Send("\r\n");
   }

   DISPATCHER.Check();
   
   return;
}


// ====================================================================
// Print help message for client.
// ====================================================================

void Client::cmd_help(char cc)
{ 
   Send("help]\r\n\r\n");

   Send(DAEMON.Version());

   Send("\r\n\r\nEscape seq (currently ");
   Send(FmtCtl(escape));
   Send(") + char\r\n");
   
   Send(".    disconnect\r\n");
   Send("a    attach readwrite\r\n");
   Send("b    display heartbeats of consoles on this server\r\n");
   Send("c    change to new console\r\n");
   Send("d    down (close tty) a console\r\n");
   Send("e    change escape sequence\r\n");
   Send("f    force attach readwrite\r\n");
   Send("g    get location infomation\r\n");
   Send("h    print this message\r\n");
   Send("i    display info about all consoles on this server\r\n");
   Send("k    set idle timeouts for this session\r\n");
   Send("l1   send break (halt host!)\r\n");
   Send("m    execute a macro (? for list)\r\n");
   Send("o    (re)open the tty file\r\n");
   Send("p    Display and select the baud rate\r\n");
   Send("qy   shutdown the server\r\n");
   Send("r    replay the last 20 lines\r\n");
   Send("s    attach readonly\r\n");
   Send("u    show status of all consoles on this server\r\n");
   Send("v    show server version info\r\n");
   Send("w    show all users connected to this server\r\n");
   Send("x    examine -- show detailed console info\r\n");
   Send("<cr> ignore/abort command\r\n");
   Send("?    print this message\r\n");

   if (!console)
   {
      Send("\r\nYou are not currently connected to a console\r\n");
      Send("Use ");
      Send(FmtCtl(escape));
      Send("a to connect to a console\r\n");
   }

   return;
}

// ====================================================================
// Give brief information about consoles on this server.
// ====================================================================

void Client::cmd_info(char cc)
{
   Buffer buffer;
   
   list<Console*>::iterator  pos = DISPATCHER.FirstConsole();
   list<Console*>::iterator epos = DISPATCHER.LastConsole();
   
   Send("Console info]\r\n\n");

   for  (; pos != epos ; pos ++)
   {
      (*pos)->ConsoleInfo(buffer);
      Send(buffer);
      Send("\r\n");
   }

   DISPATCHER.Check();
   
   return;
}

// ====================================================================
// Locate a particular console
// ====================================================================

void Client::cmd_locate(char cc)
{
   Buffer buffer;
   BConsole *cons;
   string consname;

   consname = GetWord(true);
   Send(": ");
   
   // Try to locate the named console.
   
   cons = DISPATCHER.FindConsole(consname);

   // If not local, see if it is a known remote console.

   if (!cons)
   {
      cons = DISPATCHER.FindRConsole(consname);
   }
   
   if (!cons)
   {
      Send("unknown]\r\n");
      
      return;
   }

   Send(cons->DevName() + " on " + cons->Server() + " (" +
        cons->Mode() + ") " + cons->Location() + "]\r\n");

   return;
}

// ====================================================================
// Execute a macro sequence.
// ====================================================================

struct timespec halfsecond = {0, 500000};

void Client::cmd_macro(char cc)
{
   if (!console)
   {
      Send("Not attached]\r\n");
      return;
   }
   
   if (console->ReadOnly())
   {
      Send("can\'t write to readonly line]\r\n");
      return;
   }
   if (!canWrite)
   {
      Send("attach readwrite to send character]\r\n");
      return;
   }

   Send("macro: "); 

   while ((cc = input.GetChar()) == 0)
   {
      input.Clear();
      Receive(input);
   }

   if (cc < ' ') 
   {
      Send("Cancelled]\r\n");
      return;
   }

   if (cc == '?')
   {
      Send("List]\r\n");
      Send(DAEMON.ListMacros());
      Send('[');
      cmd_macro(cc);
      return;
   }

   macrodef macro = DAEMON.Macro(cc);
   if (macro.second.empty()) macro.second = "Undefined";

   Send(cc);
   Send(": ");
   Send(macro.second);
   Send("]\r\n");

   if (macro.first.empty()) return;  // Not defined.

   for (int i = 0; i < (int) macro.first.size(); i++)
   {
      // Process special codes.
      if (macro.first[i] == '\xff') { console->SendBreak(); continue; }
      if (macro.first[i] == '\xfe') { nanosleep(&halfsecond, NULL); continue; }
      // else fall through and send literal character
      console->Write(macro.first[i]);
   }
}


// ====================================================================
// Close and re-open the lines.
// ====================================================================

void Client::cmd_open(char cc)
{ 
   Client *writer;
   
   if (!console)
   {
      // Can't, no console
      Send("no -- no console]\r\n");
      return;
   }
   
   console->Open();

   if (!console->IsUp())
   {
      Send("line to host is still down]\r\n");
   }
   else if (console->ReadOnly())
   {
      Send("up read-only]\r\n");
   }
   else if (!(writer = console->Writer()))
   {
      canWrite = true;
      Send("up -- attached]\r\n");
   }
   else if (writer == this)
   {
      Send("up]\r\n");
   }
   else
   {
      Send("up, " + writer->userid + " is attached]\r\n");
   }

   return;
}

// ====================================================================
// User requested shutdown of server.
// ====================================================================

void Client::cmd_quit(char cc)
{ 
   state = S_QUIT;              // State to look for confirmation
   Send("shutdown consoled on ");
   Send(Host::LocalNode());
   Send(" - ");
   return;
}

// ====================================================================
// Send quoted character.
// ====================================================================

void Client::cmd_quote(char cc)
{ 
   if (!console)
   {
      Send("Not attached]\r\n");
      return;
   }
   
   if (console->ReadOnly())
   {
      Send("can\'t write to readonly line]\r\n");
      return;
   }
   if (!canWrite)
   {
      Send("attach readwrite to send character]\r\n");
      return;
   }

   temp.Clear();                // Prepare buffer.
   state = S_QUOTE;             // State to look for quote
   Send("quote \\");
   return;
}

// ====================================================================
// Reload the configuration file.
// ====================================================================
 
void Client::cmd_reload(char cc)
{
   SYSLOG << Severity(Syslog::info)
          << "Reload command sent by "
          << UserAtHost()
          << endl;
   
   DAEMON.FetchConfig();        // Get a new config
   DAEMON.ReadConfig();         // Parse it.
   DISPATCHER.StartConsoles();  // Start any new consoles.

   Send("Reloaded]\r\n");
}

// ====================================================================
// Replay the last 23 lines in the current logfile.   There may
// not be that many, and we can't really go back to the previous
// one, because there's no way to know  where it is.
// ====================================================================

void Client::cmd_replay(char cc)
{ 
   if (!console)
   {
      Send("Not attached]\r\n");
      return;
   }

   Send("replay " + console->LogName() + "]\r\n");

   // Fill a buffer from the tail of the file, with enough
   // room for at least 50 80-character lines. 
   Buffer buffer(4096);
   console->ReadLog(buffer);

   // Count back 23 newline characters from the end.
   int e = buffer.Size();
   int p = e - 1;
   int cnt = 0;

   for ( ; p >= 0; p--) if (buffer[p] == '\n' && ++cnt > 23) break;
   if (p > 0) p++;
   
   // Rebuffer the output into lines and fix newlines for raw mode.
   Buffer line;
   for ( ; p > e; p++)
   {
      if (buffer[p] == '\n') 
      { 
         line += "\r\n";
         Send(line);
         line.Clear();
         p++;
      }
      else line += buffer[p++];
   }
   return;
}

// ====================================================================
// Reply with list of known servers, and drop the connection.  This is
// usually an internal call by a client to just get the list of servers
// for the next operation.
// ====================================================================

void Client::cmd_servers(char cc)
{
   Send(DAEMON.ServerList());
   Send("]\r\n");
   state = S_CLOSE;

   return;
}


// ====================================================================
// Set the speed of the line
// ====================================================================

static Entry speedtab[] =
{
   Entry("1200",  '1'),
   Entry("1800",  '2'),
   Entry("2400",  '3'),
   Entry("4800",  '4'),
   Entry("9600",  '5'),
   Entry("19200", '6'),
   Entry("38400", '7'),
   Entry("57600", '8'),
#ifdef B115200
   Entry("115200", '9'),
#endif
};

#define speedtab_end (speedtab+sizeof(speedtab)/sizeof(Entry))

void Client::cmd_speed(char cc)
{
   char response;
   Entry *sp;

   if (!console)
   {
      Send("not attached]\r\n");
      return;
   }

   Send("Speed is " + console->GetSpeed() + "]\r\n");
   Send("[Choose: ]\r\n");

   for (sp = speedtab; sp < speedtab_end; sp++)
   {
      Send("[  ");
      Send(sp->Val1());
      Send(" = ");
      Send(sp->Name());
      Send("]\r\n");
   }
   
   Send("[Speed: ");

   while ((response = input.GetChar()) == 0)
   {
      input.Clear();
      Receive(input);
   }
   
   sp = find(speedtab, speedtab_end, response);
   
   if (sp >= speedtab_end) 
   {
      Send("Not a valid speed]\r\n");
      return;
   }
   
   console->SetSpeed(sp->Name());

   Send(sp->Name());
   Send("]\r\n");
}

// ====================================================================
// Give status of consoles on this server.
// ====================================================================

void Client::cmd_status(char cc)
{
   Buffer buffer;
   
   list<Console*>::iterator  pos = DISPATCHER.FirstConsole();
   list<Console*>::iterator epos = DISPATCHER.LastConsole();
   
   Send("Console status]\r\n\n");

   for  (; pos != epos ; pos ++)
   {
      (*pos)->ConsoleStatus(buffer);
      Send(buffer);
      Send("\r\n");
   }

   DISPATCHER.Check();
   
   return;
}

// ====================================================================
// Set timeouts for this client.
// ====================================================================

void Client::cmd_timeout(char cc)
{
   OSTRSTREAM line;
   int num;

   line << "[Timeouts currently " << idleWarn << " and " << idleForce << "]\r\n";
   Send(line.str());

   Send("[Warning timeout (minutes): ");
   num = GetNum(true);
   if (num != -1) idleWarn = num;
   else Send("unchanged");
   
   Send("]\r\n[Disconnect timeout (minutes): ");
   num = GetNum(true);
   if (num != -1) idleForce = num;
   else Send("unchanged");
   Send("]\r\n");
   
   return;
}

// ====================================================================
// Unknown command sequence.
// ====================================================================

void Client::cmd_unknown(char cc)
{ 
   Send("unknown -- use ");
   Send(FmtCtl(escape));
   Send("h]\r\n");
   return;
}  


// ====================================================================
// Report our version.
// ====================================================================

void Client::cmd_version(char cc)
{ 
   Send("Server version " + DAEMON.Version() + "]\r\n");
   return;
}


// ====================================================================
// Report who is signed on to this console.
// ====================================================================

void Client::cmd_who(char cc)
{ 
   DISPATCHER.LockClientlist(MLock::Shared);
   
   Dispatcher::clientlist::iterator pos = DISPATCHER.FirstClient();
   Dispatcher::clientlist::iterator end = DISPATCHER.LastClient();
   
   Send("Who connected]\r\n\n");
   
   for  (; pos != end; pos++)
   {
      Client *client = (*pos);
      
      if (client == this) continue;
      if (!client->Running()) continue;

      Send(client->UserAtHost());
      Send(" ");
      if (client->Cons()) 
      {         
         Send(client->console->Name());
         Send(client->canWrite ? " r/w" : " r/o");
         Send(" idle ");
         Send(client->IdleTime());
         Send("\r\n");
      }
      else Send("No console\r\n");
   }

   DISPATCHER.UnlockClientlist(MLock::Shared);
   
   return;
}

// ====================================================================
// Report who is signed on to this console.
// ====================================================================

void Client::cmd_health(char cc)
{
   OSTRSTREAM line;
   extern char BUILD[];
   
   Send("Health]\r\n\n");
   
   Send(DAEMON.Version());
   Send("\r\n");
   Send(DAEMON.StartMsg());
   line << "Listening on port " << DAEMON.Port() << "\r\n";
   Send(line.str());
   Send("\r\nLog Messages:\r\n");
   std::queue<string> history = SYSLOG.History();
   while (history.size() > 0)
   {
      Send(history.front());
      history.pop();
   }
   
   return;
}

// ====================================================================
// Report on the amount of idle time
// ====================================================================

string Client::IdleTime(void)
{
   OSTRSTREAM result;
   
   int idle = time(NULL) - lastActivity;
   
   int sec = idle % 60;  idle /= 60;
   
   if (idle == 0)  
   { 
      result << sec << " seconds" << '\0';
      return result.str();
   }
   
   int min = idle % 60; idle /= 60;
   
   if (idle == 0)
   {
      result << min << " minutes";
      if (min < 10) result << " " << sec << " seconds";
      result << '\0';
      return result.str();
   }
   
   int hr = idle % 24; idle /= 24;
   
   if (idle == 0)
   {
      result << hr << " hours " << min << " minutes" << '\0';
      return result.str();
   }
   
   result << idle << " days " << hr << " hours" << '\0';
   return result.str();
}

// ====================================================================
// Check on idle time and force if idle too long.
// ====================================================================

bool Client::IdleCheck(void)
{
   int idle = time(NULL) - lastActivity;
   int idlemin = idle / 60;

   if (!Running()) return true;
   
   if (idleForce > 0 && idlemin >= idleForce)
   {
      string msg = "\r\n[Idle time exceeded, idle for ";
      msg += IdleTime();
      msg += "]\r\n";
      try { Send(msg); } catch (...) { };
      DropConsole();
      StopReader();
      return true;
   }

   if (idleWarn > 0 && idlemin >= idleWarn)
   {
      string msg = "\r\n[Warning, idle for ";
      msg += IdleTime();
      msg += "]\r\n";
      try { Send(msg); } catch (...) { };
   }
   return false;
}


// ====================================================================
// Format a control character for printing.
// ====================================================================

string FmtCtl(string cc)
{
   string buffer;
   
   for (int i = 0; i < (int) cc.size(); i++)
   {
      unsigned char c = cc[i];
   
      c &= '\377';
   
      if (c > '\177') { c -= '\200';  buffer += "M-"; }        // Meta char
   
      if (c < ' ' || c == '\177') {buffer += '^'; buffer += (char)(c + 'A' - 1);}   // Ctl char
      else buffer += c;
   }
   
   return buffer;
}
