// ====================================================================
// client.hh - Base class for clients connections to console server.
//             A client is a Socket and is a Reader
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.10 $ $Date: 2006/12/05 06:44:28 $ $Author: boeheim $
// ******************************************************************** //

#ifndef client_hh
#define client_hh

#include <string>

#include "acl.h"
#include "autoconfig.h"
#include "sockettcp.h"
#include "reader.h"
#include "buffer.h"
#include "console.h"
#include "file.h"

#ifdef HAVE_LIBSSL
  #include "socketssl.h"
  #define SOCK_TYPE SocketSSL
#else
  #define SOCK_TYPE SocketTCP
#endif

class Console;

class Client : public SOCK_TYPE, public Reader, public CECallback
{

protected:

   Console *console;            // Console connected to.

   bool canWrite;               // Client can write to TTY.
   bool wantWrite;              // Client wants to write to TTY.

   string escape;               // Escape sequence (two chars).
   int    state;                // Finite State Machine state.
   Buffer input;                // Accumulate typed chars from socket.
   SmallBuffer temp;
   string userid;               // Identity of user on socket.
   CredList credlist;           // Credentials presented by client.
   bool   terse;                // Next command produces terse output.
   int    lastActivity;         // Time of last input from user.
   int    idleForce;            // Time until disconnect
   int    idleWarn;             // Time until warning
   
   string GetWord(bool echo=false, bool num=false); // Get word from client
   int    GetNum(bool echo=false); // Get number from client
   void   DropConsole(void);    // Drop connection to a console.

public:

   Client(void);                // Constructor.

   bool OfferWrite(void);       // Offer write ability to client.

   void ReaderProc(void);       // Override Reader() method
   
   Console *Cons(void)   { return console; };
   bool   CanWrite(void) { return canWrite; };
   string Userid(void)   { return userid; };
   string UserAtHost(void) { return string(userid + "@" + Peername()); };
   string IdleTime(void);
   bool   IdleCheck(void);
   
   void SetTimeouts(int force, int warn) 
      { idleForce = force; idleWarn = warn; };

   bool ValidateCred(Cred *cred); // Validate a credential

#ifdef HAVE_LIBPAM
   int    PamConversation(int num_msg, 
                          const struct pam_message **prompt, 
                          struct pam_response **response);
#endif   

   // -----------------------------------------------------------------
   // States for a Finite State Machine, in alphabetical order.
   // -----------------------------------------------------------------
   
   enum 
   {
      S_CATTN,       // Change first character of escape sequence
      S_CESC,        // Change second character of escape sequence
      S_CLOSE,       // Close the connection.
      S_CMD,         // Second escape character received          
      S_CONN,        // Waiting for a console name to be completed
      S_ESC1,        // First escape character received              
      S_HALT1,       // First halt sequence received.          
      S_NORMAL,      // Just pass character                          
      S_QUIT,        // Quit character received, awaiting confirmation
      S_QUOTE,       // Send any character we can spell              
      S_SUSP,        // We are suspended, first char wakes us up      
      S_USER1,       // Waiting for user lead in string char 1
      S_USER2,       // Waiting for user lead in string char 2
      S_USER3,       // Waiting for user lead in string char 3
      S_USER4,       // Waiting for username.
      FSM_SIZE       // Size of FSM table.
   };
   
private:

   // -----------------------------------------------------------------
   // Routines for a Finite State Machine, in alphabetical order,
   // corresponding to the above states. They are loaded into the
   // following table at initialization.
   // -----------------------------------------------------------------
   
   void fsm_cattn       (char cc); 
   void fsm_cesc        (char cc);  
   void fsm_cmd         (char cc);   
   void fsm_conn        (char cc);
   void fsm_esc1        (char cc);  
   void fsm_halt1       (char cc); 
   void fsm_quit        (char cc);
   void fsm_quote       (char cc);
   void fsm_susp        (char cc);  
   void fsm_user        (char cc);  

   // -----------------------------------------------------------------
   // Routines for commands that can be entered following receipt of
   // a command escape sequence, listed in alphabetic order.
   // The following table is loaded with pointers to the routines for
   // the appropriate characters.
   // -----------------------------------------------------------------
   
   void cmd_abort       (char cc);  
   void cmd_attach      (char cc);  
   void cmd_connect     (char cc);
   void cmd_consoles    (char cc);  
   void cmd_control     (char cc);    
   void cmd_descape     (char cc);
   void cmd_disconnect  (char cc); 
   void cmd_down        (char cc);    
   void cmd_examine     (char cc);    
   void cmd_flow        (char cc);    
   void cmd_halt1       (char cc);  
   void cmd_hb          (char cc);  
   void cmd_help        (char cc);       
   void cmd_info        (char cc);   
   void cmd_locate      (char cc);   
   void cmd_macro       (char cc);   
   void cmd_open        (char cc);   
   void cmd_quit        (char cc);
   void cmd_quote       (char cc);  
   void cmd_reload      (char cc);  
   void cmd_speed       (char cc);
   void cmd_servers     (char cc);
   void cmd_status      (char cc);
   void cmd_timeout     (char cc);
   void cmd_replay      (char cc);
   void cmd_spy         (char cc);    
   void cmd_unknown     (char cc); 
   void cmd_version     (char cc);
   void cmd_who         (char cc);  
   void cmd_health      (char cc);  
   
};

#endif







