// ====================================================================
// General class to parse command line options.  Accepts both single
// character and extended options, with abbreviations.
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.4 $ $Date: 2005/02/09 00:48:12 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include "cmdargs.h"

#include <vector>

// ====================================================================
// Constructor for an Option Descriptor
// ====================================================================

OptionDesc::OptionDesc(int h, string p, bool tv, int min=1) 
      : handle(h), pattern(p), takesvalue(tv), minlength(min)
{
   if (p[0] != '-') pattern = '-' + p;
   
   if (p.size() == 2) single = true;
   else single = false;
}

// ====================================================================
// Determine if an option matches a token.
// ====================================================================

bool OptionDesc::Matches(string token)
{
   if (SingleChar())
   {
      if (token.substr(0,2) == pattern) return true;
      else return false;
   }
   
   if ((int) token.size() < minlength) return false;
   
   int len = (minlength < (int) token.size()) ? token.size() : minlength;
   
   if (pattern.substr(0,len) == token) return true;
   else return false;
}

// ====================================================================
// Destructor method
// ====================================================================

CmdArgs::~CmdArgs(void)
{
   while (desclist.size() > 0)
   {
      OptionDesc *op = desclist.back();
      delete op;
      desclist.pop_back();
   }
}

// ====================================================================
// Add an option to the list to be parsed.
// ====================================================================

void CmdArgs::AddOption(int h, string p, bool tv, int min)
{
   OptionDesc *op = new OptionDesc(h, p, tv, min);
   
   desclist.push_back(op);
}

// ====================================================================
// Parse an argument list with an option description list.
// ====================================================================

void CmdArgs::ParseArgs(int argc, char *argv[]) throw(CmdErr)
{
   string token;                // Copy of current token in argv
   argpair curset;              // Current option, value set
   
   if (! argc) return;          // Nothing to do.

   // The first token is the program name, always stuff that
   // into argument 0.
   
   argc--;
   arglist.push_back(*argv++);

   // Process all tokens that begin with hyphen as options.
   
   while (argc || ! token.empty())
   {
      // Get a new token from the input if necessary

      if (token.empty()) 
      { 
         if (**argv != '-') break; // end of options
         token = *argv++; 
         argc--; 
      }
      
      // Find the option in the list of patterns.

      vector<OptionDesc*>::iterator pos = desclist.begin();
      vector<OptionDesc*>::iterator epos = desclist.end();
      
      for ( ; pos < epos; pos++)
      {
         OptionDesc *op = *pos;
         
         if (op->Matches(token))
         {
            curset.first = op->Handle();
            curset.second = "";
            
            if (op->TakesValue())
            {
               if (op->SingleChar() && token.size() > 2)
                  curset.second = token.substr(2);
               else
               {
                  if (!argc-- || **argv == '-')
                     throw CmdErr("No value for option", token);
                  curset.second = *argv++;
               }
               token = "";
            }
            else 
            {
               if (op->SingleChar() && token.size() > 2) token.erase(1,1);
               else token = "";
            }
            optlist.push_back(curset); 
            break;
         }
      }
      
      if (pos == epos) throw CmdErr("Unrecognized option", token);
   }

   // What remains are command line arguments.

   if (! token.empty())
   {
      throw CmdErr("Unrecognized:", token);
   }

   while (argc--)
   {
      arglist.push_back(*argv++);
   }
   
}


      
