// ====================================================================
// console.cc - Methods to manipulate a console.
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.24 $ $Date: 2006/12/05 06:44:28 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include <algorithm>
#include <errno.h>
#include <iomanip>
#include <iostream>
#include <termios.h>
#include <unistd.h>

#include "buffer.h"
#include "console.h"
#include "consoled.h"
#include "global.h"
#include "sensorsoft.h"
#include "syslog.h"
#include "utility.h"

// ====================================================================
// Constructors for Line objects.
// ====================================================================

Line::Line(void)
{
   server = "localhost";
   speed  = SPEED;
   tsformat  = "%T ";
}

Line::Line(Line &rhs)
{
   name     = rhs.name;
   devname  = rhs.devname;
   server   = rhs.server;
   speed    = rhs.speed;
   location = rhs.location;
   classify = rhs.classify;
   filter   = rhs.filter;
   stream   = rhs.stream;
   tsformat = rhs.tsformat;
   acl      = rhs.acl;
}

Line::Line(Line &rhs, string newname, string newdev)
{
   name     = newname;
   devname  = newdev;
   server   = rhs.server;
   speed    = rhs.speed;
   location = rhs.location;
   classify = rhs.classify;
   filter   = rhs.filter;
   stream   = rhs.stream;
   tsformat = rhs.tsformat;
   acl      = rhs.acl;
}

// ====================================================================
// Assignment operators.
// ====================================================================

Line &Line::operator=(Line *rhs)
{
   *this = *rhs;
   return *this;
}

Line &Line::operator=(Line &rhs)
{
   name     = rhs.name;
   devname  = rhs.devname;
   server   = rhs.server;
   speed    = rhs.speed;
   location = rhs.location;
   classify = rhs.classify;
   filter   = rhs.filter;
   stream   = rhs.stream;
   tsformat = rhs.tsformat;
   acl      = rhs.acl;
   return *this;
}

// ====================================================================
// Parse a line or default statement from the configuration file
// and populate the Line object with the values.
//
// line <name> <ttydev> [<server>] [speed=<nnn>] 
//             [location=<string>] [class=<string>] [filter=<name>]
//             [stream=<stream>]   [timestamp=format] [acl=name]
//
// name     - Name that line is known by.  Usually name of machine on line.
// ttydev   - tty serving line, e.g. /dev/tty0
// server   - Name of server handling line, default localhost.
// speed    - Baud rate and parity setting, e.g. 9600p.
// location - Comment identifying rack location of server
// class    - Class for GUI display and other groupings.
// filter   - Filter for processing input.
// stamp    - strftime(3) style format for log; default "%T "
// acl      - Access control list
// ====================================================================

   static char *keywords[] = { "server", "speed", "class", "location", "filter", "stream", "stamp", "acl", 0 };

   static bool isKeyword(string &test)
   {
      for (char **p = keywords; *p; p++) if (test == *p) return true;
      return false;
   }

void Line::Parse(vector<string> &tokens) throw(string)
{
   int curtoken = 0;

   // Get line name and tty name.  Not allowed in a default statement,
   // required in a line or linegroup statement.

   if (tokens[curtoken++] != "default")
   {
      if (curtoken >= (int) tokens.size() || isKeyword(tokens[curtoken]))
         throw string("Missing line name");

      name = Lower(tokens[curtoken++]);  // case insignificant.

      if (curtoken >= (int) tokens.size() || isKeyword(tokens[curtoken]))
         throw string("Missing tty device");

      devname = tokens[curtoken++];      // case significant.
   }
      
   // Get server name field if present.

   if (curtoken >= (int) tokens.size()) return;
   Lower(tokens[curtoken]);

   if (! isKeyword(tokens[curtoken]))
      server = tokens[curtoken++];

   // Remaining fields may be omitted or in any order.

   while (curtoken < (int) tokens.size())
   {
      Lower(tokens[curtoken]);
      
      if (! isKeyword(tokens[curtoken]))
          throw string("Unrecognized keyword ") + tokens[curtoken];

      if (curtoken+1 >= (int) tokens.size() || isKeyword(tokens[curtoken+1])) 
         throw string("Omitted value for ") + tokens[curtoken];

      // server=nnnn field

      if (tokens[curtoken] == "server")
      {
         curtoken++;
         server = tokens[curtoken++];
      }

      // speed=nnnn field

      else if (tokens[curtoken] == "speed")
      {
         curtoken++;
         speed = tokens[curtoken++];
      }

      // location=string field

      else if (tokens[curtoken] == "location")
      {
         curtoken++;
         location = tokens[curtoken++];
      }

      // class=string field

      else if (tokens[curtoken] == "class")
      {
         curtoken++;
         classify = tokens[curtoken++];
      }

      // filter=name field

      else if (tokens[curtoken] == "filter")
      {
         curtoken++;
         filter = tokens[curtoken++];
      }

      // stream=name field

      else if (tokens[curtoken] == "stream")
      {
         curtoken++;
         stream = tokens[curtoken++];
      }

      // timestamp=format field

      else if (tokens[curtoken] == "stamp")
      {
         curtoken++;
         tsformat = tokens[curtoken++];
      }

      // acl=name

      else if (tokens[curtoken] == "acl")
      {
         curtoken++;
         acl = DAEMON.FindACL(tokens[curtoken++]);
         if (!acl) throw string("Unknown ACL name");
      }
   }
}

// ====================================================================
// Return equality of two Line objects.
// ====================================================================

bool Line::operator==(Line &comp)
{
   if (name     != comp.name    ) return false;
   if (devname  != comp.devname ) return false;
   if (server   != comp.server  ) return false;
   if (speed    != comp.speed   ) return false;
   if (location != comp.location) return false;
   if (classify != comp.classify) return false;
   return true;
}
   
// ====================================================================
// Constructor for a basic console object.
// ====================================================================

BConsole::BConsole(Line *lp) : Line(*lp)
{

}

// ====================================================================
// Destroy a basic console object
// ====================================================================

BConsole::~BConsole(void)
{

}

// ====================================================================
// Copy constructor
// ====================================================================

BConsole &BConsole::operator=(Line *lp)
{
   *this = *lp;
   return *this;
}

BConsole &BConsole::operator=(Line &lp)
{
   if (lp == *this) return *this;     // Just return if no changes.

   *this = lp;
   return *this;
}

// ====================================================================
// Constructor for a CONSOLE object.
// ====================================================================

Console::Console(Line *lp) 
      : BConsole(lp)
{
   // -----------------------------------------------------------------
   // Initialize the logfile and message service.
   // -----------------------------------------------------------------
   
   log.Dir(DAEMON.LogDir());
   log.Name(Name());

   logbuffer.Trigger(*this, '\n', 'l');
   msgbuffer.Trigger(*this, '\n', 'm');
   logbuffer.FilterMode(Buffer::Parity);
   logbuffer.FilterMode(Buffer::Chars, '\r');
   logbuffer.FilterMode(Buffer::Ansi);
   msgbuffer.FilterMode(Buffer::Parity);
   msgbuffer.FilterMode(Buffer::Chars, '\r');
   msgbuffer.FilterMode(Buffer::Ansi);
   
   logbuffer.TimeStampFormat(tsformat);
   msgbuffer.TimeStampFormat(tsformat + " " + Name() + ": ");

   state      = S_NORMAL;
   filterproc = NULL;
   isUp       = false;
   heartbeat  = NULL;
   
   if (FilterName() == "sensorsoft")
   {
      filterproc = new SensorSoft(this);
   }
   else if (! FilterName().empty())
   {
      SYSLOG << Severity(Syslog::warning)
             << "Invalid filter name "
             << FilterName()
             << endl;
   }
   
   // -----------------------------------------------------------------
   // Produce an informational log entry about this creation.
   // -----------------------------------------------------------------
   
   if (GLOBAL.Verbose(2))
   {
      Buffer info;
   
      ConsoleInfo(info);

      SYSLOG << Severity(Syslog::debug)
             << info.Cstr()
             << endl;
   }   
}

// ====================================================================
// Destroy a console object
// ====================================================================

Console::~Console(void)
{
   if (filterproc) delete filterproc;
   if (heartbeat)  delete heartbeat;

   try
   {
      CloseLog();
   }
   catch (IOErr e)
   {
      // Ignore
   }
}

// ====================================================================
// Update line information in a console structure.
// ====================================================================

Console &Console::operator=(Line *lp)
{
   *this = *lp;
   return *this;
}

Console &Console::operator=(Line &lp)
{
   if (lp == *this) return *this;     // Just return if no changes.

   *(Line *)this = lp;
   
   // -----------------------------------------------------------------
   // Update the logfile name and the timestamp in case the name changed.
   // -----------------------------------------------------------------
   
   log.Name(Name());
   msgbuffer.TimeStampFormat(tsformat + " " + Name() + ": ");
   
   // -----------------------------------------------------------------
   // Produce an informational log entry about this update.
   // -----------------------------------------------------------------
   
   if (GLOBAL.Verbose(2))
   {
      Buffer info;
   
      ConsoleInfo(info);

      SYSLOG << Severity(Syslog::debug)
             << info.Cstr()
             << endl;
   }   

   return *this;
}

// ====================================================================
// Method called when the log buffer needs to be written out.
// ====================================================================

void Console::WriteTriggered(Buffer &record, char tag)
{
   record.TimeStamp();
   
   if (tag == 'l')   log.WriteLog(record);
   else              DAEMON.SendMessage(stream, record);
}

// ====================================================================
// Add a client to a console, with locking of list.
// ====================================================================

void Console::AddClient(Client *NewClient)
{
   LockClientlist(MLock::Exclusive);

   clients.push_back(NewClient);

   UnlockClientlist(MLock::Exclusive);
}

// ====================================================================
// Drop a client from a console, with locking of list.
// ====================================================================

void Console::DropClient(Client *oldClient)
{
   LockClientlist(MLock::Exclusive);

   clientlist::iterator pos = clients.begin();
   
   pos = find(clients.begin(), clients.end(), oldClient);

   if (pos != clients.end()) clients.erase(pos);

   UnlockClientlist(MLock::Exclusive);
}

// ====================================================================
// Find the client that currently can write to the console.
// ====================================================================

Client *Console::Writer(void)
{
   Client *found = NULL;
   
   LockClientlist(MLock::Shared);
   
   clientlist::iterator pos = clients.begin();
   
   for ( ; pos != clients.end(); pos++)
   {
      if ((*pos)->CanWrite()) { found = *pos; break; };
   }

   UnlockClientlist(MLock::Shared);
   
   return found;
}


// ====================================================================
// Offer write access to the clients in the list until one accepts.
// ====================================================================

void Console::OfferWrite(void)
{
   LockClientlist(MLock::Shared);
   clientlist::iterator pos = clients.begin();
   
   for ( ; pos != clients.end(); pos++)
   {
      if ((*pos)->OfferWrite()) break; 
   }

   UnlockClientlist(MLock::Shared);
   
   return;
}


// ====================================================================
// (Re)Initialize the console.  Called on initial setup or to
// reinitialize after errors.
// ==================================================================== 

void Console::Open(void)
{
   // -----------------------------------------------------------------
   // Close previous TTY.
   // ----------------------------------------------------------------- 
   
   if (isUp) Close();
   
   isUp = false;
   
   // -----------------------------------------------------------------
   // Try to open it again
   // -----------------------------------------------------------------

   if (GLOBAL.Debug() && GLOBAL.Verbose(2))
   {
      SYSLOG << Severity(Syslog::debug)
             << "Opening "
             << DevName()
             << endl;
   }

   try
   {
      TTYFile::Open(DevName(), O_RDWR | O_NDELAY, 0600);
      Blocking();

      SetMode(Mode());
      SetRaw();
   }
   catch (IOErr e)
   {
      SYSLOG << Severity(Syslog::error) 
             << e.routine 
             << ": " << e.filename 
             << ": " << e.message 
             << endl;
      return;
   }

   if (filterproc) filterproc->Open();

   isUp = true;   
}


// ====================================================================
// Close a console.
// ==================================================================== 

void Console::Close(void)
{
   if (isUp)
   {
      try
      {
         TTYFile::Close();      // Call parent's method.
      }
      catch (IOErr e)
      {
         // Ignore
      }
   }

   if (filterproc) filterproc->Close();
   
   isUp = false;
}


// ====================================================================
// Loop reading tty line and dispatching output to all clients.
// ====================================================================

void Console::ReaderProc(void)
{
   Buffer input(256);
   Buffer output(256);
   
   try {                        // Catch all exceptions

   // ----------------------------------------------------------------
   // Loop until shut down or error on console.
   // ----------------------------------------------------------------
   
   for (;;)
   {
      // --------------------------------------------------------------
      // Read terminal line
      // --------------------------------------------------------------

      input.Clear();
      output.Clear();
      
      try
      {
         Read(input);
      }
      catch (IOErr e)
      {
	 if (e.reason != EAGAIN)
         {
            /* carrier lost */

            SYSLOG << Severity(Syslog::error) 
                   << e.routine 
                   << ": " << e.filename 
                   << ": " << e.message 
                   <<  endl;
            
            Close();
            Open();
           
	    if (!isUp) return;
         }
	 
         continue;
      }

      if (input.Size() == 0) // End of file
      {
         SYSLOG << Severity(Syslog::error)
                << "readerproc(): EOF on "
                << Name()
                << ", halting reader"
                << endl;
         Close();
         return;
      }
      
      if (filterproc) filterproc->Input(input);

      // --------------------------------------------------------------
      // Process received characters through state machine.
      // --------------------------------------------------------------

      while (unsigned char cc = input.GetChar())
      {
         switch(state)
         {
             case S_NORMAL:
                switch(cc)
                {
                    case 0x1b:
                       state = S_ESCAPE;
                       break;
                       
                    case 0xff:
                       state = S_PARITY;
                       break;
                       
                    default:
                       output += cc;
                       break;
                }
                break;

             case S_PARITY:
                if (cc == 0x00)
                {
                   if (Autobaud())
                   {
                      string msg = "[Parity Err -- trying " + NextSpeed() + " baud]";
                      output += msg;
                   }
                   else
                   {
                      output += "[Parity Err]";
                   }
                }
                else
                {
                   output += (unsigned char) 0x00;
                   output += cc;
                }
                state = S_NORMAL;
                break;
                
             case S_ESCAPE:
                switch(cc)
                {
                    // Device control sequence lead-in (used for heartbeat)
                    case 'P':
                       ctrl.Clear();
                       state = S_CTRL1;
                       break;
                       
                    default:
                       output += (unsigned char) '\x1b';
                       output += cc;
                       state = S_NORMAL;
                       break;
                }
                break;
                
             case S_CTRL1:
                switch(cc)
                {
                    case 0x1b:
                       // Device control termination is 'ESC \' 
                       state = S_CTRL2;
                       break;

                    case '\\':
                       // Early beat code sent \ e as a termination   
                       state = S_CTRL3;
                       break;
                       
                    default:
                       // Don't let a message get over 100 bytes.
                       if (ctrl.Size() > 100) 
                       {
                          ctrl.Clear();
                          state = S_NORMAL;
                       }
                       else ctrl += cc;
                       break;
                }
                break;
                
             case S_CTRL2:
             case S_CTRL3:
                if (state == S_CTRL2 && cc != '\\') 
                {
                   ctrl += 0x1b;
                   ctrl += cc;
                   state = S_CTRL1;
                   break;
                }

                if (state == S_CTRL3 && cc != 'e') 
                {
                   ctrl += '\\';
                   ctrl += cc;
                   state = S_CTRL1;
                   break;
                }

                if (!heartbeat) heartbeat = new HeartBeat(this);
                if (ctrl[0] == 'H' && ctrl[1] == 'B') heartbeat->Input(ctrl);
                ctrl.Clear();
                state = S_NORMAL;
                break;
         }
      }

      if (output.Size() == 0) continue;
      
      // --------------------------------------------------------------
      // Write data to logfile and to any waiting clients.
      // Also write to the message service if no one connected
      // and no filterproc.
      // --------------------------------------------------------------

      LockClientlist(MLock::Shared);
      
      if (clients.empty() && !filterproc) msgbuffer += output;
      else for (clientlist::iterator client = clients.begin(); 
                client != clients.end(); 
                client++)
      {
         try
         {
            (*client)->Send(output);
         }
         catch (IOErr e)
         {
            SYSLOG << Severity(Syslog::error) 
                   << "Error sending to client "
                   << e.routine 
                   << ": " << e.filename 
                   << ": " << e.message 
                   <<  endl;
         }
      }

      UnlockClientlist(MLock::Shared);

      logbuffer += output;
   }

   } // end of try

   catch (File::IOErr(e))
   {
      SYSLOG << Severity(Syslog::error)
             << "Console thread ending because of uncaught I/O error "
             << e.routine 
             << ": " << e.filename 
             << ": " << e.message 
             << endl;
   }
   catch(string s)
   {
      SYSLOG << Severity(Syslog::error)
             << "Console thread ending because of uncaught exception "
             << s
             <<endl;
   }
   catch(...)
   {
      SYSLOG << Severity(Syslog::error)
             << "Console thread ending because of uncaught unknown exception"
             << endl;
   }
}

// ====================================================================
// Perform idle checks on console.
// ====================================================================

void Console::IdleCheck(time_t now)
{
   if (filterproc) filterproc->IdleCheck(now);
   if (heartbeat)  heartbeat->IdleCheck(now);
}

// ====================================================================
// Return information about the configuration of the console.
// ====================================================================

static char *upstr[] = {" down", " up  "};
static char *rwstr[] = {" rw", " ro"};

void Console::ConsoleInfo(Buffer &buffer)
{
   buffer.Clear();
   
   buffer += Name();
   buffer += " ";
   buffer += DevName();
   buffer += " ";
   buffer += log.Fullname();
   buffer += " \"";   buffer += Location();  buffer += '"';
   buffer += " \"";   buffer += Class();     buffer += "\" ";
   if (heartbeat) heartbeat->Display(buffer);
}

// ====================================================================
// Return just the heartbeat information
// ====================================================================

void Console::ConsoleHB(Buffer &buffer)
{
   buffer.Clear();
   
   buffer += Name();
   buffer += " ";
   buffer += DevName();
   buffer += " ";
   if (heartbeat) heartbeat->Display(buffer);
}

// ====================================================================
// Return information about the current status of the console.
// ====================================================================

void Console::ConsoleStatus(Buffer &buffer)
{
   buffer.Clear();
   
   buffer.Append(Name(), 16);
   buffer += upstr[isUp];
   buffer += rwstr[ReadOnly()];
   buffer += ' ';
   buffer.Append(Mode(),6);

   if (isUp)
   {
      unsigned int stat;
      vector<bool> sigState;
      
      sigState = GetSignals();
      
      buffer += ' ';
      if (sigState[SigRTS]) buffer += "+RTS";
      if (sigState[SigCTS]) buffer += "+CTS";
      if (sigState[SigDSR]) buffer += "+DSR";
      if (sigState[SigCAR]) buffer += "+DCD";
      if (sigState[SigDTR]) buffer += "+DTR";
      if (sigState[SigRI ]) buffer += "+RI";
   }
   
   if (heartbeat)
   {
      buffer += "\r\n  ";
      heartbeat->Display(buffer);
   }
   LockClientlist(MLock::Shared);
   
   clientlist::iterator client     = clients.begin();
   clientlist::iterator lastclient = clients.end();
   
   for ( ; client != lastclient; client++) 
   {
      buffer += " ";
      buffer += (*client)->UserAtHost();
      buffer += " idle ";
      buffer += (*client)->IdleTime();
   }

   UnlockClientlist(MLock::Shared);
}

// ====================================================================
// Return detailed information about the current status of the console.
// ====================================================================

void Console::ConsoleDetail(Buffer &buffer)
{
   buffer.Clear();
   
   buffer += Name();
   buffer += " on ";
   buffer += Server();
   buffer += "\r\n  Device file: ";
   buffer += DevName();
   buffer += " (";
   buffer += upstr[isUp];
   buffer += rwstr[ReadOnly()];
   buffer += ' ';
   if (isUp) buffer += GetSpeed();
   buffer += ")\r\n  Thread ID: ";
   buffer += (int) Thread();
   buffer += "\r\n  Log file: ";
   buffer += log.Fullname();
   buffer += "\r\n  Location: ";
   buffer += Location();
   buffer += "\r\n  Class: ";
   buffer += Class();
   if (heartbeat)
   {
      buffer += "\r\n  ";
      heartbeat->Display(buffer);
   }
   
   if (isUp)
   {
      buffer += "\r\n\nSignal  Pin  Pin  Pin  Direction  Status  Full\r\n";
      buffer +=       "Name    DB25 DB9  RJ45 (server)           Name\r\n";
      buffer +=       "------  ---- ---- ---- ---------  ------  ---------\r\n";
                      
      unsigned int stat = 0;
      vector<bool> sigState;
      
      sigState = GetSignals();
      
      buffer += "FG        1    -    -     -        -     Frame Ground\r\n";
      buffer += "TxD       2    3    4     out      -     Transmit Data\r\n";
      buffer += "RxD       3    2    5     in       -     Receive Data\r\n";
      buffer += "RTS       4    7    2     out      ";
      buffer += (sigState[SigRTS] ? '1' : '0');
      buffer += "     Request to Send\r\n";
      buffer += "CTS       5    8    7     in       ";
      buffer += (sigState[SigCTS] ? '1' : '0');
      buffer += "     Clear to Send\r\n";
      buffer += "DSR       6    6    1Alt  in       ";
      buffer += (sigState[SigDSR] ? '1' : '0');
      buffer += "     Data Set Ready\r\n";
      buffer += "SIG       7    5    6     -        -     Signal Ground\r\n";
      buffer += "DCD       8    1    1     in       ";
      buffer += (sigState[SigCAR] ? '1' : '0');
      buffer += "     Data Carrier Detect\r\n";
      buffer += "DTR      20    4    8     out      ";
      buffer += (sigState[SigDTR] ? '1' : '0');
      buffer += "     Data Terminal Ready\r\n";
      buffer += "RI       22    9    -     in       ";
      buffer += (sigState[SigRI ] ? '1' : '0');
      buffer += "     Ring Indicator\r\n";
   }

   LockClientlist(MLock::Shared);
   
   clientlist::iterator client     = clients.begin();
   clientlist::iterator lastclient = clients.end();
   
   buffer += "\r\nUser List:\r\n";
   
   for ( ; client != lastclient; client++) 
   {
      buffer += "  ";
      buffer += (*client)->UserAtHost();
      buffer += (*client)->CanWrite() ? " r/w" : " r/o";
      buffer += " idle ";
      buffer += (*client)->IdleTime();
      buffer += "\r\n";
   }

   UnlockClientlist(MLock::Shared);
}

// ====================================================================
// Save the state in a form suitable for persistant storage.
// Return true if state has changed since last Get or Set.
// ====================================================================

bool Console::GetState(Buffer &statebuf)
{
   bool changed = false;
   statebuf.Clear();
   
   if (heartbeat) changed = heartbeat->GetState(statebuf);
   
   return(changed);
}

// ====================================================================
// Set the state from a previously set record.
// ====================================================================

void Console::SetState(Buffer &statebuf)
{
   if (!heartbeat) heartbeat = new HeartBeat(this);
   
   heartbeat->SetState(statebuf);
}

// ====================================================================
// Determine if requestor is authorized
// ====================================================================

bool Console::Authorized(CredList *creds)
{
   if (!acl) return true;       // Default allow; may want default deny
   
   return acl->Allow(creds);
}
