// ====================================================================
// console.hh - Class to manipulate a console.
// A console is a tty line to a remote machine, and has state
// attributes and tracks a list of clients of that tty line.
// ==================================================================== 

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.8 $ $Date: 2006/12/05 06:44:28 $ $Author: boeheim $
// ******************************************************************** //

#ifndef console_hh
#define console_hh

#include <list>
#include <pthread.h>
#include <string>
#include <time.h>

#include "acl.h"
#include "client.h"
#include "defaults.h"
#include "filter.h"
#include "heartbeat.h"
#include "logfile.h"
#include "msgclient.h"
#include "mlock.h"
#include "slock.h"
#include "ttyfile.h"
#include "user.h"

class Client;
class Filter;

// --------------------------------------------------------------------
// The Line class holds the descriptive information about a console 
// line that comes from the line statement in the configuration file.
// The actual console will have a pointer to this object.
// --------------------------------------------------------------------

class Line
{
protected:
   string name;                 // Name of console on this line.
   string devname;              // Name of /dev/tty file to open
   string server;               // Server that services it.
   string speed;                // Speed setting for this line.
   string location;             // Location on shelf.
   string classify;             // Class for displays.
   string filter;               // Filter name.
   string stream;               // Name of message stream.
   string tsformat;             // Format for timestamp.
   ACL    *acl;                 // Access list for this line.

public:
   // Constructor

   Line(void);                  // Blank line.
   Line(Line &rhs);             // Copy defaults from another line.
	Line(Line &rhs, string newname, string newdev); // Copy defaults, override name.
   Line &operator=(Line *rhs);  // Assignment operators
   Line &operator=(Line &rhs);  // Assignment operators

   void Parse(vector<string>&) throw(string);

   bool operator==(Line &comp);

   // Return information from the line to non-friends

   string Name      (void) { return name; };
   string DevName   (void) { return devname; };
   string Server    (void) { return server; };
   string Speed     (void) { return speed; };
   string Mode      (void) { return speed; };
   string Location  (void) { return location; };
   string Classify  (void) { return classify; };
   string Class     (void) { return classify; };
   string FilterName(void) { return filter; };
   string StreamName(void) { return stream; }
};

// --------------------------------------------------------------------
// Base class for a console for the behaviors of both local and
// remote consoles.
// --------------------------------------------------------------------

class BConsole : public Line
{
public:

   BConsole(Line *lp);          // Constructor
   ~BConsole(void);             // Destructor
   BConsole &operator=(Line &lp); // Update line info.
   BConsole &operator=(Line *lp); // Update line info.
};
   
// --------------------------------------------------------------------
// Main class for a local console.  It is a subclass of ttyfile so
// it inherits all those methods for handling the tty, and it is
// a subclass of reader so that it can have a reader thread.
// --------------------------------------------------------------------

class Console : public BConsole, 
                public TTYFile, 
                public Reader, 
                public TriggerCallback
{
public: // typedefs
   typedef list<Client *> clientlist;

private:
   
   Logfile log;                 // The log file for this console.
   Buffer logbuffer;            // The time-stamped log buffer connected to logfile.
   SLock  logopen;              // True if log is open, lock when changing.

   Buffer msgbuffer;            // The time-stamped message buffer.
   
   bool isUp;                   // State of console connection.
   
   clientlist clients;          // clients on this console
   MLock clientLock;            // Lock clientList against simultaneous upd.
   
   Filter *filterproc;          // Filter object if needed.

   SmallBuffer ctrl;            // Control message.
   enum {S_NORMAL, S_PARITY, S_ESCAPE, S_CTRL1, S_CTRL2, S_CTRL3} 
       state;                   // Processing state   
   HeartBeat *heartbeat;        // Keep track of heartbeat from host.

public:

   Console(Line *lp);           // Constructor
   ~Console(void);              // Destructor
   Console &operator=(Line &lp); // Update line info.
   Console &operator=(Line *lp); // Update line info.

   // Client list functions.

   void LockClientlist(MLock::locktype how) { clientLock.Lock(how); }; 
   void UnlockClientlist(MLock::locktype how) { clientLock.Unlock(how); };
   
   void AddClient(Client *);    // Add and remove clients.
   void DropClient(Client *);   // 

   clientlist::iterator FirstClient(void) { return clients.begin(); };
   clientlist::iterator LastClient(void)  { return clients.end();   };

   void Open(void);             // Open files and other setup
   void Close(void);            // Close everything.

   void OpenLog(void)            { log.OpenLog(); };
   void CloseLog(void)           { log.CloseLog(); };
   void WriteLog(Buffer &record) { log.WriteLog(record); };
   void ReadLog(Buffer &buffer)  { log.ReadLog(buffer, Logfile::Tail); };
   string LogName(void)          { return log.Fullname(); };
   string TSFormat(void)         { return tsformat; };
   
   void ReaderProc(void);           // Function to read console

   Client *Writer(void);        // Return client that can write.
   void OfferWrite(void);       // Offer write to clients on list.

   void WriteTriggered(Buffer &record, char tag); 
                                // Method called when log buffer needs 
                                // to be written.

   void IdleCheck(time_t now);     // Called by dispatcher when idle.
   void ConsoleInfo(Buffer &);     // Return line of info about console.
   void ConsoleHB(Buffer &);       // Return heartbeat status.
   void ConsoleStatus(Buffer &);   // Return console status.
   void ConsoleDetail(Buffer &);   // Return detailed console status.

   bool IsUp(void) { return isUp; };

   bool GetState(Buffer &);     // Get current state, return whether state changed.
   void SetState(Buffer &);     // Set state from saved state record.

   bool Authorized(CredList *);
};

#endif
