// ====================================================================
// consolec.cc -- Main routine for the console client.  When compiled
//                and linked with the other routines, the binary will
//                be called simply 'console'
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.22 $ $Date: 2007/04/16 01:59:05 $ $Author: boeheim $
// ******************************************************************** //

// Connect to one or more console servers to connect to a serial console
// line, or to query/control the console server.  One or more master
// servers may be contacted to find the actual server(s) to which the
// connection should be addressed.  The order of preference for
// determining the master servers is:
// 1. The -M or -host option on the command line.
// 2. The 'master' statement in the console.cf file (can be a list).
// 3. The compiled-in default 'console@localhost'.
// The preference order for determining the port to use for the
// connection:
// 1. A specific port number associated with a server: conserv7:6767
// 2. The -P or -port option on the command line
// 3. The 'port' statement in the console.cf file
// 4. The port number in /etc/services for service name 'console'
// 5. The compiled in default.
// The connection sequence varies depending on the operation:
// CONNECT: connect to master servers in the master list.  Take the
// first one to respond either that it owns the serial line, or with
// the name of the server that does own it.  Connect to the final
// server.
// SHUTDOWN: A specific server must be named; no redirection allowed.
// LOCATE: connect to master servers; report the first response.
// QUERY: If no server name is specified, or * is specified, send the
// query to all known servers.  If a server name is specified, query
// only that server.

using namespace std;

#include "autoconfig.h"
#include "buffer.h"
#include "cmdargs.h"
#include "consolec.h"
#include "defaults.h"
#include "file.h"
#include "global.h"
#include "host.h"
#include "server.h"
#ifdef HAVE_LIBSSL
#include "socketssl.h"
#endif
#include "syslog.h"
#include "terminal.h"
#include "utility.h"

#include <ctype.h>
#include <errno.h>
#include <fstream>
#include <iostream>
#include <netdb.h>
#include <signal.h>
#include <stdlib.h>
#include <string>
#include <unistd.h>

// Storage space for globals

Global        GLOBAL;

ConsoleClient CC;

Syslog        SYSLOG(CLIENT_PROGNAME, Syslog::LOGCLASS, Syslog::info);

// Constants

string CONNECT    = string(ESCAPE) + "a";
string GETSERVERS = string(ESCAPE) + "%";
string RELOAD     = string(ESCAPE) + "@";
string SHOWHB     = string(ESCAPE) + "b";
string SHOWHEALTH = string(ESCAPE) + "z";
string SHOWINFO   = string(ESCAPE) + "i";
string SHOWLOC    = string(ESCAPE) + "g";
string SHOWSTATUS = string(ESCAPE) + "u";
string SHOWUSERS  = string(ESCAPE) + "w";
string SHUTDOWN   = string(ESCAPE) + "qy";
string SIGNOFF    = string(ESCAPE) + ".";

// ====================================================================
// A function object with an argument and generic Execute function.
// ====================================================================

class funcobj
{
  protected:
   string     target;
   string     cmd;
  public:
   funcobj(string t, string c) : target(t), cmd(c) { };
   virtual string Execute(void) = 0;
};

// A Connection function

class Connection : public funcobj
{
  public:
   Connection(string t, string c="a") : funcobj(t,c) { };
   string Execute(void) { return CC.Connect(target); };
};

// A Command function

class Command : public funcobj
{
  public:
   Command(string t, string c) : funcobj(t,c) { };
   string Execute(void) { return CC.Distribute(target, cmd); };
};

// ====================================================================
// Main routine: Launch the object that does the work.
// ====================================================================

int main(int argc, char *argv[])
{
   CC.ParseArgs(argc, argv);
   CC.ReadConfigs();
   CC.ProcessArgs();
   CC.Run();
   
   exit(0);
}

// ====================================================================
// ConsoleClient constructor
// ====================================================================

ConsoleClient::ConsoleClient(void)
{
   extern char BUILD[];

   GLOBAL.Progname(CLIENT_PROGNAME);
	
   version = VERSION;
   version += " ";
   version += BUILD;
#ifdef HAVE_LIBSSL
   version += " (with SSL";
#else
   version += " (without SSL";
#endif
#ifdef HAVE_LIBGSSAPI
   version += ", with GSSAPI";
#else
   version += ", without GSSAPI";
#endif
#ifdef HAVE_LIBPAM
   version += ", with PAM)";
#else
   version += ", without PAM)";
#endif

   // Indicate default values for option fields.
	
   char *env = getenv("HOME");
   
   configs.push_back("/usr/local/etc/console.cf");
   configs.push_back("/etc/console.cf");

   string myconfig;
   if (env) { myconfig = env; myconfig += "/"; }
   myconfig  += ".console/console.cf";
   configs.push_back(myconfig);
   
   // Set the pointers in the Server and Terminal classes
   // so that they can communicate with each other. 
   
   terminal.RegisterServer(&server);
   server.RegisterTerminal(&terminal);

   // By default, output is to tty device.
   
   StdOut = &terminal;

   username = FindUsername();

   // Push a default host with no port on the master list.
   masterlist = MakeHostlist(CONSOLE_SERVER);

   // Set the default port, but don't put it on the master list.
   // Late binding makes the defaults work correctly.
   Host s;
   port = CONSOLE_PORT;
   if (s.FindPort(CONSOLE_SERVICE, "tcp")) port = s.Port();
}

// ====================================================================
// Parse command line arguments.
// ====================================================================

void ConsoleClient::ParseArgs(int argc, char *argv[])
{
   // Long options.
   myargs.AddOption('d', "-debug",    false, 3);
   myargs.AddOption('h', "-help",     false, 3);
   myargs.AddOption('b', "-heartbeat",false, 3);
   myargs.AddOption('z', "-health",   false, 3);
   myargs.AddOption('M', "-host",     true, 3);
   myargs.AddOption('i', "-info",     false);
   myargs.AddOption('l', "-locate",   true, 1);
   myargs.AddOption('P', "-port",     true);
   myargs.AddOption('q', "-shutdown", true,3);
   myargs.AddOption('r', "-reload",   false);
   myargs.AddOption('S', "-service",  true, 3);
   myargs.AddOption('s', "-stdout",   false,3);
   myargs.AddOption('x', "-status",   false),3;
   myargs.AddOption('v', "-verbose",  false, 3);
   myargs.AddOption('V', "-version",  false, 3);
   myargs.AddOption('w', "-who",      false, 3);

   // Single-char options.
   myargs.AddOption('M', "-M",        true);
   myargs.AddOption('P', "-P",        true);
   myargs.AddOption('V', "-V",        false);
   myargs.AddOption('b', "-b",        false);
   myargs.AddOption('d', "-d",        false);
   myargs.AddOption('h', "-h",        false);
   myargs.AddOption('i', "-i",        false);
   myargs.AddOption('l', "-l",        true);
   myargs.AddOption('q', "-q",        true);
   myargs.AddOption('s', "-s",        false);
   myargs.AddOption('v', "-v",        false);
   myargs.AddOption('w', "-w",        false);
   myargs.AddOption('x', "-x",        false);
	
   try
   {
      myargs.ParseArgs(argc, argv);
   }
   catch (CmdErr e)
   {
      SYSLOG << Severity(Syslog::error)
		<< e.error 
		<< ": " 
		<< e.option 
		<< endl; 
      UsageShort();
      exit(1);
   }      
	
   // Extract the program name from the first argument, and
   // determine the path by which we were invoked.
	
   if (myargs.ArgCount() > 0)
   {
      string fullname = myargs.Arg(0);
      
      string::size_type pos = fullname.find_last_of("/");
      
      if (pos == string::npos)
      {
         GLOBAL.Progname(fullname);
      }
      else
      {
         GLOBAL.Progname(fullname.substr(pos+1));
      }
      SYSLOG.ProgName(GLOBAL.Progname());
   }
	
}

// ====================================================================
// Process command line arguments.
// ====================================================================

void ConsoleClient::ProcessArgs(void)
{
   bool usestdout = false;
   string arg = "*";
   
   if (myargs.ArgCount() > 1) 
   {
      arg = myargs.Arg(1);
   }
   
   if (myargs.ArgCount() > 2) 
   {
      cerr << "Extraneous: " << myargs.Arg(2) << endl;
   }

   for (int i = 0; i < myargs.OptCount(); i++)
   {
      int opt = myargs.Option(i);
      
      switch(opt)
      {
          case 'b':             // -b | -heartbeat
             usestdout = true;
             function = new Command(arg, SHOWHB);
             break;
             
          case 'd':             // -d | -debug
             GLOBAL.Debug(true);
             break;
             
          case 'h':             // -h | -help
             UsageLong();
             exit(0);
             
          case 'l':             // -l | -locate
             usestdout = true;
             function = new Command("%", SHOWLOC + myargs.Value(i));
             break;
             
          case 'q':             // -q | -shutdown
             usestdout = true;
             function = new Command(myargs.Value(i), SHUTDOWN);
             break;
             
          case 'r':             // -reload
             function = new Command(arg, RELOAD);
             
          case 's':             // -s | -stdout
             usestdout = true;
             break;
             
          case 'w':             // -w | -who
             usestdout = true;
             function = new Command(arg, SHOWUSERS);
             break;
             
          case 'v':             // -v | -verbose
             // Increments by one every time called.
             GLOBAL.Verbosity(+1);
             break;
             
          case 'x':             // -x | -status
             usestdout = true;
             function = new Command(arg, SHOWSTATUS);
             break;
             
          case 'i':             // -i | -info
             usestdout = true;
             function = new Command(arg, SHOWINFO);
             break;
             
          case 'z':             // -health
             usestdout = true;
             function = new Command(arg, SHOWHEALTH);
             break;
             
          case 'M':             // -M | -host
             masterlist = MakeHostlist(myargs.Value(i));
             break;

          case 'S':             // -service
             gssservice = myargs.Value(i);
             break;

          case 'P':             // -P | -port
             try { port = StrToInteger(myargs.Value(i)); }
             catch(string str) { cerr << str << endl; exit(1); }
             break;

          case 'V':             // -V | -version
             cout << version << endl;
             exit(0);
      }
   }

   if (usestdout) 
   {
      StdOut = new File(1); // Point to process's stdout.
      server.RegisterTerminal(StdOut);
   }
   
   if (! function) 
   {
      if (arg == "*") 
      {
         cerr << "No console specified for connection" << endl;
         exit(1);
      }
      function = new Connection(arg);
   }
}

// ====================================================================
// Parse an options file.  There may be multiple overriding ones.
// ====================================================================

void ConsoleClient::ReadConfigs(void)
{
   for (int i=0; i<configs.size(); i++)
      ReadConfig(configs[i]);
}

void ConsoleClient::ReadConfig(string filename)
{
   if (! CheckFile(filename, "optional"))
   {
      string::size_type pos = filename.find("bin/");
      
      if (pos != string::npos)
      {
         filename.replace(pos, 4, "etc/");
      
         if (! CheckFile(filename, "optional")) return;
      }
   }
   
      if (GLOBAL.Verbose()) cout << GLOBAL.Progname() 
                                 << ": Reading options from " 
                                 << filename << endl;

   ifstream configstream(filename.c_str());
	
   vector<string> tokens;
   string line;
   int lineNum;
   
   while (getline(configstream, line, '\n'))
   {
      ++lineNum;
      
      // Tokenize the line.
      
      tokens = Tokenize(line);
      
      // Skip blank lines and comment lines (no tokens found).
      
      if (tokens.empty()) continue;
      
      Lower(tokens[0]);
      
      if (GLOBAL.Verbose()) cout << line << endl;
         
      if (tokens[0] == "master")
      {
         masterlist = MakeHostlist(tokens);
      }
      
      else if (tokens[0] == "port")
      {
         try { port = StrToInteger(tokens[1]); }
         catch(string str) { cerr << str << endl; exit(1); }
      }

      else if (tokens[0] == "service")
      {
         gssservice = tokens[1];
      }

#ifdef HAVE_LIBSSL      
      else if (tokens[0] == "cafile" ||
               tokens[0] == "cert"   ||
               tokens[0] == "key")
      {
         if (tokens.size() < 2) 
            cout << GLOBAL.Progname() 
                 << ": Missing " << tokens[0] << " name in " 
                 << filename << ": " << line << endl;
         else
         {
            if (CheckFile(tokens[1], "optional"))
            {
               if (tokens[0] == "cafile") SocketSSL::SetCAFile  (tokens[1].c_str());
               if (tokens[0] == "cert"  ) SocketSSL::SetCertFile(tokens[1].c_str());
               if (tokens[0] == "key"   ) SocketSSL::SetKeyFile (tokens[1].c_str());
            }
            else
            {
               if (GLOBAL.Verbose())
               {
                  cout << GLOBAL.Progname()
                       << ": cafile " << tokens[1] 
                       << " not found, "
                       << (tokens[0] == "cafile" ? "host certs will not be verified" :
                                                   "client will not be authenticated"
                          )
                       << endl;
               }
            }
         }
         
      }
#endif

      else
      {
         cout << GLOBAL.Progname() << ": Invalid config file line in " 
              << filename << ": " << line << endl;	
      }
   }
}

// ====================================================================
// Run the command now.
// ====================================================================

void ConsoleClient::Run(void)
{
   if (GLOBAL.Verbose())
      cout << version << endl;

   server.SetGSSService(gssservice);
   
   // Mask unwanted signals.
      
   sigset_t signalset, signalnew;
   sigemptyset(&signalset);
   sigaddset(&signalset, SIGHUP);
   sigaddset(&signalset, SIGPIPE);
   pthread_sigmask(SIG_BLOCK, &signalset, &signalnew);

   if (isatty(terminal.FD())) terminal.SaveState();
   
   try
   {
      // Execute the requested function.

      response = function->Execute();

      StdOut->Write("[");
      StdOut->Write(response);
      StdOut->Write("]\r\n");

      // Join so that execution of the main thread will be suspended
      // until the server.ReaderProc() thread finishes. 
      
      server.Join();
      
      // if the socket thread is finished then there is no need for
      // terminal thread to continue
      
      terminal.StopReader();
   }
   
   
   catch (File::IOErr e)
   {
      if (isatty(terminal.FD())) terminal.RestoreState();
      
      SYSLOG << Severity(Syslog::error) 
             << e.routine 
             << ": " << e.filename
             << ": " << e.message 
             << endl;
      
      exit (1);
   }

   catch (string msg)
   {
      if (isatty(terminal.FD())) terminal.RestoreState();

      cerr << msg << endl;
      
      exit (1);
   }

   catch (char *msg)
   {
      if (isatty(terminal.FD())) terminal.RestoreState();
      cerr << msg << endl;
      exit(1);
   }

   if (isatty(terminal.FD())) terminal.RestoreState();
}


// ====================================================================
// Issue commands to connect to a console.
// ====================================================================

string ConsoleClient::Connect(const string &target)
{
   Hostlist hl(masterlist);

   
   // Loop trying to connect to client; may get redirected to other servers.
      
   for (;;)
   {
      if (GLOBAL.Verbose()) cerr << "Requesting console " << target << "\r\n";
      
      response = Cmd(hl, CONNECT + target + string("\n"));

      // Catch a few negative responses, let the rest fall through.
      
      string::size_type pos = response.find_first_of(":");
      if (pos == string::npos) pos = 0;
      else pos += 2;

      if (response.substr(pos) == "unknown") return string("Unknown console: ") + target;
      if (response.substr(pos) == "not allowed") return string("Not authorized to connect to ") + target;
      
      
      if (response.substr(pos, 10) == "redirect: ")
      {
         server.Close();
         string referhost = response.substr(pos+10);
         
         if (referhost.find_first_of(".") == string::npos)
            referhost += "." + Host::LocalDomain();
         
         if (GLOBAL.Verbose()) cerr << "Console " << target
                                    << " is served by console server " 
                                    << referhost 
                                    << "\r\n";
         hl = MakeHostlist(referhost);
      }
      else
      {
         break;
      }
   }

   // Next, the terminal must be put in a raw state so that
   // all keystrokes will be sent as-is to the server. 
   
   if (isatty(terminal.FD())) 
   {
      terminal.SetMode(SPEED);
      terminal.SetRaw();
      terminal.Blocking();
   }
   
   terminal.StartReader();
      
   if (GLOBAL.Verbose()) cerr << "Terminal reader thread started\r\n";

   server.StartReader();
      
   if (GLOBAL.Verbose()) cerr << "Server reader thread started\r\n";

   if (GLOBAL.Verbose()) cerr << "Entering interactive session\r\n";
      
   return response;
}

// ====================================================================
// Send a command to multiple servers.  If target is '*', ask a 
// master server for the list of all available servers. Then send
// the command to each server in the list.  If there is an error with
// a given server, proceed to the next.
// ====================================================================

string ConsoleClient::Distribute(const string &t, const string &cmd)
{
   string response;
   Hostlist hl;
   vector<string> serverlist;
   string target = t;
   
   if (target == "*")
   {
      target = Cmd(masterlist, GETSERVERS); // Get server list.

      server.Close();
   }

   if (GLOBAL.Verbose()) cerr << "Distribute " << cmd << " to " << target << endl;

   serverlist = Tokenize(target);
   
   vector<string>::iterator srv = serverlist.begin();
   
   for ( ; srv != serverlist.end(); srv++)
   {
      StdOut->Write(*srv);
      StdOut->Write(":\r\n");
      
      hl = MakeHostlist(*srv);
      
      try
      {
         response = Cmd(hl, cmd + SIGNOFF);
         
         server.ReadToEnd();
         
         server.Close();
      }
      catch(File::IOErr e)
      {
         cerr << e.routine 
              << ": " << e.filename
              << ": " << e.message 
              << endl;
      }
      catch(string s)
      {
         cerr << s << "\n" << endl;
      }
   }

   return response;
}

// ====================================================================
// Issue a command to a server and get a response.  Leave the connection
// open when returning to the caller so it can read additional 
// information from the stream.  This is the routine that takes care
// of all initial signin protocol with the server.
// The routine takes a list of servers for failover; use the first
// server that responds.  
// ====================================================================

string ConsoleClient::Cmd(Hostlist &hl, const string &cmdstring)
{
   // Connect to the server.  The first server in the list that responds is 
   // what we'll use.

   int i;
   string attempts;
   
   for (i = 0; i < hl.size(); i++)
   {
      if (hl[i].Port() == 0) hl[i].SetPort(port);
      
      try 
      {
         attempts += (attempts.empty() ? "" : ", ") + hl[i].Name();
         if (GLOBAL.Verbose()) cerr << "Attempting to connect to "
                                    << hl[i].Name() 
                                    << " port "
                                    << hl[i].Port()
                                    << endl;
         server.Connect(hl[i]);
         break;
      }
      catch (File::IOErr e)
      {
         if (!(e.reason != EHOSTDOWN ||
               e.reason != ECONNREFUSED)
             || GLOBAL.Verbose()) 
            cerr << "Error " << e.reason << ": " 
                 << e.message 
                 << " connecting to " 
                 << hl[i].Name() << endl;
         server.Close();
         continue;
      }
   }

   if (i >= hl.size())
   {
      throw string("Error, unable to connect to server on " + attempts);
   }

#ifdef HAVE_LIBSSL
   Buffer sslmsg;
   if (hl[i].Name() != "localhost" ) {
      
      server.GetSSLCipher(sslmsg);
   }
#endif
         
   if (GLOBAL.Verbose()) cerr << "Connected to console server " << hl[i].Name();
#ifdef HAVE_LIBSSL
   if (hl[i].Name() != "localhost" ) {
	   if (GLOBAL.Verbose()) cerr << " using " << sslmsg.Cstr() ;
   }
#endif
   if (GLOBAL.Verbose()) cerr << "\r\n";

   // Initial sign in to server, send our username.

   server.Send(ESCAPE);
   server.Send(';');  
   server.Send(username);
   server.Send('\n');
   if (GLOBAL.Verbose()) cerr << "Username " << username << " sent\r\n";

   response = server.GetNextResponse();

   if (response != "ok")  // response to username
   {
      if (GLOBAL.Verbose()) cerr << "Unexpected response from server: " << response << "\r\n";
      
      throw(response);
   }
   
   server.Send(cmdstring);
         
   response = server.GetNextResponse();

   // Server may have requested our password
   
   if (response == "pass")
   {
      if (GLOBAL.Verbose()) cerr << "Sending password to server\r\n";
      
      terminal.Write("Enter password: ");
      pass = terminal.GetLine();
      server.Send(pass);
      server.Send('\n');
      response = server.GetNextResponse();
   }

   return response;
}

// ====================================================================
// Make a hostlist from a string.
// ====================================================================

Hostlist ConsoleClient::MakeHostlist(const string &s)
{
   vector<string> names = Tokenize(s);
   return MakeHostlist(names);
}

Hostlist ConsoleClient::MakeHostlist(vector<string> &names)
{
   Hostlist newlist;
   vector<string>::iterator n = names.begin();
   
   for ( ; n != names.end(); n++)
   {
      Host host;

      // Host can have an optional port number.
      
      string::size_type pos = (*n).find_first_of(":");
      if (pos != string::npos)
      {
         try { host.SetPort(StrToInteger((*n).substr(pos+1))); }
         catch(string str) 
         { 
            cerr << "Invalid port in " << *n << ": " << str << endl; 
            exit(1); 
         }
         (*n).erase(pos);
      }

      // Make sure the host exists.

      if (!host.Find(*n))
      {
         cout << GLOBAL.Progname() << ": server " << *n << " not found" << endl;
         exit(1);
      };

      newlist.push_back(host);
   }
   return newlist;
}

// ====================================================================
// Usage help
// ====================================================================

void ConsoleClient::UsageShort(void)
{
   cerr << GLOBAL.Progname() << " [-bdhivwx] [-q server] [-M server] [-P port] console" << endl;
}

void ConsoleClient::UsageLong(void)
{
   UsageShort();
   
   cout
      << "-b | -heartbeat    get heartbeat info from servers\r\n"
      << "-d | -debug        debug -- trace internal protocol\r\n"
      << "-h | -help         output this message\r\n"
      << "     -health       check health of server\r\n"
      << "-i | -info         display configuration of all lines\r\r"
      << "-l | -locate name  locate a console\r\n"
      << "-q | -shutdown srv shutdown consoled on server srv\r\n"
      << "-v | -verbose      be verbose on startup\r\n"
      << "-w | -who          display who is on all consoles\r\n"
      << "-x | -status       display status of all consoles\r\n"
      << "-M | -host         specify console server to contact first\r\n"
      << "-P | -port         specify port number of console server\r\n"
      << "-V | -version      output version info\r\n"
      << endl;
};
