// ====================================================================
// consoled.cc - process options, process configuration file, startup the
//               dispatcher to create the threads to listen to the serial
//               lines and await incoming connections.
// ==================================================================== 

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.22 $ $Date: 2006/12/07 07:47:29 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include "cmdargs.h"
#include "console.h"
#include "consoled.h"
#include "defaults.h"
#include "file.h"
#include "global.h"
#ifdef HAVE_LIBGSSAPI
#include "gssapi.h"
#endif
#include "host.h"
#include "psstream.h"
#include "syslog.h"
#include "utility.h"

#include <fstream>
#include <iostream>
#include <netdb.h>
#include <pwd.h>
#include <set>
#include <signal.h>
#include <stdlib.h>
#include <string>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <unistd.h>
#include <vector>

extern int errno;

// --------------------------------------------------------------------
// Storage for global objects.
// --------------------------------------------------------------------

Global        GLOBAL;

Syslog        SYSLOG(CONSOLE_PROGNAME, Syslog::LOGCLASS, Syslog::info);

ConsoleDaemon DAEMON;

Dispatcher    DISPATCHER; 

// ====================================================================
// Main routine, start the primary object.
// ====================================================================

int main(int argc, char *argv[])
{ 
   // -----------------------------------------------------------------
   // Run the program now.
   // -----------------------------------------------------------------

   DAEMON.Run(argc, argv);
   
   exit(0);
}

// ====================================================================
// Constructor method to initialize.
// ====================================================================

   // Small functions to start the threads.

   static void *MonitorSignals(void *data);
   static void *RestartMsgserv(void *data);


ConsoleDaemon::ConsoleDaemon(void)
{
   extern char BUILD[];

   // Set default string values.

   GLOBAL.Progname(CONSOLE_PROGNAME);
   configfile = GLOBAL.Progname() + ".cf";

   // Set socket constants.

   escape  = ESCAPE;

   monitorThread = 0;

   startmsg.TimeStampFormat("%D %T ");

   SYSLOG.SetHistory(40);

   dosetuid = false;
   idleForce = IDLE_FORCE;
   idleWarn  = IDLE_WARN;

   version = VERSION;
   version += " ";
   version += BUILD;
#ifdef HAVE_LIBSSL
   version += " (with SSL";
#else
   version += " (without SSL";
#endif
#ifdef HAVE_LIBGSSAPI
   version += ", with GSSAPI";
#else
   version += ", without GSSAPI";
#endif
#ifdef HAVE_LIBPAM
   version += ", with PAM)";
#else
   version += ", without PAM)";
#endif

}

   
// ====================================================================
// Destructor
// ====================================================================

ConsoleDaemon::~ConsoleDaemon(void)
{
   // Later.
}

// ====================================================================
// Start services running.
// ====================================================================

void ConsoleDaemon::Run(int argc, char **argv)
{
   gssrealm = Host::LocalDomain();
   Upper(gssrealm);
   
   gssservice = "console/" + Host::LocalFQName();
   gsskeytab  = KEYTAB;

   try
   {
      // -----------------------------------------------------------------
      // Parse the command line arguments into the global object.
      // Report on the startup if being verbose.
      // -----------------------------------------------------------------
      
      ParseArgs(argc, argv);
      
      if (!GLOBAL.SyntaxOnly())
      {
         SYSLOG << Severity(Syslog::info)
                << version
                << endl;
         SYSLOG << Severity(Syslog::info)
                << GLOBAL.Progname()
                << " starting up on "
                << Host::LocalNode()
                << " with config file: " 
                << ConfigFile()
                << endl;
      }
      
      startmsg += GLOBAL.Progname() + " started with config file " + ConfigFile() + "\r\n";
      startmsg.TimeStamp();

      // -----------------------------------------------------------------
      // Read the config file and parse it.
      // -----------------------------------------------------------------
      
      FetchConfig();
      ReadConfig();

      // If only doing a syntax check, exit now.

      if (GLOBAL.SyntaxOnly()) exit(0);
      
      // If using GSSAPI, load credentials now.
#ifdef HAVE_LIBGSSAPI
      if (GLOBAL.Verbose())
      {
         SYSLOG << Severity(Syslog::info)
                << "GSSAPI Realm:   " << gssrealm << "\n"
                << "       Keytab:  " << gsskeytab << "\n"
                << "       Service: " << gssservice << endl;
      }
      // Should we override the env variable, or let the environment override?
      static char setkeytab[256]; // Env string must be static
      strcpy(setkeytab, "KRB5_KTNAME=");
      strncat(setkeytab, gsskeytab.c_str(), 240);
      putenv(setkeytab);
      GSSAPI::DefaultCreds(gssservice);
#endif
      // If requested, setuid to a (presumably) less-privileged userid.

      if (dosetuid)
      {
         struct passwd *pw;
         pw = getpwnam(runas.c_str());
         
         if (!pw) 
         {
            SYSLOG << Severity(Syslog::error)
                   << "Userid  "             << runas
                   << " not found!"
                   << endl;
         }
         
         if (setgid(pw->pw_gid) < 0)
         {
            SYSLOG << Severity(Syslog::error)
                   << "Cannot switch to gid "  << (int) pw->pw_gid
                   << " (user "                << runas
                   << ")!"
                   << endl;
            exit(1);
         }
         
         if (setuid(pw->pw_uid) < 0)
         {
            SYSLOG << Severity(Syslog::error)
                   << "Cannot switch to uid "  << (int) pw->pw_uid
                   << " (user "                << runas
                   << ")!"
                   << endl;
            exit(1);
         }

         SYSLOG << Severity(Syslog::info)
                << "Switched to user "         << runas
                << endl;
      }
      
      // Become a daemon now.  Do first before any other threads are spawned.
      // Also need to do before we save the PID so we get the right PID.
   
      if (GLOBAL.Daemonize())
      {
         SYSLOG.Echo(false);
         Daemonize();
      }
   
      // Save current PID if requested.
      
      if (! pidfile.empty())
      {
         ofstream pidf;
         pidf.open(pidfile.c_str());
         pidf << getpid() << endl;
         pidf.close();
      }
      
      // Mask unwanted signals.
      
      sigset_t signalset, signalnew;
      sigemptyset(&signalset);
      sigaddset(&signalset, SIGHUP);
      sigaddset(&signalset, SIGINT);
      sigaddset(&signalset, SIGQUIT);
      sigaddset(&signalset, SIGILL);
      sigaddset(&signalset, SIGSEGV);
      sigaddset(&signalset, SIGPIPE);
      sigaddset(&signalset, SIGTERM);
      pthread_sigmask(SIG_BLOCK, &signalset, &signalnew);

      // Start a thread to monitor for signals.
      
      int status = pthread_create(&monitorThread, NULL, MonitorSignals, NULL);
      
      if (status != 0)
      {
         SYSLOG << Severity(Syslog::error) 
                << "Thread error " 
                << status << strerror(status) 
                << endl;
         exit(1);
      }
      
      pthread_detach(monitorThread);
   
      // Open the message service if one was specified.
      
      if (msgserv.AddrSet())
      {
         if (GLOBAL.Debug())
         {
            SYSLOG << Severity(Syslog::debug)
                   << "Opening message service "
                   << msgserv.Name() << ":" << msgserv.Port()
                   << endl;
         }
         
         streamlist::iterator cur, last;
         last = streams.end();
         for (cur = streams.begin(); cur != last; cur++)
         {
            string streamname = cur->first;
            streamblock *sb = cur->second;
            
            try 
            {
               sb->mc.OpenStream(SUPPLIER, msgserv, streamname);
               sb->streamOpen = true;
               sb->restartNeeded = false;
            }
            catch(File::IOErr e)
            {
               SYSLOG << Severity(Syslog::error)
                      << "Error opening Message Service "
                      << msgserv.Name() << ":" << msgserv.Port()
                      << ": " << e.message 
                      << endl;
               SYSLOG << "Will continue trying" << endl;
               sb->streamOpen = false;
               sb->restartNeeded = true;
            }
            
            // Start a thread to monitor the message service if tcp.
            
            sb->restartThread = 0;
            if (sb->mc.ConnectionIs(TCP))
            {
               int status = pthread_create(&sb->restartThread, NULL, RestartMsgserv, sb);
               
               if (status != 0)
               {
                  SYSLOG << Severity(Syslog::error) 
                         << "Thread error " 
                         << status << strerror(status) 
                         << endl;
                  exit(1);
               }
               
               pthread_detach(sb->restartThread);
            }
         }
      }
   }
      
   catch (File::IOErr(e))
   {
      SYSLOG << Severity(Syslog::error)
             << "Uncaught I/O error "
             << e.routine 
             << ": " << e.filename 
             << ": " << e.message 
             << endl;
      
      exit(1);
   }
   catch(string s)
   {
      SYSLOG << Severity(Syslog::error)
             << "Uncaught exception "
             << s
             <<endl;
      exit(1);
   }
   catch(...)
   {
      SYSLOG << Severity(Syslog::error)
             << "Uncaught unknown exception"
             << endl;
      
      exit(1);
   }
   
   // -----------------------------------------------------------------
   // Run the dispatcher now, which will set up the console lines and
   // the threads that service them.  It will then wait for incoming
   // client connections and start threads to service them.  When
   // it receives a quit command it will return here.
   // ----------------------------------------------------------------- 
   
   DISPATCHER.Run();

   // On return we are exiting.  Cancel our sub-threads on the way out.

   if (monitorThread != 0) pthread_cancel(monitorThread);
   streamlist::iterator cur;
   for (cur = streams.begin(); cur != streams.end(); cur++)
   {
      if (cur->second->restartThread != 0) pthread_cancel(cur->second->restartThread);
   }
   
   // Remove any pid file.

   if (! pidfile.empty()) unlink(pidfile.c_str());
}

// ====================================================================
// Parse ConsoleDaemon option list.
// ====================================================================

void ConsoleDaemon::ParseArgs(int argc, char *argv[])
{
   CmdArgs myargs;

   // Long options.
   myargs.AddOption('C',  "-config",  true);
   myargs.AddOption('D',  "-nodaemon",false, 3);
   myargs.AddOption('d',  "-debug",   false, 3);
   myargs.AddOption('h',  "-help",    false, 3);
   myargs.AddOption('m',  "-msgserv", true,  3);
   myargs.AddOption('p',  "-pid",     true,  3);
   myargs.AddOption('P',  "-port",    true);
   myargs.AddOption('r',  "-runas",   true,  3);
   myargs.AddOption('S',  "-syntax",  false, 3);
   myargs.AddOption('v',  "-verbose", false, 3);
   myargs.AddOption('V',  "-version", false, 3);   

   // Single-char options.
   myargs.AddOption('C',  "-C",       true);
   myargs.AddOption('d',  "-d",       false);
   myargs.AddOption('D',  "-D",       false);
   myargs.AddOption('h',  "-h",       false);
   myargs.AddOption('P',  "-P",       true);
   myargs.AddOption('r',  "-r",       true);
   myargs.AddOption('S',  "-S",       false);
   myargs.AddOption('V',  "-V",       false);
   myargs.AddOption('v',  "-v",       false);
   
   try { myargs.ParseArgs(argc, argv);  }
   catch (CmdErr e)
   {
      SYSLOG << Severity(Syslog::error)
             << e.error 
             << ": " 
             << e.option 
             << endl; 
      UsageShort();
      exit(1);
   }      

   // Extract the program name from the first argument, and
   // determine the path by which we were invoked.

   if (myargs.ArgCount() > 0)
   {
      string fullname = myargs.Arg(0);
      
      string::size_type pos = fullname.find_last_of("/");
      
      if (pos == string::npos)
      {
         GLOBAL.Progname(fullname);
      }
      else
      {
         GLOBAL.Progname(fullname.substr(pos+1));
         bindir = fullname.substr(0,pos+1);
      }
      SYSLOG.ProgName(GLOBAL.Progname());
   }

   if (myargs.ArgCount() > 1)
   {
      cerr << "Unrecognized: " << myargs.Arg(1) << endl;
      exit(1);
   }

   configfile = GLOBAL.Progname() + ".cf";

   // Loop through the remaining options
   
   for (int i = 0; i < myargs.OptCount(); i++)
   {
      int port;
      string ms;
      string::size_type pos;
      
      int opt = myargs.Option(i);
      
      switch(opt)
      {
          case 'C':             // -C | -config
             configfile = myargs.Value(i);
             break;

          case 'd':             // -d | -debug
             GLOBAL.Daemonize(false);
             GLOBAL.Debug(true);
             break;
             
          case 'D':             // -D | -nodaemon
             GLOBAL.Daemonize(false);
             break;
             
          case 'h':             // -h | -help
             UsageLong();
             exit(0);
             
          case 'm':             // -msgserv
             ms = myargs.Value(i);
             pos = ms.find_first_of(":");
             if (pos != string::npos)
             {
                try { msgserv.SetPort(StrToInteger(ms.substr(pos+1))); }
                catch(string str) { cerr << str << endl; exit(1); }
                ms.erase(pos);
             }
             if (!msgserv.Find(ms))
             {
                cerr << "Cannot find message server " << ms << endl;
                exit(1);
             }
             break;             

          case 'p':             // -pid
             pidfile = myargs.Value(i);
             break;
             
          case 'P':             // -P | -port
             try { port = StrToInteger(myargs.Value(i)); }
             catch(string str) { cerr << str << endl; exit(1); }
             localport.SetPort(port);
             break;
             
          case 'r':             // -r | -runas
             dosetuid = true;
             runas = myargs.Value(i);
             break;
             
          case 'S':             // -S | -syntax
             GLOBAL.SyntaxOnly(true);
             break;
             
          case 'v':             // -v | -verbose
             // Increments by one every time called.
             GLOBAL.Verbosity(+1);
             break;
             
          case 'V':             // -V | -version
             cout << version << endl;
             exit(0);

          default:
             cerr << "Unrecognized option";
             exit(1);
      }
   }

   // Now that we had a chance to get a port number from the command line
   // see if we should use that or our default.

   if (localport.Port() == 0)
   {
      localport.SetPort(CONSOLE_PORT);
      localport.FindPort(CONSOLE_SERVICE, "tcp");
   }

   if (msgserv.Port() == 0)
   {
      msgserv.SetPort(MESSAGE_PORT);
      msgserv.FindPort(MESSAGE_SERVICE, "tcp");
   }
}

// ====================================================================
// Fetch the configuration file from the message service if requested
// and cache it on local disk.  If not available, use the cached 
// version.  The configfile name can contain a relative or absolute
// path, which is used to specify where the cached copy is placed,
// but is stripped from the name requested from the server.  The
// name may also have a server named following the "@", and a port
// number following a colon.  E.g. "lib/consoled.cf@console:611"
// says to fetch file "consoled.cf" from a server named console
// at port 611, and to cache the file in "lib/consoled.cf.cached".
// ====================================================================

void ConsoleDaemon::FetchConfig(void)
{
   string::size_type pos = configfile.find_first_of("@");
   
   if (pos == string::npos) return; // No '@', requesting local file.
   
   // Default values

   Host cfgserv = msgserv;
   string cfgclass = configfile.substr(0,pos);
   string cfgservname;

   // If a server name specified, pull it off and use it.

   if (pos+1 < configfile.size()) cfgservname = configfile.substr(pos+1);
   configfile.erase(pos);
   configfile += ".cache";

   if (! cfgservname.empty())
   {
      // If a port number followed the sever name, pull it off.
      
      pos = cfgservname.find_first_of(":");
      if (pos != string::npos)
      {
         int    cfgport;
         try  { cfgport = StrToInteger(cfgservname.substr(pos+1)); }
         catch(string s) { cerr << s << endl; exit(1);  }
         cfgservname.erase(pos);
         cfgserv.SetPort(cfgport);
      }

      cfgserv.Find(cfgservname);
   }
   

   // If a path name was specified, remove it from the requested name.

   pos = cfgclass.find_last_of("/");
   if (pos != string::npos) cfgclass.erase(0,pos+1);

   // Request the file from the message server now.

   MessageClient cfg;
   try {  cfg.OpenStream(SUBSCRIBER, cfgserv, cfgclass); }
   catch(File::IOErr e)
   {
      SYSLOG << Severity(Syslog::error)
             << "Cannot open Message Service "
             << cfgserv.Name()
             << " to fetch configuration file: " << cfgclass
             << endl;
      SYSLOG << "Using cached copy: " << configfile << endl;
      return;
   }

   // Now read the config file and copy it to disk.  

   Buffer buffer;
   File cache;
   
   try
   {
      cache.Open(configfile, O_RDWR | O_CREAT | O_TRUNC);
      cfg.Receive(buffer);
      string temp = (char *) buffer.GetLine();
      if (temp != "[ok]")
      {
         cerr << temp << endl;
         exit(1);
      }
      if (buffer.Dirty()) cache.Write(buffer);
      buffer.Clear();
      
      while (cfg.Receive(buffer))
      {
         cache.Write(buffer);
         buffer.Clear();
      }
      cache.Close();
      cfg.CloseStream();
   }
   catch(File::IOErr(e))
   {
      SYSLOG << Severity(Syslog::error)
             << "Error caching configuration file: "
             << e.message
             << endl;
      exit(1);
   }
   
   SYSLOG << Severity(Syslog::info)
          << "Configuration file "
          << cfgclass
          << " retrieved from "
          << cfgserv.Name()
          << endl;
}

// ====================================================================
// Read in the configuration file, record the data in the object
// ====================================================================

void ConsoleDaemon::ReadConfig(void)
{
   int lineNum = 0;
   string newline, line;
   struct stat statbuf;
   Line dflt;                   // Default line settings.
   bool parseError = false;     // Set if parsing error.
   set<string> consoles;        // Check for duplicate lines
   set<string> lines;           //   or consoles.

   // Open input file

   if (stat(configfile.c_str(), &statbuf) < 0)
   {
      SYSLOG << Severity(Syslog::error)
             << "Cannot open configuration file " 
             << configfile
             << ": "
             << strerror(errno)
             << endl;
      exit(1);
   }

   ifstream configstream(configfile.c_str());

   vector<string> tokens;

PARSE:
   while (! configstream.eof()) 
   {
      // Gather lines.  Trailing comma indicates a continuation.
      // Comments aren't removed yet.
      line = "";
      while (getline(configstream, newline, '\n'))
      {
         ++lineNum;
         line += newline;
         string::size_type pos = line.find_last_not_of(" \t\n\r");
         if (pos == string::npos || line[pos] != ',')
            break;     
      }
      
      // Tokenize the line.
      
      tokens = Tokenize(line);
      
      // Skip blank lines and comment lines (no tokens found).
      
      if (tokens.empty()) continue;
      
      // Make sure at least one argument follows

      if (tokens.size() < 2)
      {
         SYSLOG << Severity(Syslog::error)
                << configfile 
                << "(" 
                << lineNum
                << "): Invalid statement: " 
                << tokens[0] 
                << endl;

         parseError = true;
         continue;
      }

      Lower(tokens[0]);

      // This is a logdir statement

      if (tokens[0] == "logdir")
      {
         logdir = tokens[1];
      }
#ifdef HAVE_LIBSSL
      // This is a CAFile statement

      else if (tokens[0] == "cafile")
      {
         SocketSSL::SetCAFile(tokens[1].c_str());
      }

      // This is a Cert statement

      else if (tokens[0] == "cert")
      {
         SocketSSL::SetCertFile(tokens[1].c_str());
      }

      // This is a Key statement

      else if (tokens[0] == "key")
      {
         SocketSSL::SetKeyFile(tokens[1].c_str());
      }
#endif
#ifdef HAVE_LIBGSSAPI
      // This is a GSSAPI statement

      else if (tokens[0] == "gssapi")
      {
         int i = 1;
         while (i < tokens.size())
         {
            Lower(tokens[i]);
            if (tokens[i] == "service")
            {
               gssservice = tokens[++i];
            }
            else if (tokens[i] == "realm")
            {
               gssrealm = tokens[++i];
            }
            else if (tokens[i] == "keytab")
            {
               gsskeytab = tokens[++i];
            }
            else
            {
               SYSLOG << Severity(Syslog::error)
                      << configfile 
                      << "(" 
                      << lineNum
                      << "): Invalid option " << tokens[i] 
                      << " on gssapi statment"
                      << endl;

               parseError = true;
               continue;
            }
            i++;
         }
      }

#endif
      // This is a speed statement

      else if (tokens[0] == "speed")
      {
         defaultSpeed = tokens[1];
      }

      // This is an acl

      else if (tokens[0] == "acl")
      {
         string name = tokens[1];
         ACL *acl = new ACL;
         acls[name] = acl;
         
         for (int i = 2; i < (int) tokens.size(); i++)
         {
            // Find a prefix on the acl that gives the type of
            // entry this is.  The default is host.
            string temp = tokens[i];
            string::size_type pos = temp.find_first_of(":");
            string type = "host";
            if (pos != string::npos)
            {
               type = temp.substr(0, pos);
               temp.erase(0,pos+1);
            }
            // Find the security token string and host part
            // separated by '@'.
            string tok = temp;
            string host = "";
            pos = temp.find_last_of("@");
            if (pos != string::npos)
            {
               tok  = temp.substr(0, pos);
               host = temp.substr(pos+1);
            }
            if (type == "host")
               acl->Add(new ACE_Host(tok, host));
            else if (type == "k5")
            {
               if (tok.find_first_of("@") == string::npos)
                  tok += "@" + gssrealm;
               acl->Add(new ACE_GSSAPI(tok, host));
            }
            else if (type == "cert")
            {
               acl->Add(new ACE_CERT(tok, host));
            }
            else if (type == "acl")
            {
               ACL *subacl = acls[temp];
               if (subacl == NULL)
               {
                  SYSLOG << Severity(Syslog::error)
                         << configfile 
                         << "(" 
                         << lineNum 
                         << "): undefined ACL:" 
                         << tok
                         << endl;
                  parseError = true;
               }
               else acl->Add(new ACE_ACL(subacl));
            }
            else
            {
               SYSLOG << Severity(Syslog::error)
                      << configfile 
                      << "(" 
                      << lineNum 
                      << "): Unknown ACL type: " 
                      << type
                      << endl;
               parseError = true;
               continue;
            }
            if (GLOBAL.Verbose())
               SYSLOG << Severity(Syslog::info)
                      << configfile 
                      << "(" 
                      << lineNum 
                      << "): Added ACL of type " 
                      << type 
                      << " for "
                      << ((tok == "") ? "ANYONE" : tok)
                      << " from "
                      << ((host == "") ? "ANYWHERE" : host)
                      << endl;
         }
      }      
      
      // This is an idle_timeout statement
      
      else if (tokens[0] == "idle_timeout")
      {
         if (tokens.size() > 1)
         {
            idleWarn  = StrToInteger(tokens[1]); 
         }
         if (tokens.size() > 2)
         {
            idleForce = StrToInteger(tokens[2]);
         }
      }
      
      // This is a line statement
      
      else if (tokens[0] == "line" || tokens[0] == "linegroup")
      {
         Line *lp = new Line(dflt);
         vector<Line *> linelist;
         linelist.push_back(lp);
         
         try { lp->Parse(tokens); }
         catch (string s)
         {
            SYSLOG << Severity(Syslog::error)
                   << configfile 
                   << "(" 
                   << lineNum 
                   << "): " 
                   << s 
                   << endl;
            parseError = true;
            delete lp;
            continue;
         }
         
         if (lp->Name() == "unused")
         {
            delete lp;
            continue;
         }
         
         if (tokens[0] == "linegroup")
         {
            try { ExpandLines(linelist); }
            catch(string s)
            {
               SYSLOG << Severity(Syslog::warning)
                      << configfile 
                      << "(" 
                      << lineNum 
                      << "): " << s
                      << endl;
               delete lp;
               if (GLOBAL.SyntaxOnly()) parseError = true;
               continue;
            }
         }
         
         vector<Line *>::iterator li = linelist.begin();
         
         for( ; li != linelist.end(); li++)
         {
            lp = *li;
            set<string>::iterator pos;
            
            pos = consoles.find(lp->Name());
            if (pos == consoles.end())
               consoles.insert(lp->Name());
            else
            {
               SYSLOG << Severity(Syslog::warning)
                      << configfile 
                      << "(" 
                      << lineNum 
                      << "): Console " << lp->Name()
                      << " previously defined"
                      << endl;
               delete lp;
               if (GLOBAL.SyntaxOnly()) parseError = true;
               continue;
            }
               
            string fqline = lp->DevName() + "@" + lp->Server();
            pos = lines.find(fqline);
            if (pos == lines.end())
               lines.insert(fqline);
            else
            {
               SYSLOG << Severity(Syslog::warning)
                      << configfile 
                      << "(" 
                      << lineNum 
                      << "): Line " << fqline
                      << " previously defined"
                      << endl;
               delete lp;
               if (GLOBAL.SyntaxOnly()) parseError = true;
               continue;
            }
            
            DISPATCHER.RegisterConsole(lp);
            servers.insert(lp->Server());
            delete lp;
         }
      }      
      
      // This is a default statement
      
      else if (tokens[0] == "default")
      {
         try { dflt.Parse(tokens); }
         catch (string s)
         {
            SYSLOG << Severity(Syslog::error)
                   << configfile 
                   << "(" 
                   << lineNum 
                   << "): " 
                   << s 
                   << endl;
            parseError = true;
            continue;
         }
      }      
      
      // This is a macro statement
      
      else if (tokens[0] == "macro")
      {
         if ((int) tokens.size() < 3 || (int) tokens[1].size() != 1)
         {
            SYSLOG << Severity(Syslog::error)
                   << configfile
                   << "("
                   << lineNum
                   << "): Invalid macro definition:"
                   << line
                   << endl;
            if (GLOBAL.SyntaxOnly()) parseError = true;
            continue;
         }

         // "compile" the string by replacing any escapes with
         // the values they represent.  \n cannot appear in the
         // string, so append it as a stopping value.

         string defn = ((int) tokens.size() > 3) ? tokens[3] : tokens[2];
         string comp; // Compiled string

         defn += '\n';
         int state = 1;
         int i = 0;
         string numstr;
         while (i < defn.size())
         {
            char cc = defn[i];
            switch (state)
            {
                case 1: // Just a character
                   if (cc == '\n') break;
                   if (cc == '\\') state = 2;
                   else comp += cc;
                   break;

                case 2: // Escape sequence
                   state = 1; // unless exception found
                   switch (cc)
                   {
                       case '\n': state = -9;     break;
                       case '0':  state = 3;      break;
                       case 't':  comp += '\t';   break;
                       case 'n':  comp += '\n';   break;
                       case 'r':  comp += '\r';   break;
                       case 'a':  comp += '\a';   break;
                       case 'e':  comp += '\e';   break;
                       case 'p':  comp += '\xfe'; break; // pause
                       case 'b':  comp += '\xff'; break; // break
                       default:   comp += cc;     break;
                   }
                   break;
                   
                case 3: // Octal Escape \0nn
                   if (isdigit(cc) && numstr.size() < 2)
                      numstr += cc;
                   else if (numstr.size() > 0)
                   {
                      comp += StrToOctal(numstr);
                      state = -1;
                      numstr = "";
                   }
                   else state = -9;
                   break;
                   
                case 9:
                   SYSLOG << Severity(Syslog::error)
                          << configfile
                          << "("
                          << lineNum
                          << "): Invalid escape in macro definition:"
                          << line
                          << endl;
                   if (GLOBAL.SyntaxOnly()) parseError = true;
                   break;
            }
            if (state >=0) i++;   // negative state means 'redo'
            state = abs(state);
         }
         
         int index = (int) (tokens[1][0]);
         macros[index].first = tokens[2];
         macros[index].second = defn;
      }      
      
      // This is a port statement
      
      else if (tokens[0] == "port")
      {
         int port;
         
         try 
         { 
            port = StrToInteger(tokens[1]); 
         }
         catch(string str) 
         { 
            SYSLOG << Severity(Syslog::error)
                   << configfile
                   << "("
                   << lineNum
                   << "): Invalid port number in:"
                   << line
                   << endl;
            if (GLOBAL.SyntaxOnly()) parseError = true;
            continue;
         }
         
         localport.SetPort(port);
      }
      
      else if (tokens[0] == "runas")
      {
         dosetuid = true;
         runas = tokens[1];
      }
      
      // This is a msgserv statement
      
      else if (tokens[0] == "msgserv")
      {
         string ms;
         string::size_type pos;
         
         ms = tokens[1];
         pos = ms.find_first_of(":");
         if (pos != string::npos)
         {
            try { msgserv.SetPort(StrToInteger(ms.substr(pos+1))); }
            catch(string str) 
            { 
               SYSLOG << Severity(Syslog::error)
                      << configfile
                      << "("
                      << lineNum
                      << "): Invalid port number in:"
                      << line
                      << endl;
               if (GLOBAL.SyntaxOnly()) parseError = true;
               continue;
            }
            ms.erase(pos);
         }
         if (!msgserv.Find(ms))
         {
            SYSLOG << Severity(Syslog::error)
                   << configfile
                   << "("
                   << lineNum
                   << "): Cannot find message server:"
                   << line
                   << endl;
            if (GLOBAL.SyntaxOnly()) parseError = true;
            continue;
         }
         for (int i=2; i < tokens.size(); i++)
         {
            string streamname = tokens[i];
            streamblock *sb = new streamblock;
            if (streamname.substr(0,4) == "udp:")
            {
               sb->mc.SetConnection(UDP);
               streamname.erase(0,4);
            }
            streams.insert(streamitem(streamname, sb));
         }
      }
      
      // Otherwise, we can't recognize what it is.
      
      else
      {
         SYSLOG << Severity(Syslog::error)
                << configfile 
                << "(" 
                << lineNum
                << "): Invalid statement type " 
                << tokens[0] 
                << endl;
         parseError = true;
      }
   }

   if (parseError)
   {
      SYSLOG << Severity(Syslog::error)
             << "Error parsing file, not starting server"
             << endl;
      exit(1);
   }

}

//=====================================================================
// Expand a linegroup statement by iterating ranges.
// Hostnames can have numeric ranges such as host001-099.
// Line numbers are incremented using the starting line number
// as a base.
// ====================================================================

void ConsoleDaemon::ExpandLines(vector<Line *> &linelist)
{
   Line *base;
   base = linelist.front();
   linelist.clear();
   vector<string> names;
   vector<string> ttys;
   
   names = Makenames(base->Name());
   ttys  = Makenames(base->DevName(), names.size());

   if (names.size() > ttys.size())
      throw string("Not enough tty names for line names");
   
   for (unsigned int i=0; i < names.size(); i++)
   {
      Line *newline = new Line(*base, names[i], ttys[i]);
      linelist.push_back(newline);
   }
}

// ====================================================================
// Return the name of the logging directory, replacing escapes.
// ====================================================================

string ConsoleDaemon::LogDir(void)
{
   string tempname = logdir;
   
   return tempname;
}

// ====================================================================
// Return a macro for a character, or the null string.
// ====================================================================

macrodef ConsoleDaemon::Macro(int cc)
{
   return macros[cc];
}

// ====================================================================
// List all macros
// ====================================================================

string ConsoleDaemon::ListMacros(void)
{
   string buffer;
   
   for (int cc = 0; cc < 256; cc++)
   {
      if (! macros[cc].first.empty())
      {
         buffer += cc;
         buffer += ": ";
         buffer += macros[cc].second;
         buffer += "\r\n";
      }
   }

   return buffer;
}


// ====================================================================
// Return the list of unique servers.
// ====================================================================

string ConsoleDaemon::ServerList(void)
{
   string response;
   
   set<string>::iterator pos = servers.begin();
   set<string>::iterator epos = servers.end();
   
   bool first = 1;
   
   for ( ; pos != epos; pos++)
   {
      if (!first) response += ' ';
      response += *pos;
      first = 0;
   }
   
   return response;
}
// --------------------------------------------------------------------
// Short and long usage texts.
// -------------------------------------------------------------------- 

void ConsoleDaemon::UsageShort(void)
{
   cerr << GLOBAL.Progname() 
        << " [-C|-config filename] [-d|-debug] [-v|-verbose] [-V|-version]\n"
        << "         [-h|-help] [-P|-port port] [-D|-nodaemon] [-syntax]\n"
        << "         [-msgserv host:port] [-pid filename]\n"
        << endl;
};

void ConsoleDaemon::UsageLong(void)
{
   UsageShort();
   
   cout
      << "-C | -config <filename>  give a new config file to the server process\n"
      << "-d | -debug              debug -- don't become daemon, and do extra tracing\n"
      << "-D | -nodaemon           don't become daemon\n"
      << "-h | -help               output this message\n"
      << "-P | -port <port>        specify port to listen on\n"
      << "-r | -runas <user>       specify the username to run as\n"
      << "-v | -verbose            be verbose on startup\n"
      << "-V | -version            output version info\n"
      << "     -pid <filename>     store pid of program in named file\n"
      << "     -msgserv host:port  supply name of message service\n"
      << "     -syntax             only syntax check the config file and exit.\n"
      << endl;
};

// ====================================================================
// Send a message to the global message service.
// ====================================================================

void ConsoleDaemon::SendMessage(string streamname, Buffer &record)
{
   streamblock *sb = streams[streamname];

   if (!sb) return;
   
   if (sb->streamOpen && sb->streamOpen.TryLock(MLock::Shared))
   {
      try { sb->mc.Send(record);  }
      catch(File::IOErr(e))
      {
         SYSLOG << Severity(Syslog::error)
                << "Error sending to Message Service "
                << ": " << e.filename 
                << ": " << e.message 
                << endl;
         sb->streamOpen = false;
         sb->restartNeeded = true;
         sb->restartNeeded.Signal();
      }
      
      sb->streamOpen.Unlock(MLock::Shared);
   }
}

// ====================================================================
// Monitor the message service connection and restart it if necessary.
// This is a separate thread so that no other thread blocks on the
// message service if the daemon or the network is down.
// ====================================================================

void *RestartMsgserv(void *data)
{
   streamblock *sb = (streamblock *) data;

   for (;;)
   {
      sb->restartNeeded.Lock();
      
      if (! sb->restartNeeded) sb->restartNeeded.Wait();
      
      if (sb->restartNeeded)
      {
         sb->streamOpen.Lock(MLock::Exclusive);
         sb->mc.ReopenStream();
         sb->streamOpen = true;
         sb->restartNeeded = false;
         SYSLOG << Severity(Syslog::info)
                << "Connection to Message Service re-established "
                << endl;
         sb->streamOpen.Unlock(MLock::Exclusive);
      }
   }

   return data;
}

// ====================================================================
// Thread that blocks waiting for a signal and takes appropriate action.
// ====================================================================

static void *MonitorSignals(void *data)
{
   // Compose the set of signals that we look for.  They must be
   // blocked in the program constructor.
   sigset_t signalset;
   sigemptyset(&signalset);
   sigaddset(&signalset, SIGHUP);
   sigaddset(&signalset, SIGINT);
   sigaddset(&signalset, SIGQUIT);
   sigaddset(&signalset, SIGILL);
   sigaddset(&signalset, SIGSEGV);
   sigaddset(&signalset, SIGTERM);

   int sigint = 0;
   int sigerr = 0;
   int signumber;
   
   for (;;)
   {
      sigwait(&signalset, &signumber);
      
      switch(signumber)
      {
          case SIGHUP:
             SYSLOG << Severity(Syslog::info)
                    << "Received SIGHUP signal" << endl;
             // Ignore for now; maybe cause re-read later.
             break;
             
          case SIGINT:
          case SIGQUIT:
          case SIGTERM:
             sigint++;
             SYSLOG << Severity(Syslog::info)
                    << GLOBAL.Progname()
                    << " stopping on signal "
                    << signumber
                    << endl;
             DISPATCHER.Stop("killed by termination signal");
             sleep(1);
             exit(1);
             break;
             
          case SIGILL:
          case SIGSEGV:
             sigerr++;
             SYSLOG << Severity(Syslog::error)
                    << GLOBAL.Progname()
                    << " stopping on signal "
                    << signumber
                    << endl;
             DISPATCHER.Stop("killed by error signal");
             sleep(1);
             exit(1);
             break;
      }
   }
   
   return data;
}
