//====================================================================
// Global definitions for Console Server
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.9 $ $Date: 2006/08/21 23:59:13 $ $Author: boeheim $
// ******************************************************************** //

#ifndef consoled_h
#define consoled_h

#include <map>
#include <pthread.h>
#include <set>
#include <string>
#include <vector>

#include "acl.h"
#include "defaults.h"
#include "dispatcher.h"
#include "host.h"
#include "mlock.h"
#include "msgclient.h"

class Line;

// Macro definition, comment.
typedef pair<string, string> macrodef;

// Stream handling definitions.
typedef struct
{
   MessageClient mc;            // Client object
   MLock streamOpen;            // Shared lock while sending, 
                                //   exclusive while stream not open.
                                //   Returns true if open.
   SLock restartNeeded;         // For signalling restart thread.
   pthread_t restartThread;     // ID of restart thread.
}  streamblock;
typedef map<string, streamblock*> streamlist;
typedef pair<string, streamblock*> streamitem;

class ConsoleDaemon
{
private:
   string escape;               // Default escape.
   string configfile;           // Config file name.
   string bindir;               // Directory program started from.
   Host   localport;            // Local address to bind to.
   Host   msgserv;              // Address of message server.
   string pidfile;              // File to write PID into.

   streamlist streams;          // Set of streams to slacmsg.
   pthread_t monitorThread;     // ID of signal monitor thread.

   string logdir;               // Default logging directory
   string defaultSpeed;         // Default speed setting.

   map <string, ACL *> acls;    // Set of named access lists.

   set<string> servers;         // List of all unique servers.

   macrodef macros[256];        // Array of macros.

   SmallBuffer startmsg;        // Message about startup time.

   void UsageShort(void);       // Short and long
   void UsageLong(void);        //   usage messages.

   bool dosetuid;               // Whether to setuid on startup.
   string runas;                // User to run as

   int idleWarn;                // Time until warning of idleness
   int idleForce;               // Time until force disconnect.

   string gssrealm;             // Domain for authentication
   string gssservice;           // Service for authentication.
   string gsskeytab;            // Location of keytab.

   string version;              // Constructed version string
   
public:

   // Constructor to initialize.

   ConsoleDaemon(void);
   ~ConsoleDaemon(void);

   // Option parsing

   void ParseArgs(int, char *[]); // Parse global options.
   
   // Configuration file parsing.

   void FetchConfig(void);
   void ReadConfig(void);
   void ExpandLines(vector<Line *>&);

   // Action routines.

   void Run(int argc, char **argv); // Run service.

   // Global values.

   string ConfigFile(void) { return configfile; };
   string Escape(void)     { return escape; };
   string Pidfile(void)    { return pidfile; };
   string LogDir (void);
   macrodef Macro(int cc); 
   string ListMacros(void);
   int Port(void)         { return localport.Port(); };
   Buffer &StartMsg(void)  { return startmsg; }
   int IdleWarn(void)      { return idleWarn; };
   int IdleForce(void)     { return idleForce; };
   string ServerList(void);
   ACL *FindACL(string name) { return acls[name]; };
   string Version(void) { return version; };
   
   // Message service routines.

   void SendMessage(string streamname, Buffer &record);
};

extern ConsoleDaemon DAEMON;    // Reference to single global object.

extern Dispatcher    DISPATCHER; // Reference to global dispatcher

#endif

