// ====================================================================
// dispatcher.cc - Object to own the console lines.
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.20 $ $Date: 2007/04/16 01:57:53 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include "autoconfig.h"
#include "consoled.h"
#include "dispatcher.h"
#include "file.h"
#include "global.h"
#include "host.h"
#include "mydbm.h"
#include "syslog.h"
#include "utility.h"

#include <algorithm>
#include <iostream>
#include <unistd.h>

// ====================================================================
// Main dispatcher method, which will set up the console lines and
// the threads that service them.  It will then wait for incoming
// client connections and start threads to service them.  When
// it receives a quit command it will return here.
// ====================================================================

   // Small function to start the thread.

   static void *starter(void *data) 
   { ((Dispatcher *) data)->WaitForConnection(); return data; }

void Dispatcher::Run(void)
{
   SmallBuffer statebuf;
   bool changed;
   bool stateopen = false;
   MyDBM statefile(STATEFILE);
   
   // Start the console threads.

   StartConsoles();

   // -----------------------------------------------------------------
   // Set up the socket to listen for connections and start the
   // thread to service those connections.
   // -----------------------------------------------------------------
   
   int status;
   void *result;
   
   status = pthread_create(&listenThread, NULL, starter, this);
   
   if (status != 0)
   {  
      SYSLOG << Severity(Syslog::error) 
             << "Thread error " << status << strerror(status) << endl;
   }
   
   // -----------------------------------------------------------------
   // Main running state now: wait for deaths of client threads or
   // the signal that the dispatcher should be halted.
   // -----------------------------------------------------------------

   halt = false;
   halt.Lock();                 // Lock the condition.
   
   while (! halt)
   {
      time_t tv = time(NULL); // Time until next minute.
      tv = tv + (60 - tv % 60);

      if (halt.TimedWait(tv)) continue;
      else
      {
         // Check log files and drive idle proc.
         LockConsolelist(MLock::Shared);
         consolelist::iterator pos = consoles.begin();
         consolelist::iterator epos = consoles.end();

         time_t now = time(NULL);
         for ( ; pos != epos; pos++)
         {
            Console *console = *pos;
            
            console->CloseLog();
            console->IdleCheck(now);

            statebuf.Clear();
            changed = console->GetState(statebuf);
            if (changed)
            {
               if (statefile.Open())
               {
                  statefile.Store(console->DevName(), statebuf);
               }
            }
         }
         statefile.Close();
         
         UnlockConsolelist(MLock::Shared);

         // Check idle timeouts.
         LockClientlist(MLock::Exclusive);
         clientlist::iterator cl = clients.begin();
         
         while (cl != clients.end())
         {
            if (!(*cl)->IdleCheck()) 
            {
               cl++;
               continue;
            }

            Client *client = (*cl);
            
            client->Shutdown(SHUT_WR);
            client->Close();       // Close open connections.

            if (GLOBAL.Verbose())
               SYSLOG << Severity(Syslog::info)
                      << "Dropped connection from "
                      << client->Userid() << "@" << client->Peername()
                      << endl;

            cl = clients.erase(cl); // Moves cl to next element.
            delete client;
         }
         UnlockClientlist(MLock::Exclusive);
      }
   }
   
   // -----------------------------------------------------------------
   // Halt has been posted, so kill our listener and clients and return.
   // -----------------------------------------------------------------
   
   pthread_cancel(listenThread);
//   status = pthread_join(listenThread, &result);

   listenSocket.Close();

   LockClientlist(MLock::Shared);
   
   for (clientlist::iterator iclient = clients.begin();
        iclient != clients.end(); iclient++)
   {
      (*iclient)->StopReader();
      (*iclient)->Close();
   }

   UnlockClientlist(MLock::Shared);
   LockConsolelist(MLock::Shared);

   for (consolelist::iterator iconsole = consoles.begin();
        iconsole != consoles.end(); iconsole++)
   {
      (*iconsole)->StopReader();
      (*iconsole)->Close();
   }

   UnlockConsolelist(MLock::Shared);
}

// ====================================================================
// Method to gracefully halt the dispatcher.
// ====================================================================

void Dispatcher::Stop(string reason)
{
   Buffer message;
   
   message.TimeStampFormat("%T ");
   message = GLOBAL.Progname();
   message += " shutting down: ";
   message += reason;
   message += "\r\n";
   message.TimeStamp();
   
   LockClientlist(MLock::Shared);
   for (clientlist::iterator iclient = clients.begin();
        iclient != clients.end(); iclient++)
   {
      try { (*iclient)->Send(message); }
      catch (...) { } // Ignore errors.
   }
   UnlockClientlist(MLock::Shared);
   
   halt = true;
   halt.Signal();
}

// ====================================================================
// Method to run in a thread to wait for connections.
// ====================================================================

void Dispatcher::WaitForConnection(void)
{
   Client *newclient;
      
   SYSLOG << Severity(Syslog::info)
          << GLOBAL.Progname()
          << " listening on port "
          << DAEMON.Port()
          <<endl;
   
   // Establish the listening socket.

   try
   {
      Host myport;
      myport.SetPort(DAEMON.Port());
      listenSocket.Bind(myport);
      listenSocket.Listen();
   }
   catch( File::IOErr e )
   {
      SYSLOG << Severity(Syslog::error)
             << e.routine 
             << ": " << e.filename 
             << ": " << e.message 
             << endl;

      abort();
   }

   // Loop waiting for connections.

   for (;;)
   {
      newclient = new Client;
      
      // In following code, if newclient is non-null, it should be
      // freed at the end of the loop.  This allows exceptions to jump out of
      // the try block and the memory will still get cleaned up.  Care is taken
      // to put the new client on the linked list and zero this reference to it
      // so that a good client is preserved.

      try
      {
         // Accept the incoming connection.

         listenSocket.Accept(newclient);
      
         // Add them to the list and start the reader thread.
         
#ifdef HAVE_LIBSSL
         newclient->CheckCert();

         Buffer sslmsg;
            
         newclient->GetSSLCipher(sslmsg);
#endif
            
         if (GLOBAL.Verbose())
            SYSLOG << Severity(Syslog::info)
                   << "Connection from "
                   << newclient->Peername()
#ifdef HAVE_LIBSSL
                   << " using "
                   << sslmsg.Cstr()
#endif
                   << endl;
         
         // Lock list not only around the list update, but until thread is started.
         // We don't want the once-per-minute process removing it because it 
         // doesn't say 'running' yet.
         
         LockClientlist(MLock::Exclusive);
         
         clients.push_back(newclient);
         
         newclient->StartReader();
         newclient->Detach();
         
         UnlockClientlist(MLock::Exclusive);
         
         newclient = 0;
      }
      catch (File::IOErr e)
      {
         SYSLOG << Severity(Syslog::error)
                << e.routine 
                << ": " << e.filename 
                << ": " << e.message 
                << endl;
      }
      
      if (newclient) delete newclient;
      newclient = 0;
   }
}

// ====================================================================
// Run the console list and open and startup any consoles that are
// not yet started.
// ====================================================================

void Dispatcher::StartConsoles(void)
{
   MyDBM statefile(STATEFILE);
   SmallBuffer staterec;
   
   statefile.Open();
   
   LockConsolelist(MLock::Shared);

   consolelist::iterator pos = consoles.begin();
   
   for ( ; pos != consoles.end(); pos++)
   {
      Console *console = *pos;
      
      if (console->Started()) continue;

      if (statefile.Open()) 
      {
         if (statefile.Fetch(console->DevName(), staterec))
            console->SetState(staterec);
      }
      
      console->Open();
      if (console->IsUp()) 
      {
         console->StartReader();
         console->Detach();
      }
   }

   UnlockConsolelist(MLock::Shared);

#ifdef ANY_DBM
   if (statefile) MYDBM_CLOSE(statefile);
#endif
}

// ====================================================================
// Register a console by adding a new console to the list, or updating
// if it already exists.  Local consoles are updated by device name,
// so that they keep open connections and logs if necessary; Remote
// consoles are updated by name.
// ====================================================================

void Dispatcher::RegisterConsole(Line *lp)
{
   if (IsLocalHost(lp->Server().c_str()))
   {
      Console *console;

      if ((console = FindConsole(lp->DevName())))
      {
         *console = *lp;
      }
      else
      {
         console = new Console(lp);
         
         consoles.push_back(console);
      }
   }
   else
   {
      BConsole *rconsole;

      if ((rconsole = FindRConsole(lp->Name())))
      {
         *rconsole = *lp;
      }
      else
      {
         rconsole = new BConsole(lp);
         
         rconsoles.push_back(rconsole);
      }
   }
}

// ====================================================================
// Search the console list for a console matching the specified name.
// ====================================================================

Console *Dispatcher::FindConsole(string search)
{
   Console *console = NULL;
   bool searchdev = (search[0] == '/');
   
   LockConsolelist(MLock::Shared);
   consolelist::iterator pos = consoles.begin();

   for ( ; pos != consoles.end(); pos++)
   {
      if (searchdev)
      {
         if ((*pos)->DevName() == search) { console = *pos;  break;  }
      }
      else
      {
         if ((*pos)->Name() == search) { console = *pos;  break;  }
      }
   }
   UnlockConsolelist(MLock::Shared);

   return console;
}

BConsole *Dispatcher::FindRConsole(string search)
{
   BConsole *rconsole = NULL;

   rconsolelist::iterator pos = rconsoles.begin();

   for( ; pos != rconsoles.end(); pos++)
   {
     if ((*pos)->Name() == search)
     {
       rconsole = *pos;
       break;
     }
   }

   return rconsole;
}

// ====================================================================
// Check lists for consistancy
// ====================================================================

void Dispatcher::Check(void)
{
   consolelist::iterator console;
   clientlist::iterator client;
   Console *consp;

   LockConsolelist(MLock::Shared);
   
   for (console = consoles.begin(); console != consoles.end(); console++)
   {
      for (client = (*console)->FirstClient(); client != (*console)->LastClient(); client++)
      {
         if ((*client)->Cons() != (*console))
         {
            SYSLOG << Severity(Syslog::error)
                   << "Internal error: console "
                   << (*console)->Name()
                   << " points to client "
                   << (*client)->UserAtHost()
                   << " but client points to ";
            
            if ((*client)->Cons()) SYSLOG << (*client)->Cons()->Name() << endl;
            else SYSLOG << "NULL" << endl;
         }
      }
   }

   UnlockConsolelist(MLock::Shared);
   LockClientlist(MLock::Shared);

   for (client = clients.begin(); client != clients.end(); client++)
   {
      bool found = false;
      list<Client *>::iterator cl;
      
      consp = (*client)->Cons();
      
      if (!consp) continue;
      
      for (cl = consp->FirstClient(); cl != consp->LastClient(); cl++)
      {
         if ((*cl) == (*client)) { found = true; break; }
      }

      if (!found)
      {
         SYSLOG << Severity(Syslog::error)
                << "Internal error: client "
                << (*client)->UserAtHost()
                << " points to console "
                << consp->Name()
                << " but client is not in console's list"
                << endl;
      }
   }

   UnlockClientlist(MLock::Shared);
}
