// ====================================================================
// Class for encapsulating files and their actions.
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.5 $ $Date: 2005/02/09 00:48:12 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include <fcntl.h>
#include <iostream>
#include <unistd.h>
#include <errno.h>

#include "file.h"
#include "buffer.h"

extern int errno;

// ====================================================================
// Constructor
// ====================================================================

File::File(int f) : fd(f) 
{ 
   holding = new Buffer(256);
}

// ====================================================================
// Copy constructor
// ====================================================================

File::File(const File &f)
{ 
   fd       = f.fd;
   filename = f.filename;
   holding  = new Buffer(256);
}

// ====================================================================
// Assignment operator.
// ====================================================================

File &File::operator=(const File &f)
{ 
   fd       = f.fd;
   filename = f.filename;
   holding  = new Buffer(256);
   return *this;
}

// ====================================================================
// Destructor
// ====================================================================

File::~File(void)
{
   delete holding;
}

// ====================================================================
// Open a file and save the file descriptor.
// ====================================================================

void File::Open(const string fn, int flags, int mode)
{
   int value;

   filename = fn;
   
   value = ::open(filename.c_str(), flags, mode);

   if (value == -1) throw IOErr("open", errno, filename);

   fd = value;
}

// ====================================================================
// Dup another file descriptor to this one.
// ====================================================================

void File::Dup(File &fp)
{
   filename = fp.FileName();
   
   fd = ::dup(fp.FD());
   
   if (fd < 0) throw IOErr("dup", errno, filename);
}

// ====================================================================
// Close a file and mark the file descriptor invalid.
// ====================================================================

void File::Close(void)
{
   int error;

   if (fd == -1) return;        // No operation on already closed file.
   
   error = ::close(fd);

   if (error == -1) throw IOErr("close", errno, filename);

   fd = -1;
}

// ====================================================================
// Set non-blocking mode on a file.
// ====================================================================

void File::NonBlocking(void)
{
   int error;
   int flags;

   flags = ::fcntl(fd, F_GETFL);
   
   flags |= O_NONBLOCK;
   
   error = ::fcntl(fd, F_SETFL, flags);

   if (error == -1) throw IOErr("fcntl", errno, filename);
}

// ====================================================================
// Set blocking mode on a file.
// ====================================================================

void File::Blocking(void)
{
   int error;
   int flags;
   
   flags = ::fcntl(fd, F_GETFL);
   
   flags &= ~O_NONBLOCK;
   
   error = ::fcntl(fd, F_SETFL, flags);

   if (error == -1) throw IOErr("fcntl", errno, filename);
}

// ====================================================================
// Set/Query buffering mode
// ====================================================================

int File::BufferSize(void)
{
   return holding->Capacity();
}

int File::BufferSize(int newsize)
{
   int prev = holding->Capacity();

   holding->Resize(newsize);
   
   return prev;
}

// ====================================================================
// Write the contents of a buffer to a file.  If non-blocking I/O
// returns EAGAIN, stick the contents into a buffer for later.
// ====================================================================

void File::Write(const Buffer &buffer)
{
   Write((char *)buffer.RPos(), buffer.Size());
}

void File::Write(const string &buffer, int len)
{
   Write(buffer.c_str(), len);
}

void File::Write(const char cc)
{
   Write(&cc, 1);
}

void File::Write(const char *buffer, int len)
{
   int error = 0;

   // Default buffer length is string length.

   if (len == 0) len = strlen(buffer);

   // If previously held output, try to write now.

   if (holding->Dirty())
   {
      error = ::write(fd, holding->RPos(), holding->Size());
      if (error != -1) holding->Clear();
   }

   // If buffer clear, try to write
   
   if (error != -1)
      error = ::write(fd, buffer, len);

   // If EAGAIN, save output in buffer.  Handle other errors.

   if (error == -1) 
   {
      if (errno == EAGAIN)
      {
         holding->Append(buffer, len);
         return;
      }
     
      throw IOErr("write", errno, filename);
   }
}

// ====================================================================
// Read from a file into a buffer.
// ====================================================================

int File::Read(char *buffer, int len)
{
   len = ::read(fd, buffer, len);

   if (len == -1)  throw IOErr("read", errno, filename);

   return len;
}

int File::Read(Buffer &buffer)
{
   int len = ::read(fd, buffer.WPos(), buffer.Space());

   if (len == -1)  throw IOErr("read", errno, filename);

   buffer.WSeek(Buffer::Current, len);

   return len;
}

// ====================================================================
// Perform ioctl operations on the file.
// ====================================================================

int File::IOCtl(int flag, void *data)
{
   int value;
   
   value = ::ioctl(fd, flag, data);

   if (value == -1) throw IOErr("ioctl", errno, filename);

   return value;
}

// ====================================================================
// Perform fstat operation on a open file.
// ====================================================================

void File::Stat(struct stat &statbuf)
{
   int error;

   error = ::fstat(fd, &statbuf);

   if (error == -1) throw IOErr("stat", errno, filename);
}

// ====================================================================
// Change ownership of an open file.
// ====================================================================

void File::Chown(int uid, int gid)
{
   int error;

   error = ::fchown(fd, uid, gid);

   if (error == -1) throw IOErr("chown", errno, filename);
}

// ====================================================================
// Change modes of an open file.
// ====================================================================

void File::Chmod(int mode)
{
   int error;

   error = ::fchmod(fd, mode);

   if (error == -1) throw IOErr("chmod", errno, filename);
}
