// ====================================================================
// gssapi.cc -- A class for gssapi authentication
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.6 $ $Date: 2006/12/12 07:46:20 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include "file.h"
#include "global.h"
#include "gssapi.h"
#include "socket.h"
#include "syslog.h"

#ifdef HAVE_GSSAPI_GSSAPI_KRB5_H
# include <gssapi/gssapi_krb5.h>
#endif
#ifdef HAVE_GSSAPI_GSSAPI_EXT_H
# include <gssapi/gssapi_ext.h>
#endif
#ifndef HAVE_GSS_C_NT_USER_NAME
# define GSS_C_NT_USER_NAME gss_nt_user_name
#endif

#include <iostream>

#define MAXMSG 2048

gss_cred_id_t GSSAPI::gssdefaultcred = NULL;

// ====================================================================
// GSSAPI Constructor. 
// ====================================================================

GSSAPI::GSSAPI(Socket *s)                // Constructor
{
   socket           = s;
   gsscontext       = GSS_C_NO_CONTEXT;
   gsscredential    = gssdefaultcred;
   
   majorStatus      = 0;
   minorStatus      = 0;
   *gsserror        = '\0';
}

// ====================================================================
// Delete context;
// ====================================================================

GSSAPI::~GSSAPI(void)               // Destructor
{
   gss_release_name(&minorStatus, &gssservice);
   gss_delete_sec_context(&minorStatus, &gsscontext, GSS_C_NO_BUFFER);
}

// ====================================================================
// Given the service name of the remote server, establish a GSSAPI 
// security context.
// ====================================================================

bool GSSAPI::Connect(void)
{
   gss_buffer_desc input = {0,NULL};    // Input token to functions.
   gss_buffer_desc response = {0,NULL}; // Output token from function.
   gss_buffer_desc *bptr = GSS_C_NO_BUFFER; 
   OM_uint32       ignoreStatus;
   bool            success = true;
#if !HAVE_DECL_GSS_KRB5_MECHANISM
#  ifdef HAVE_GSS_MECH_KRB5
#    define GSS_KRB5_MECHANISM gss_mech_krb5
#    undef HAVE_DECL_GSS_KRB5_MECHANISM
#    define HAVE_DECL_GSS_KRB5_MECHANISM 1
#  endif
#endif
#if ! HAVE_DECL_GSS_KRB5_MECHANISM
   gss_OID    k5OID;
   const char *mech = "kerberos_v5";
   __gss_mech_to_oid(mech, &k5OID);
#  define GSS_KRB5_MECHANISM k5OID
#endif

   // Perform the context-establishment loop.

   majorStatus = GSS_S_CONTINUE_NEEDED; // Get the loop started.

   while (majorStatus == GSS_S_CONTINUE_NEEDED)
   {
      // Prepare the token that we will send to the remote server.
      majorStatus = gss_init_sec_context(&minorStatus, 
                                         GSS_C_NO_CREDENTIAL, 
                                         &gsscontext, 
                                         gssservice, 
                                         (const gss_OID) GSS_KRB5_MECHANISM,
                                         GSS_C_MUTUAL_FLAG | GSS_C_REPLAY_FLAG, 
                                         0, 
                                         NULL,     // no channel bindings 
                                         bptr,     // previous response from server. 
                                         NULL,     // ignore mech type
                                         &input,   // to be sent to remote server
                                         &gssflags, 
                                         NULL);    // ignore time_rec 
      
      gss_release_buffer(&ignoreStatus, &response);  // Done with this.
      
      if (NotOK("Initializing context"))
      {
         string errmsg = "Client aborting handshake: ";
         errmsg += gsserror;
         gss_buffer_desc errbuf = { errmsg.size()+1, (void *) errmsg.c_str() };
         try 
         {
            SendPacket(DATA, &errbuf);
         }
         catch (string s) {  }
         
         success = false;
         break;
      }
      
      // If there is anything to send to the server, send it.
      if (input.length != 0) 
      {
         if (input.length > MAXMSG)
         {
            SYSLOG << Severity(Syslog::error)
                   << "Security token size "       << (int) input.length
                   << " greater than maximum "     << (int) MAXMSG
                   << endl;
            success = false;
            break;
         }
         if (GLOBAL.Debug()) 
            SYSLOG << Severity(Syslog::debug) 
                   << "GSSAPI: sending init_sec_context token, size=" << (int) input.length
                   << endl;
         try 
         {
            SendPacket(HANDSHAKE, &input);
         }
         catch(string s)
         {
            SYSLOG << Severity(Syslog::error)
                   << "GSSAPI error: "  << s
                   << " sending token to server "
                   << endl;
            success = false;
            break;
         }
      }
      
      gss_release_buffer(&ignoreStatus, &input);
      
      if (majorStatus == GSS_S_CONTINUE_NEEDED) 
      {
         if (GLOBAL.Debug())
            SYSLOG << Severity(Syslog::debug) << "GSSAPI: continue needed" << endl;
         try
         {
            char type = ReceivePacket(&response);
            if (type == DATA) 
            {
               int size = (response.length > sizeof(gsserror)) ? sizeof(gsserror) 
                                                               : response.length;
               memcpy(gsserror,response.value,size);
               SYSLOG << Severity(Syslog::warning) << 
                  "GSSAPI: server returned error indicator: " << gsserror << endl;
               success = false;
               break;
            }
         }
         catch (string s)
         {
            SYSLOG << Severity(Syslog::error)
                   << "GSSAPI error: "       << s
                   << " receiving token from server "
                   << endl;
            success = false;
            break;
         }
         
         if (GLOBAL.Debug())
            SYSLOG << Severity(Syslog::debug) 
                   << "GSSAPI: received token, size=" << (int) response.length << endl;

         bptr = &response;
      }
   }
   
   gss_release_buffer(&ignoreStatus, &input);
   gss_release_buffer(&ignoreStatus, &response);
//   if (k5OID != GSS_C_NULL_OID)
//      (void) gss_release_oid(&ignoreStatus, &k5OID);

   return success;
}

// ====================================================================
// Sets the service name for the client to use.
// ====================================================================

bool GSSAPI::SetService(const char *service_name)
{
   gss_buffer_desc input;

   service = service_name;

   // Import the service name into gssservice.
   input.value  = (void *) service_name;
   input.length = strlen(service_name) + 1;
   majorStatus = gss_import_name(&minorStatus, 
                                 &input, 
                                 GSS_C_NT_USER_NAME,
                                 &gssservice);
   if (NotOK("Parsing name"))  return false;
   return true;
}

// ====================================================================
// Given a service name, imports it and acquires credentials for it.
// Returns true on success and false on failure, logging an error message.
// ====================================================================

bool GSSAPI::AcquireServerCreds(const char *service_name)
{
   if (! SetService(service_name)) return false;

   // Acquire credentials for that name.
   majorStatus = gss_acquire_cred(&minorStatus, gssservice, 0,
                                  GSS_C_NULL_OID_SET, GSS_C_ACCEPT,
                                  &gsscredential, NULL, NULL);
   if (NotOK("Acquiring credentials")) return false;
   
   if (GLOBAL.Verbose())
      SYSLOG << Severity(Syslog::debug) 
             << "GSSAPI: Acquired credentials for "
             << service_name
             << endl;
   
   return true;
}

// ====================================================================
// Given a service name, gets the default credentials for new objects
// ====================================================================

bool GSSAPI::DefaultCreds(const char *service_name)
{
   Socket *s = NULL;
   GSSAPI g(s);
   if (! g.AcquireServerCreds(service_name)) return false;

   gssdefaultcred = g.Credential();
   
   return true;
}

// ====================================================================
// Establish a GSS-API context as a specified service with an incoming
// client, sets the context handle, associated client name, and
// associated connection flags.  Returns true on success and false on
// failure, logging an error message.
//
//  Any valid client request is accepted.  If a context is
//  established, its handle is returned in context, the type of the
//  conncetion established is described in the ret_flags, and the
//  client name is returned in client_name.
// ====================================================================

bool GSSAPI::Accept(void)
{
   gss_buffer_desc input    = {0,NULL};
   gss_buffer_desc response = {0,NULL};
   gss_name_t      gssclient;
   OM_uint32       ignoreStatus;
   bool            success = true;
   
   majorStatus = GSS_S_CONTINUE_NEEDED;
   
   while (majorStatus == GSS_S_CONTINUE_NEEDED)
   {
      try
      {
         char type = ReceivePacket(&input);
         if (type != HANDSHAKE) 
         {
            SYSLOG << Severity(Syslog::warning) << 
               "GSSAPI: " << (char *) input.value << endl;
            success = false;
            break;
         }
      }
      catch (string s)
      {
         SYSLOG << Severity(Syslog::error)
                << "GSSAPI error: "       << s
                << " receiving token from client "
                << endl;
         success = false;
         break;
      }
      if (input.length == 0)
      {
         SYSLOG << Severity(Syslog::warning)
                << "GSSAPI: Client closed connection" << endl;
         success = false;
         break;
      }
      
      if (GLOBAL.Debug())
         SYSLOG << Severity(Syslog::debug) 
                << "GSSAPI: received token, size=" << (int) input.length << endl;
      PrintToken(&input);
      majorStatus = gss_accept_sec_context(&minorStatus, 
                                           &gsscontext, 
                                           gsscredential, 
                                           &input,    // Input from client
                                           GSS_C_NO_CHANNEL_BINDINGS, 
                                           &gssclient,// Principle of client 
                                           NULL,      // No preferred mechanism
                                           &response, // Response to client
                                           &gssflags, // Connection settings 
                                           NULL,      // ignore time_rec 
                                           NULL);     // ignore del_cred_handle 

      gss_release_buffer(&ignoreStatus, &input);      // Done with this now.
         
      if (NotOK("Accepting context")) { success = false; break; }
      
      if (response.length != 0) 
      {
         if (GLOBAL.Debug())
            SYSLOG << Severity(Syslog::debug) 
                   << "GSSAPI: sending accept_sec_context token, size="
                   << (int) response.length
                   << endl;
         PrintToken(&response);
         try
         {
            SendPacket(HANDSHAKE, &response);
         }
         catch (string s)
         {
            SYSLOG << Severity(Syslog::error)
                   << "GSSAPI error: "       << s
                   << " receiving token from client "
                   << endl;
            success = false;
            break;
         }
         gss_release_buffer(&ignoreStatus, &response);
      }

      if (majorStatus == GSS_S_CONTINUE_NEEDED && GLOBAL.Debug())
         SYSLOG << Severity(Syslog::debug) << "GSSAPI: continue needed" << endl;
   }

   if (!success)
   {
      if (strlen(gsserror) > 0)
      {
         gss_delete_sec_context(&ignoreStatus, &gsscontext, GSS_C_NO_BUFFER);
         gss_buffer_desc buffer = {strlen(gsserror), gsserror};
         SendPacket(DATA, &buffer);
      }
   }
   else
   {
      DisplayCtxFlags(gssflags);
      gss_buffer_desc clientbuf = {0,NULL};
         
      majorStatus = gss_display_name(&minorStatus, gssclient, &clientbuf, NULL);
      if (NotOK("displaying name")) return false;
      clientname = (char *) clientbuf.value;
      if (GLOBAL.Verbose())
      {
         SYSLOG << Severity(Syslog::debug)
                << "GSSAPI: Authenticated client name "
                << clientname
                << endl;
      }
   }
   
   gss_release_name(&ignoreStatus, &gssclient);
   gss_release_buffer(&ignoreStatus, &input);
   gss_release_buffer(&ignoreStatus, &response);

   return success;
}

// ====================================================================
// Check for one of the normal status.  Return false if all ok, 
// true and a diagnostic if not ok.
// ====================================================================

bool GSSAPI::NotOK(const char *msg)
{
   if (majorStatus == GSS_S_COMPLETE || majorStatus == GSS_S_CONTINUE_NEEDED) return false;

   if (majorStatus != GSS_S_FAILURE)  // Info is in the mech code only.
      LogMsg(msg, majorStatus, GSS_C_GSS_CODE);
   if (minorStatus != 0)
      LogMsg(msg, minorStatus, GSS_C_MECH_CODE);
   return true;
}

// ====================================================================
// Log a GSS message with warning status
// ====================================================================

void GSSAPI::LogMsg(const char *m, OM_uint32 code, int type)
{
   OM_uint32       s1, s2;
   OM_uint32       msg_ctx = 0;
   gss_buffer_desc msg;
   
   do 
   {
      s1 = gss_display_status(&s2, code, type, GSS_C_NULL_OID, &msg_ctx, &msg);
      SYSLOG << Severity(Syslog::warning)
             << "GSSAPI error " 
             << m 
             << ": " 
             << (char *) msg.value 
             << endl;

      int size = sizeof(gsserror) - 1;
      if ((int) msg.length < size) size = msg.length;
      memcpy(gsserror, msg.value, size);
      gsserror[size] = '\0';
      gss_release_buffer(&s2, &msg);
   }  while (msg_ctx != 0);
}

// ====================================================================
// Write a GSSAPI context flag to the log in a human-readable form,
// preceeded by "context flag: ".
// ====================================================================

void GSSAPI::DisplayCtxFlags(OM_uint32 flags)
{
   if (! GLOBAL.Debug()) return;
   
   if (flags & GSS_C_DELEG_FLAG)
      SYSLOG << Severity(Syslog::debug) 
             << "GSSAPI context flag: GSS_C_DELEG_FLAG" << endl;
   if (flags & GSS_C_MUTUAL_FLAG)
      SYSLOG << Severity(Syslog::debug) 
             << "GSSAPI context flag: GSS_C_MUTUAL_FLAG" << endl;
   if (flags & GSS_C_REPLAY_FLAG)
      SYSLOG << Severity(Syslog::debug) 
             << "GSSAPI context flag: GSS_C_REPLAY_FLAG" << endl;
   if (flags & GSS_C_SEQUENCE_FLAG)
      SYSLOG << Severity(Syslog::debug) 
             << "GSSAPI context flag: GSS_C_SEQUENCE_FLAG" << endl;
   if (flags & GSS_C_CONF_FLAG)
      SYSLOG << Severity(Syslog::debug) 
             << "GSSAPI context flag: GSS_C_CONF_FLAG" << endl;
   if (flags & GSS_C_INTEG_FLAG)
      SYSLOG << Severity(Syslog::debug) 
             << "GSSAPI context flag: GSS_C_INTEG_FLAG" << endl;
}


// ====================================================================
//  Print the contents of a raw token in hex, byte by byte, to the log.
// ====================================================================

void GSSAPI::PrintToken(gss_buffer_t tok)
{
   unsigned char *p = (unsigned char *) tok->value;
   char *buf;
   size_t length, i, offset;
   return;
   if (! GLOBAL.Debug()) return;

   length = tok->length * 3 + 1;
   buf = (char *) malloc(length);
   for (offset = 0, i = 0; i < tok->length; i++, p++)
      offset += snprintf(buf + offset, length - offset, "%02x ", *p);
   SYSLOG << Severity(Syslog::debug)
          << "Token: "
          << buf
          << endl;
   free(buf);
}

// ====================================================================
// Send a packet to the other end.
// ====================================================================

typedef struct
{
   OM_uint32 length;
   char      type;
}  HEADER;

void GSSAPI::SendPacket(char type, gss_buffer_t buffer)
{
   HEADER header;
   
   header.length = htonl(buffer->length);
   header.type = type;
   
   try
   {
      socket->Send(&header, sizeof(header));
      socket->Send(buffer->value, buffer->length);
   }
   catch (File::IOErr e)
   {
      throw e.message + " sending packet";
   }
   
}

char GSSAPI::ReceivePacket(gss_buffer_t buffer)
{
   HEADER header;
   int hdramt, datamt;
   
   try
   {
      hdramt = socket->Receive(&header, sizeof(header));
   }
   catch (File::IOErr e)
   {
      throw e.message + " receiving packet";
   }

   if (hdramt == 0) throw string("partner closed connection");
   if (hdramt != sizeof(header))
      throw string("protocol error with partner");
   if (header.type != HANDSHAKE && header.type != DATA)
      throw string("protocol packet type error with partner");
   datamt = ntohl(header.length);
   if (datamt <= 0 || datamt > MAXMSG) 
      throw string("protocol length error with partner");
   buffer->value = malloc(datamt);
   buffer->length = socket->Receive(buffer->value, datamt);
   if (buffer->length == 0) throw string("partner closed connection");
   
   return header.type;
}

   
