// ====================================================================
// heartbeat.cc - listen for a heartbeat from a system.
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.8 $ $Date: 2007/04/02 16:06:58 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include "console.h"
#include "consoled.h"
#include "heartbeat.h"
#include "syslog.h"
#include "utility.h"

#include <iostream>

// ====================================================================
// Constructor: start in state where no heartbeat detected.
// ====================================================================

HeartBeat::HeartBeat(Console *cons)
{
   console  = cons;
   state    = HB_NONE;
   lastseq  = -1;
   lasttime = 0;
   changed  = false;

   logmsg.TimeStampFormat(console->TSFormat() + console->Name() + " Heartbeat ");
}

// ====================================================================
// Parse a heartbeat signal from our line.
// ====================================================================

void HeartBeat::Input(Buffer &buffer)
{
   int    newseq;
   string newseqstr;
   string newhost;

   logmsg.Clear();

   unsigned char *p = buffer.RPos();
   unsigned char *e = buffer.WPos();
   
   p += 3;
   
   while (p < e && *p != ':') newhost += *p++;
   
   p++;
   if (*p == '-') newseq = -1;
   else 
   {
      while (p < e && *p >= '0' && *p <= '9') newseqstr += *p++;
      if (newseqstr == "") newseqstr = "1";
      try 
      {
         newseq = StrToInteger(newseqstr);
      }
      catch (string s)
      {
         SYSLOG << Severity(Syslog::error)
                << "Invalid heartbeat number in Heartbeat::Input: "
                << s
                << endl;
         newseq = 1;
      }
   }

   if (newseq == -1)
   {
      logmsg += "Shutdown " + newhost;
      state = HB_SHUTDOWN;
      changed = true;
   }
   else
   {
      switch (state)
      {
          case HB_NONE:
             // If sequence number is non-zero, it's likely that the console server
             // just restarted and is getting caught up.  If it is zero, it's
             // a server just starting up.  To keep log noise down, only log
             // the second case.
             if (newseq == 0) logmsg += "Started " + newhost;
             state = HB_OK;
             changed = true;
             break;

          case HB_SHUTDOWN:
             logmsg += "Started " + newhost;
             state = HB_OK;
             changed = true;
             break;

          case HB_STOPPED:
             logmsg += "Resumed " + newhost + " ";
             state = HB_OK;
             changed = true;
             // fall through

          case HB_OK:
             if (newseq > lastseq + 1) 
             {
                logmsg += "Missed ";
                logmsg += newseq - lastseq - 1;
                logmsg += " heartbeats ";
             }
             
             if (newhost != host)
             {
                logmsg += "Host changed from " + host 
                       +  " to " + newhost;
                changed = true;
             }
             break;
      }
   }

   if (logmsg.Dirty()) 
   {
      logmsg += "\n";
      logmsg.TimeStamp();
      DAEMON.SendMessage("heartbeat", logmsg);
//      logmsg.RSeek(Buffer::Beginning, 11);
      console->WriteLog(logmsg);
   }

   host     = newhost;
   lastseq  = newseq;
   lasttime = time(NULL);
}

// ====================================================================
// Check state of heartbeat on idle calls.
// ====================================================================

void HeartBeat::IdleCheck(time_t now)
{
   switch (state)
   {
       case HB_NONE:
       case HB_STOPPED:
       case HB_SHUTDOWN:
          break;

       case HB_OK:
          if (now - lasttime > 90)
          {
             logmsg.Clear();
             logmsg += "Stopped " + host + "\n";
             logmsg.TimeStamp();
             DAEMON.SendMessage("heartbeat", logmsg);
             console->WriteLog(logmsg);
             state = HB_STOPPED;
             changed = true;
          }
          break;
   }
   
}

// ====================================================================
// Display heartbeat status
// ====================================================================

void HeartBeat::Display(Buffer &buffer)
{
   switch (state)
   {
       case HB_NONE:
          buffer += "No heartbeat ";
          break;
          
       case HB_OK:
          buffer += "Heartbeat OK " + host;
          break;
          
       case HB_STOPPED:
       case HB_SHUTDOWN:
          buffer += "Heartbeat ";
          buffer += (state == HB_STOPPED) ? "Stopped  " : "Shutdown ";

          struct tm tod;
          
          unsigned char *p = buffer.WPos();
          
          buffer += "dd mon yyyy hh:mm:ss";
          
          localtime_r(&lasttime, &tod);
          
          strftime((char *)p, buffer.WPos() - p + 1, "%e %b %Y %T", &tod);
   }
}

// ====================================================================
// Save the state in a form suitable for persistant storage.
// Return true if state has changed since last Get or Set.
// ====================================================================

bool HeartBeat::GetState(Buffer &statebuf)
{
   bool retchanged = changed;
   changed = false;
   
   statebuf += "HB";
   statebuf += '\0';

   statebuf += (state == HB_OK) ? 'U' : 'D';
   statebuf += '\0';
   
   statebuf += host;
   statebuf += '\0';

   statebuf += lastseq;
   statebuf += '\0';
   
   statebuf += (int) lasttime;
   statebuf += '\0';
   
   return(retchanged);
}

// ====================================================================
// Set the state from a previously set record.  If the previous state
// was HB_OK, set the state to HB_NONE until it is re-acquired.  This
// will prevent noise as if the host was down.  If the previous state
// was HB_STOPPED, record that and the last time we heard from it.
// ====================================================================

void HeartBeat::SetState(Buffer &statebuf)
{
   changed = false;
   
   char *p = (char *) statebuf.RPos();
   
   p += strlen(p) + 1; // Skip HB label string.

   state = (*p == 'U') ? HB_NONE : HB_STOPPED;
   p += strlen(p) + 1;
   
   host = p;
   p += strlen(p) + 1;
   
   string temp = p;
   p += strlen(p) + 1;
   
   int oldseq;

   try 
   {
      oldseq = StrToInteger(temp);
   }
   catch (string s)
   {
      oldseq = 0;
      SYSLOG << Severity(Syslog::error)
             << "Invalid heartbeat number in Heartbeat::SetState: "
             << s
             << endl;
   }
   
   temp = p;
   
   int oldtime = StrToInteger(temp);

   if (state == HB_STOPPED)
   {
      lastseq  = oldseq;
      lasttime = oldtime;
   }
}
