// ====================================================================
// Class for host name/address lookups
// Use re-entrant functions for thread safety when available.  
// These routines generally return a pointer to static data, so
// all data must be copied to a safe place under a lock so that
// nothing else can wipe it out.  All threads in the program MUST
// use this interface to preserve thread safety.
// gethostbyname_r is very non-standard so use a global lock instead.
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.9 $ $Date: 2007/04/16 01:58:32 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include "autoconfig.h"
#include "host.h"
#include "psstream.h"
#include "slock.h"
#include "utility.h"

#include <string.h>
#include <iostream>
#include <sys/socket.h>
#include <unistd.h>

extern int h_errno;             // Errno defined by gethostbyxxxx

static SLock hostEntLock;       // Global lock

bool   Host::initialized = false;

string Host::thishostname;
string Host::thisdomainname;
string Host::thisfqname;
string Host::thisipaddr;

// ====================================================================
// Constructor:  First time only initialize the static information.
// ====================================================================

Host::Host(void)
{
   sa.sin_family = AF_INET; 
   sa.sin_addr.s_addr = INADDR_ANY;
   sa.sin_port = 0;
   addrset = false; 
}   

// ====================================================================
// Initialization of the global class data: hostname and ip address
// of the local host.  Function that fetch this data call this method
// to make sure that it's set up.  This is not needed for the 
// other address structure manipulation.
// ====================================================================

void Host::initialize(void)
{
   hostEntLock.Lock();
   
   if (! initialized)
   {
      initialized = true;

      // Get this host's name and domain, if available.
      // gethostname() may only return the unqualified host name on
      // some systems.  If so, we'll have another chance in a minute.

      char buf[256];
   
      gethostname(buf, sizeof(buf));
      Host::thishostname = buf;
      Lower(thishostname);
      string::size_type pos = thishostname.find_first_of(".");
      if (pos != string::npos)
      {
         thisfqname     = thishostname;
         thishostname   = thisfqname.substr(0,pos);
         thisdomainname = thisfqname.substr(pos+1);
      }

      // Look up this host in name service.
      
      struct hostent *hostp;
      
      if (!(hostp = gethostbyname(thishostname.c_str())))
      {
         cerr << "Unable to lookup name and address for localhost" << endl;
         hostEntLock.Unlock();
         return;
      }

      // Get domainname from this entry if not obtained before and available now.

      if (thisdomainname.empty())
      {
         thisfqname = hostp->h_name;
         Lower(thisfqname);
         pos = thisfqname.find_first_of(".");
         if (pos != string::npos)
            thisdomainname = thisfqname.substr(pos+1);
      }   
   
      // Get the local host address.
      
      unsigned char *addr = (unsigned char *) hostp->h_addr;

      OSTRSTREAM fmt_addr;
      
      fmt_addr << (int) *addr++ << "."
               << (int) *addr++ << "."
               << (int) *addr++ << "."
               << (int) *addr++ << '\0';
      
      thisipaddr = fmt_addr.str();
      
   }

   hostEntLock.Unlock();
}

// ====================================================================
// Static access to local host information.
// ====================================================================

string Host::LocalNode(void)
{
   if (! initialized) initialize();
   
   return thishostname;
}

string Host::LocalDomain(void)
{
   if (! initialized) initialize();
   
   return thisdomainname;
}

string Host::LocalFQName(void)
{
   if (! initialized) initialize();
   
   return thisfqname;
}

string Host::LocalAddr(void)
{
   if (! initialized) initialize();
   
   return thisipaddr;
}


// ====================================================================
// Internal function to set hostname and address from gethost** reply.
// ====================================================================

void Host::SetNameAddr(struct hostent *reply)
{
   if (reply)
   {  
      // Copy the primary name from the reply
      name = reply->h_name;
      Lower(name);

      // Copy the primary address from the reply.
      addrt **q = (addrt **) reply->h_addr_list;
      sa.sin_addr.s_addr = **q;
      addrset = true;

      // Recognize localhost specially, because there's no telling
      // what it's going to return for the lookup.
      if (sa.sin_addr.s_addr == htonl(INADDR_LOOPBACK))
      {
         name = "localhost";
      }
   }

}

// ====================================================================
// Find host information given either the hostname or the
// address in dotted-decimal notation.
// ====================================================================

bool Host::Find(string search)
{
   struct hostent *reply;

   name = "";
  
   hostEntLock.Lock();
   
   reply = gethostbyname(search.c_str());

   error = h_errno;             // record status.

   SetNameAddr(reply);

   hostEntLock.Unlock();

   return (reply) ? true : false; // Return whether name was resolved
}

// ====================================================================
// Find host information given the address in network byte order.
// ====================================================================

bool Host::Find(addrt search)
{
   struct hostent *reply;

   name = "";
  
   hostEntLock.Lock();
   
   reply = gethostbyaddr((char *) &search, sizeof(search), AF_INET);

   error = h_errno;             // record status.

   SetNameAddr(reply);

   hostEntLock.Unlock();

   return (reply) ? true : false; // Return whether name was resolved
}

// ====================================================================
// Get the peer name of a connected socket.
// ====================================================================

bool Host::GetPeerName(int fd)
{
   socklen_t  addrlen = Len();
   bool result = false;
   int rc;

   hostEntLock.Lock();
   
   rc = getpeername(fd, Ptr(), &addrlen);
   
   hostEntLock.Unlock();
   
   if (rc == 0)
   {
      result = Find(sa.sin_addr.s_addr);
   }
   
   return result;
}

// ====================================================================
// Get the peer name given an address structure, as from accept() or
// recvfrom().
// ====================================================================

bool Host::GetPeerName(struct sockaddr_in addr)
{
   bool result = false;

   memcpy(&sa, &addr, sizeof(sa));

   result = Find(sa.sin_addr.s_addr);
   
   return result;
}

// ====================================================================
// Get the peer name of the address already filled in.
// ====================================================================

bool Host::GetPeerName(void)
{
   bool result = false;

   result = Find(sa.sin_addr.s_addr);
   
   return result;
}

// ====================================================================
// Return the primary host name.
// ====================================================================

string Host::Name(void)
{
   string myname = name;

   if (! initialized) initialize();

   string::size_type pos = myname.find_first_of(".");
   if (pos != string::npos)
   {
      if (myname.substr(pos+1) == thisdomainname)
         myname.erase(pos);
   }
   
   return myname;
}

string Host::Node(void)
{
   string myname = name;

   string::size_type pos = myname.find_first_of(".");
   if (pos != string::npos) myname.erase(pos);
   
   return myname;
}

string Host::Domain(void)
{
   string myname = name;

   string::size_type pos = myname.find_first_of(".");
   myname.erase(0, pos);
   
   return myname;
}

string Host::FQName(void)
{
   return name;
}

// ====================================================================
// Return the primary address in network order.
// ====================================================================

addrt Host::Addr(void)
{
   return sa.sin_addr.s_addr;
}

// ====================================================================
// Return the primary address in dotted-decimal form.
// ====================================================================

string Host::AddrStr(void)
{
   OSTRSTREAM temp;
   unsigned char a[4];
   memmove(a, &sa.sin_addr.s_addr, 4);
   
   temp << (a[0] & 0xff) << "."
        << (a[1] & 0xff) << "."
        << (a[2] & 0xff) << "."
        << (a[3] & 0xff);
   
   return temp.str();
}

// ====================================================================
// Find service information given the service name and protocol.
// ====================================================================

bool Host::FindPort(const string &search, const string &pr)
{
   struct servent *reply;

   name = "";
  
   hostEntLock.Lock();
   
   reply = getservbyname(search.c_str(), pr.c_str());

   if (reply)
   {  
      sa.sin_port = reply->s_port;
   }

   hostEntLock.Unlock();

   return (reply) ? true : false; // Return whether name was resolved
}

// ====================================================================
// Find service information given port number and protocol.
// ====================================================================

bool Host::FindPort(int search, const string &pr)
{
   struct servent *reply;

   name = "";
  
   hostEntLock.Lock();
   
   reply = getservbyport(search, pr.c_str());

   if (reply)
   {  
      sa.sin_port = reply->s_port;
   }

   hostEntLock.Unlock();

   return (reply) ? true : false; // Return whether name was resolved
}

// ====================================================================
// Set the port.
// ====================================================================

void Host::SetPort(short port)
{
   sa.sin_port = htons(port);
}

// ====================================================================
// Return the port.
// ====================================================================

short Host::Port(void)
{
   return ntohs(sa.sin_port);
}

