// ====================================================================
// Logfile class.  Log is opened on demand by writing to the file, and
// supports autoclosing.
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.4 $ $Date: 2005/02/09 00:48:13 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include "global.h"
#include "logfile.h"
#include "syslog.h"

#include <iostream>
#include <unistd.h>

// ====================================================================
// Constructor
// ====================================================================

Logfile::Logfile(void)
{
   owner     = getuid();
   ownergrp  = getgid();
   mode      = 0644;
}

string Logfile::Fullname(void)
{
   string fullname;
   char logfilename[200];
   
   // Put together the directory name and the file name.

   if (name[0] != '/')
   {
      fullname = dir;
      if (! fullname.empty()) fullname += '/';
   }
   
   fullname += name;

   // Substitute any % escapes in the full path name.

   time_t now = time(NULL);

   strftime(logfilename, sizeof(logfilename), fullname.c_str(), localtime(&now));

   fullname = logfilename;
   
   return fullname;
}

// ====================================================================
// Open a log file.  This needs locking to open the file and update
// the flag atomically, because another thread might try to close it
// on us.  However, the Distribute() routine needs to lock in a larger
// scope, so we take a flag available to internal routines only to 
// skip the locking.
// ====================================================================

void Logfile::OpenLog(void)
{
   OpenLog(true);               // External callers always lock.
}

void Logfile::OpenLog(bool locklog)
{
   if (locklog) logopen.Lock(); // Lock if requested.

   // Now open the log file.  Make sure the owner is set correctly.

   try 
   {
      Open(Fullname().c_str(), O_RDWR | O_CREAT | O_APPEND | O_NONBLOCK, mode);
      if (GLOBAL.Verbose())
      {
         SYSLOG << Severity(Syslog::debug)
                << "Opened log file "
                << filename
                << endl;
      }
      logopen = true;
      Chown(owner, ownergrp);
      Chmod(mode);
   }
   catch(File::IOErr e)
   {
      SYSLOG << Severity(Syslog::error)
             << e.routine 
             << ": " << e.filename 
             << ": " << e.message 
             << endl;
   }
   
   if (locklog) logopen.Unlock(); // Unlock if we locked it.
}

// ====================================================================
// Close a log file.
// ====================================================================

void Logfile::CloseLog(void)
{
   CloseLog(true);              // External callers always lock.
}

void Logfile::CloseLog(bool locklog)
{
   if (! logopen) return;
   
   if (locklog) logopen.Lock();           // Make sure no one is writing.

   try
   {
      Close();
      logopen = false;

      if (GLOBAL.Verbose())
      {
         SYSLOG << Severity(Syslog::debug)
                << "Closed log file "
                << FileName()
                << endl;
      }
   }
   catch(File::IOErr e)
   {
      // Ignore
   }

   if (locklog) logopen.Unlock();
}

// ====================================================================
// Write a record to a logfile, opening if necessary.
// ====================================================================

void Logfile::WriteLog(Buffer &record)
{
   try
   {
      logopen.Lock();
      if (! logopen) OpenLog(false); // Open without locking
      if (logopen)   Write(record);
      logopen.Unlock();
   }
   catch(File::IOErr e)
   {
      SYSLOG << Severity(Syslog::error)
             << e.routine 
             << ": " << e.filename 
             << ": " << e.message 
             << endl;
      CloseLog(false);
      logopen.Unlock();
   }
}

// ====================================================================
// Write a record to a logfile, opening if necessary.
// ====================================================================

void Logfile::ReadLog(Buffer &buffer, howval how)
{
   try
   {
      logopen.Lock();
      if (! logopen) OpenLog(false); // Open without locking
      if (logopen)   
      {
         int off = 0;
         if (how == Tail) off = - buffer.Space();
         
         lseek(fd, how, off);
         
         Read(buffer);
      }
      CloseLog(false);
      logopen.Unlock();
   }
   catch(File::IOErr e)
   {
      SYSLOG << Severity(Syslog::error)
             << e.routine 
             << ": " << e.filename 
             << ": " << e.message 
             << endl;
      CloseLog(false);
      logopen.Unlock();
   }
}

