// ====================================================================
// Class to implement multiple locks.  A lock can be locked for
// shared access or exclusive access.  Any number of threads
// can hold a shared lock; only one thread can hold an exclusve
// lock.  Exclusive locks have priority over shared locks.
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.6 $ $Date: 2005/02/09 00:48:13 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include "mlock.h"

void rcleanup(void *);
void wcleanup(void *);

// ====================================================================
// Constructor: initialize the lock structures.
// ====================================================================

MLock::MLock(void)
{
   pthread_mutex_init(&mutex, NULL);
   pthread_cond_init(&swait, NULL);
   pthread_cond_init(&xwait, NULL);

   numLocks = 0;
   waitingXLocks = 0;
   predicate = false;
}

// ====================================================================
// Copy constructor: this is a new instance so needs new locks.
// ====================================================================

MLock::MLock(MLock &rhs)
{
   pthread_mutex_init(&mutex, NULL);
   pthread_cond_init(&swait, NULL);
   pthread_cond_init(&xwait, NULL);

   numLocks = 0;
   waitingXLocks = 0;
   predicate = false;
}

// ====================================================================
// Destructor: deallocate storage for the object.
// ====================================================================

MLock::~MLock(void)
{
   scleanup(this);
   xcleanup(this);
}

// ====================================================================
// Lock cleanup handlers
// ====================================================================

void scleanup(void *l)
{
   MLock *lock = (MLock *)l;
   
   if (lock->numLocks != 0) pthread_mutex_unlock(&lock->mutex);
}

void xcleanup(void *l)
{
   MLock *lock = (MLock *)l;

   lock->waitingXLocks--;

   if ((lock->waitingXLocks == 0) && (lock->numLocks >= 0))
      // Cancelled while still waiting for shared locks to finish,
      // let any blocked shared locks try to resume.
      pthread_cond_broadcast(&lock->xwait);
   
   if (lock->numLocks != 0) pthread_mutex_unlock(&lock->mutex);
}


// ====================================================================
// Obtain a lock.
// ====================================================================

void MLock::Lock(locktype type)
{
   pthread_mutex_lock(&mutex);

   if (type == Shared)
   {
      // Register a cleanup in case this thread is cancelled.
      pthread_cleanup_push(&scleanup, this);

      // Wait while any current writers, or if anyone wants to write.
      while ((numLocks < 0) || (waitingXLocks > 0))
      {
         pthread_cond_wait(&swait, &mutex);
      }

      numLocks++;

      // Pop and call the cleanup routine to release resources.
      pthread_cleanup_pop(1);
   }
   else // type == Exclusive
   {
      waitingXLocks++;

      // Register a cleanup in case this thread is cancelled.
      pthread_cleanup_push(xcleanup, this);

      // Wait while there's still active shared or exclusive locks.
      while (numLocks != 0)
      {
         pthread_cond_wait(&xwait, &mutex);
      }
      
      numLocks = -1;

      // Pop and call the cleanup routine to release resources.
      pthread_cleanup_pop(1);
   }
}

// ====================================================================
// Try to obtain a lock, but return false instead of waiting if it
// is already locked.  Return true if lock obtained.
// ====================================================================

bool MLock::TryLock(locktype type)
{
   pthread_mutex_lock(&mutex);
   bool gotlock = false;

   if (type == Shared)
   {
      // Fail if any current writers, or if anyone wants to write.
      if ((numLocks < 0) || (waitingXLocks > 0)) 
      {
         gotlock = false;
      }
      else
      {
         gotlock = true;
         numLocks++;
      }
   }
   else // type == Exclusive
   {
      // Fail if there's active shared or exclusive locks.
      if (numLocks != 0)
      {
         gotlock = false;
      }
      else
      {
         gotlock = true;
         numLocks = -1;
      }
   }
   pthread_mutex_unlock(&mutex);
   return gotlock;
}

// ====================================================================
// Release a lock
// ====================================================================

void MLock::Unlock(locktype type)
{
   pthread_mutex_lock(&mutex);
   
   if (type == Shared)
   {
      numLocks--;
      if (numLocks == 0)
         pthread_cond_signal(&xwait);
   }
   else
   {
      numLocks = 0;
      if (waitingXLocks == 0)
         pthread_cond_broadcast(&swait);
      else
         pthread_cond_signal(&xwait);
   }

   pthread_mutex_unlock(&mutex);
}
