// ====================================================================
// msgclient.cc -- Send/Receive records to Message Service Dispatcher
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.6 $ $Date: 2005/02/09 00:48:13 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include "msgclient.h"
#include "servent.h"
#include "utility.h"

#include <string>
#include <sys/socket.h>
#include <unistd.h>

// ====================================================================
// Destructor
// ====================================================================

MessageClient::~MessageClient(void)
{
   if (server) 
   {
      delete server;
   }
}


// ====================================================================
// Perform the requested send or receive operation.
// I/O exceptions are not handled; they are allowed to escape to
// the caller to handle.
// ====================================================================

void MessageClient::OpenStream(MessageDirection dir, 
                               Host &h,
                               string stream)
{
   vector<string> streamlist;

   streamlist.push_back(stream);

   OpenStream(dir, h, streamlist);
}

void MessageClient::OpenStream(MessageDirection dir, 
                               Host &h,
                               vector<string> &streamlist)
{
   host = h;
   
   // Set up control record for the connection.
   // First char indicates the type of connection.
   control.Clear();
   control += (char) dir;
   
   // Next is username and semicolon.
   username = FindUsername();
   control += username;
   control += ';';

   // Next is comma-delimited stream names.
   vector<string>::iterator pos = streamlist.begin();
   vector<string>::iterator epos = streamlist.end();
   
   for ( ; pos < epos; pos++)
   {
      control += *pos;
      control += ',';
   }
   
   // Back up and change last comma to newline.
   control.WSeek(Buffer::Current, -1);
   control += '\n';
   
   // Open connection
   if (type == TCP) server = new SocketTCP;
   else             server = new SocketUDP;
   
   server->Connect(host);

   // Send control record if required.
   if (dir == SUBSCRIBER || type == TCP) server->Send(control);

   // Keep the control record in case we need to reopen.
}
   
// ====================================================================
// Reopen a stream if closed by a communication error.
// ====================================================================

void MessageClient::ReopenStream(void)
{
   for (;;)
   {
      CloseStream();
   
      try
      {
         server->Connect(host);
         server->Send(control);
         break;
      }
      catch(File::IOErr e)
      {
         sleep(60);
      }
   }
}

// ====================================================================
// Close a stream.
// ====================================================================

void MessageClient::CloseStream(void)
{
   server->Shutdown(SHUT_WR);
   server->Close();
}

// ====================================================================
// Send a record to the server.
// ====================================================================

void MessageClient::Send(Buffer &record)
{
   if (type == TCP)
   {
      server->Send(record);
   }
   else
   {
      Buffer output;
      output = control;
      output += record;
      server->Send(output);
   }
}


// ====================================================================
// Receive a record from the server.
// ====================================================================

int MessageClient::Receive(Buffer &record)
{
   return server->Receive(record);
}
