// ====================================================================
// reader.cc - Virtual class to define behaviors of other classes that
//             can create threads to read from a file.
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.5 $ $Date: 2005/02/09 00:48:13 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include "consoled.h"
#include "reader.h"
#include "syslog.h"
#include "file.h"

#include <iostream>

// ====================================================================
// Start a thread running the reader and return.
// ====================================================================

static void threxit(void *data)
{
   ((Reader *) data)->Exited();
}

static void *starter(void *data) 
{ 
   pthread_cleanup_push(&threxit, data);
   ((Reader *) data)->ReaderProc(); 
   pthread_cleanup_pop(1);
   return data; 
};

void Reader::StartReader(void)
{
   int status;
   pthread_attr_t attrib;
   pthread_attr_init(&attrib);
   pthread_attr_setstacksize(&attrib, 512*1024);
   
   status = pthread_create(&thread, &attrib, starter, this);
   
   if (status != 0)
   {
      thread = 0;
      
      SYSLOG << Severity(Syslog::error)
             << "Thread error " 
             << status 
             << " in Reader::StartReader: "
             << strerror(status) 
             << endl;
     
      throw File::IOErr("StartReader", status, "pthread_create");
   }
}

// ====================================================================
// Stop the reader thread.
// ====================================================================

void Reader::StopReader(void)
{
   int status;

   if (thread == 0) return;
   
   status = pthread_detach(thread);
   status = pthread_cancel(thread);
}

// ====================================================================
// Detach the reader thread.
// ====================================================================

void Reader::Detach(void)
{
   if (thread == 0) return;
   
   pthread_detach(thread);
}

// ====================================================================
// Join the reader thread, wait for its completion.
// ====================================================================

void Reader::Join(void)
{
   void *result;
   int  status;

   if (thread == 0) return;
   
   status = pthread_join(thread, &result);
}


