// ====================================================================
// sensorsoft.cc - A filter for the sensorsoft temperature sensors.
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.10 $ $Date: 2006/07/03 12:50:30 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include "defaults.h"
#include "consoled.h"
#include "msgclient.h"
#include "sensorsoft.h"

#include <iostream>
#include <stdio.h>

enum {WAIT_ID, WAIT_STATUS, WAIT_DATA, NONE}; // States

// ====================================================================
// Structure of sensorsoft response buffer.
//
// unsigned char code;        (single byte response code)
// unsigned char length[2];   (little endian short integer)
// unsigned char data[*];     (data bytes)
// unsigned char crc[2];      (little endian short integer)
// ====================================================================

typedef struct response
{
   unsigned char code;
   unsigned char length[2];
   unsigned char data[1];
} Response;

static unsigned char   NORMAL_RESPONSE = '\x90';
static unsigned char ABNORMAL_RESPONSE = '\x94';

static unsigned char POWER_LOW = '\x01';
static unsigned char POWER_UP  = '\x08';
static unsigned char BROKEN    = '\x10';

// ====================================================================
// Constructor -- Create a new filter.  
// ====================================================================

SensorSoft::SensorSoft(Console *cons) 
      : Filter(cons)
{
   nextQuery = int((time(NULL) + INTVL) / INTVL) * INTVL;
   logbuffer.Prefix(console->Name().size() + 9, console->Name() + " AMBIENT ");
   
//   db.Open("/tmp/db." + console->Name(), O_RDWR | O_CREAT | O_APPEND  | O_NONBLOCK, 0600);
   
   state = NONE;
   isOpen = 0;
}

// ====================================================================
// Receive notification that console is open.
// ====================================================================

void SensorSoft::Open(void)
{
   try 
   {
      console->Write(QID, QIDLEN);
   }
   catch(File::IOErr e)
   {
      return;
   }

   isOpen = 1;
   state  = WAIT_ID;  
}

// ====================================================================
// Receive notification that console is closed.
// ====================================================================

void SensorSoft::Close(void)
{
   isOpen = 0;
}

// ====================================================================
// On idle call from the dispatcher, drive the query.
// ====================================================================

void SensorSoft::IdleCheck(time_t now)
{
   if (! isOpen) return;
   
   if (state == WAIT_ID)
   {
      console->Write(QID, QIDLEN);
      return;
   }
   
   if (now >= nextQuery)
   {
      state = WAIT_DATA;
      console->Write(QUERY, QUERYLEN);
      nextQuery = int((time(NULL) + INTVL) / INTVL) * INTVL;
   }
}

// ====================================================================
// Receive the response from the interface.
// ====================================================================

void SensorSoft::Input(Buffer &in)
{
//   db.Write(in);

   // Accumulate in buffer.

   input += in;
   in.Clear();

   // Map structure over buffer.
   
   Response *rp = (Response *) input.RPos();

   // If more bytes to read, go around again.

   if (input.Size() < rp->length[0]) return;   

   bool abnormal = (rp->code == ABNORMAL_RESPONSE);

   // Got back something unexpected.

   if (rp->code != NORMAL_RESPONSE && rp->code != ABNORMAL_RESPONSE)
   {
      string error = "Bad data: ";
      for (unsigned char *p = input.RPos(); p < input.WPos(); p++) 
      {
         error += (*p >> 8) + (*p++ & '\xf0');
      }
      error += "\r\n";
      in = error;
      return;
   }

   // Got back a response identifying the sensor.
   // Extract the fields and reformat to be readable.

   if (state == WAIT_ID)
   {
      char *p = (char *) rp->data+6;
      name = p;
      while(*p++ != '\x00') {  }
      while(*p++ != '\x00') {  }
      model = p;
      while(*p++ != '\x00') {  }
      firmware = p;
      
      in += name + ' ' + model + ' ' + firmware + "\r\n";
   }

   // Got back a status record

   else if (state == WAIT_STATUS)
   {
      char data = rp->data[0];

      if (data && data != POWER_UP)
      {
         Buffer error;
         error.TimeStampFormat("%T ");
         error += "Error status received: ";
         if ((data & POWER_LOW) == POWER_LOW) error += "low-voltage ";
         if ((data & POWER_UP ) == POWER_UP ) error += "power-up ";
         if ((data & BROKEN   ) == BROKEN   ) error += "sensor-broken ";
         error += "\r\n";
         
         error.TimeStamp();
         console->WriteLog(error);
      }
   }

   // Otherwise, got back a data record.

   else if (state == WAIT_DATA)
   {
      int data = rp->data[0];
      float temp = data;

      if (rp->data[1] == '\x00') temp /= 2;
      
      sprintf((char *)in.WPos(), "%4.1f\r\n", temp);
      in.WSeek(Buffer::Beginning,6);

      logbuffer.Clear();
      logbuffer += in;

      DAEMON.SendMessage(console->StreamName(), logbuffer);
   }

   // Error indicator is on, read status.

   if (abnormal)
   {
      state = WAIT_STATUS;
      console->Write(STATUS, STATLEN);
   }

   input.Clear();
}

