// ====================================================================
// server.cc - Implementation of the Server class
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.6 $ $Date: 2006/12/05 06:44:28 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include "global.h"
#include "consolec.h"
#ifdef HAVE_LIBGSSAPI
#include "gssapi.h"
#endif
#include "server.h"
#include "syslog.h"
#include "defaults.h"

#include <iostream>

// ====================================================================
// Constructor and Destructor
// ====================================================================

Server::Server(void)
{
}

Server::~Server(void)
{
}

// ====================================================================
// Get next response from server, delimited by [ ... ]\n
// ====================================================================

string Server::GetNextResponse(void)
{
   string response;
   unsigned char c;
   enum {leading, main, trailing, ending, auth1, auth2, auth3} state;

   state = leading;

   for (;;)
   {
      while (Receive(&c, 1))
      {
         if (GLOBAL.Debug()) terminal->Write(c);
         
         switch(state)
         {
             case leading:
                if (c == '[') state = main;
                break;
                
             case main:
                if (c == ']') state = trailing;
                else if (c == '{') state = auth1;
                else response += c;
                break;

             case trailing:
                if (c == '\n') { return response; };
                break;

             // The authorization requests have to be valid
             // or the characters are just passed through.
             case auth1:
                state = main;
                if (c == 'p') state = auth2;
                else if (c == 'g') state = auth3;
                else response += '{' + c;
                break;
                
             case auth2:
                if (c == '}') AuthPass();
                else response += "{p" + c;
                state = main;
                break;

             case auth3:
                if (c == '}') AuthGSSAPI();
                else response += "{g" + c;
                state = main;
                break;

             case ending:
                return response;
                break;
         }
      }

   }
}


// ====================================================================
// Read from server and echo to terminal until received [disconnect].
// ====================================================================

void Server::ReadToEnd(void)
{
   string response;
   unsigned char c;
   enum {leading, main, trailing, ending} state;

   state = leading;

   for (;;)
   {
      while (Receive(&c, 1))
      {
         if (state == leading)
         {
            if (c == '[') state = main;
            else terminal->Write(c);
            continue;
         }
         if (state == main)
         {
            if (c == ']') 
            {
               if (response == "disconnect") state = trailing;
               else 
               {
                  terminal->Write(response);
                  state = leading;
               }
            }
            
            else response += c;
            continue;
         }
         if (state == trailing)
         {
            if (c == '\n') { state = ending; break; };
         }
      }

      if (state == ending) break;
   }
}


// ====================================================================
// Get next word from server.
// ====================================================================

string Server::GetNextWord(void)
{
   string word;
   unsigned char c;
   enum {leading, main, trailing, ending} state;

   state = leading;

   for (;;)
   {
      while (Receive(&c, 1))
      {
         if (state == leading)
         {
            if (! isspace(c)) state = main;
         }
         if (state == main)
         {
            if (isspace(c))   { state = ending; break; }
            else word += c;
         }
      }

      if (state == ending) break;
   }

   return word;
}

// ====================================================================
// Authenticate with password
// ====================================================================

void Server::AuthPass(void)
{
   string word;
}

// ====================================================================
// Authenticate with GSSAPI
// ====================================================================

void Server::AuthGSSAPI(void)
{
#ifdef HAVE_LIBGSSAPI
   GSSAPI g(this);
   if (gssservice == "")
      gssservice = "console/" + Peername();
   if (GLOBAL.Verbose()) cerr << "GSSAPI authentication attempted with service " 
                              << gssservice << "\r\n";
   g.SetService(gssservice);
   if (g.Connect())
   {
      if (GLOBAL.Verbose()) cerr << "GSSAPI authentication succceed with " 
                                 << Peername() << "\r\n";
   }
   else cerr << "GSSAPI authentication failed\r\n";
#endif
}

// ====================================================================
// ReaderProc() - loops through reading from the socket
// connection and writing that data to the terminal. 
// ====================================================================

void Server::ReaderProc(void)
{
   int len;
   
   for(; ;)
   {
      termbuffer.Clear();
      
      try
      {
         // Read from the socket connection 
         len = Receive(termbuffer);
         if (len == 0) break;
         
         // Write to the terminal

         terminal->Write(termbuffer);
      }
      catch (IOErr e)
      {  
         cout << "socket got error\r\n";
         
         SYSLOG << Severity(Syslog::error)
                << e.routine 
                << ": " << e.filename 
                << ": " << e.message 
                << endl;
         
         return;
      }
      
      // If len == 0, the connection has been closed 

      if (len == 0) break;
   }

   if (GLOBAL.Verbose()) cerr << "Server reader thread ended -- Server closed connection\r\n";
}



