// ====================================================================
// Simple lock class.  Provides a single lock (mutex) and wait
// condition (condition variable).  Locks a resource for exclusive
// access and waits for its release.  For a lock that handles 
// read-write and read-only locking, see mlock. 
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.1 $ $Date: 2005/02/09 00:38:50 $ $Author: boeheim $
// ******************************************************************** //

#ifndef slock_h
#define slock_h

#include <pthread.h>
#include <time.h>
#include <errno.h>

class SLock
{
private:
   pthread_mutex_t lock;
   pthread_cond_t  wait;
   int status;
   bool predicate;
   
public:
   // Initialize the pthreads objects.
   SLock(void)
   {
      pthread_cond_init(&wait, NULL);
      pthread_mutex_init(&lock, NULL);
      status = 0;
      predicate = 0;
   }
   // Copy constructor
   SLock(SLock &rhs)
   {
      pthread_cond_init(&wait, NULL);
      pthread_mutex_init(&lock, NULL);
      status = 0;
      predicate = 0;
   }

   // Provide a Predicate for use with the condition variable.
   // The predicate is the name of the class itself, i.e.
   // SLock mylock;  mylock = true;   if (mylock) ...
   inline bool operator= (bool p) { return predicate = p; };
   inline      operator bool()    { return predicate; };
   
   // Return status of last operation.
   inline int  Status(void)    { return status; };

   // Lock and Wait operations, return true if success, false if failure.
   inline bool Lock(void)      { return (status = pthread_mutex_lock(&lock))       == 0; };
   inline bool Unlock(void)    { return (status = pthread_mutex_unlock(&lock))     == 0; };
   inline bool Wait(void)      { return (status = pthread_cond_wait(&wait, &lock)) == 0; };
   inline bool Signal(void)    { return (status = pthread_cond_signal(&wait))      == 0; };
   inline bool Broadcast(void) { return (status = pthread_cond_broadcast(&wait))   == 0; };

   // Wait with timeout, return true if not timeout, false if timeout.
   // (Check Status() for other failures.)
   inline bool TimedWait(int sec)      
      { struct timespec timeout; 
        timeout.tv_sec = sec;
        timeout.tv_nsec = 0;
        return (status = pthread_cond_timedwait(&wait, &lock, &timeout)) != ETIMEDOUT; 
      };
};

#endif
