// ====================================================================
// Socket.cc -- An abstract class for sockets that is the basis
//              for specific types of sockets.
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.13 $ $Date: 2006/12/05 06:45:38 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include <iostream>
#include <arpa/inet.h>
#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>

#include "autoconfig.h"
#include "buffer.h"
#include "defaults.h"
#include "host.h"
#include "slock.h"
#include "socket.h"
#include "syslog.h"
#include "utility.h"

extern int errno;

// ====================================================================
// Constructor: create a socket
// ====================================================================

Socket::Socket(void)
{
   cs = UNKNOWN;
}

Socket::Socket(int forfd)
{
   
   fd = forfd;
   cs = UNKNOWN;
}

// ====================================================================
// Destroy a socket
// ====================================================================

Socket::~Socket(void)
{
}

// ====================================================================
// Peername() - Find the host address and name of the other end
//              of a socket connection.
// ====================================================================

string Socket::Peername(void)
{
   // May have already looked it up and cached it.
   if (peername.size() != 0) return peername;

   peername = "unknown";

   Host peerhost;
   
   if (! peerhost.GetPeerName(fd))
   {
      SYSLOG << Severity(Syslog::error)
	     << "Error getting peername: "
	     << strerror(errno)
	     << endl;
      return peername; // Go with 'unknown'
   }
      
   peername = peerhost.FQName();
      
   Lower(peername);

   
   return peername;
}

// ====================================================================
// Bind a socket to a specified port and address.
// ====================================================================

void Socket::Bind(Host host)
{
   cs = SERVER;
   if (fd < 0) Create();

   SmallBuffer temp;
   temp += "*:";
   temp += host.Port();
   temp += "(LISTEN)";
   filename = (char *) temp.Cstr();
   
   if (bind(fd, host.Ptr(), host.Len()) < 0)
      throw IOErr("bind()", errno, filename);
}

// ====================================================================
// Initialize a socket to connect to a remote service.
// Used for the client side of a connection.
// ====================================================================

void Socket::Connect(Host host)
{
   cs = CLIENT;
   if (fd < 0) Create();

   SmallBuffer temp;
   temp += host.Name();
   temp += ":";
   temp += host.Port();
   filename = (char *) temp.Cstr();
   
   if (connect(fd, host.Ptr(), host.Len()) < 0)
      throw IOErr("connect()", errno, filename);
}

// ====================================================================
// Write the contents of a buffer to connected socket.
// ====================================================================

void Socket::Send(const unsigned char *buffer, int len)
{
   int error = 0;

   // Default buffer length is string length.

   if (len == 0) len = strlen((char *)buffer);

   error = ::send(fd, (char *) buffer, len, 0);
   
   if (error == -1) 
   {
      throw IOErr("send", errno, filename);
   }
}

// ====================================================================
// Write the contents of a buffer to an unconnected socket.
// ====================================================================

void Socket::SendTo(Host &dest, const unsigned char *buffer, int len)
{
   int error = 0;
   int destlen = dest.Len();

   // Default buffer length is string length.

   if (len == 0) len = strlen((char *)buffer);

   error = ::sendto(fd, (char *) buffer, len, 0, dest.Ptr(), destlen);
   
   if (error == -1) 
   {
      throw IOErr("sendto", errno, filename);
   }
}

// ====================================================================
// Receive from a socket into a buffer.
// ====================================================================

int Socket::Receive(unsigned char *buffer, int len)
{
   int amt;

   amt = ::recv(fd, (char *) buffer, len, 0);
   
   if (amt == -1) 
   {
      throw IOErr("recv", errno, filename);
   }
   
   return amt;
}

// ====================================================================
// Receive from a socket into a buffer with the return address.
// ====================================================================

int Socket::ReceiveFrom(Host &addr, unsigned char *buffer, int len)
{
   int amt;
   socklen_t addrlen = addr.Len();

   amt = ::recvfrom(fd, (char *) buffer, len, 0, addr.Ptr(), &addrlen);
   
   if (amt == -1) 
   {
      throw IOErr("recvfrom", errno, filename);
   }
   
   return amt;
}
