// ====================================================================
// Socket.hh -- An abstract class for sockets that is the basis
//              for specific types of sockets.
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.3 $ $Date: 2006/07/13 08:07:25 $ $Author: boeheim $
// ******************************************************************** //

#ifndef socket_hh
#define socket_hh

#include <sys/socket.h>

#include "file.h"
#include "buffer.h"
#include "host.h"

typedef enum {UNKNOWN, CLIENT, SERVER} csparm;

class Socket : public File
{
protected:
   csparm cs;                   // Indicate client or server.
   string peername;             // Name of host connected to.

   void SetFD(int newfd) { fd = newfd; };

   // Hide public methods from File, use Send/Receive instead.
   void Write(const Buffer &buffer);
   void Write(const string &buffer, int len=0);
   void Write(const char *buffer, int len=0);
   void Write(const char cc);
   int  Read(Buffer &buffer);
   int  Read(char *buffer, int len);

public:
   Socket(void);                // Constructor.
   Socket(int fd);              // Constructor.
   virtual ~Socket(void);       // Destructor.
   
   // Setup functions.
   virtual void Create(void) = 0; // Create OS socket.
   void Bind(Host host);        // Bind to address.
   void Connect(Host host);     // Connect to a remote host.
   void Shutdown(int how) { ::shutdown(fd, how); }; // Shut down a socket.

   // Send function overloaded to take different data types.
   void Send(const Buffer &buffer)                   { Send(buffer.RPos(), buffer.Size()); };
   void Send(const string &buffer, int len=0)        { Send((unsigned char *)buffer.c_str(), len); };
   void Send(const unsigned char cc)                 { Send(&cc, 1); };
   void Send(const unsigned char *buffer, int len=0);
   void Send(const void *buffer, int len=0) { Send((unsigned char *) buffer, len); };
   void SendTo(Host &dest, const Buffer &buffer)            { SendTo(dest, buffer.RPos(), buffer.Size()); }
   void SendTo(Host &dest, const string &buffer, int len=0) { SendTo(dest, (unsigned char *)buffer.c_str(), len); }
   void SendTo(Host &dest, const unsigned char cc)          { SendTo(dest, &cc, 1); }
   
   void SendTo(Host &dest, const unsigned char *buffer, int len);

   // Receive function overloaded to take different data types.
   int  Receive(Buffer &buffer) 
      { int len = Receive(buffer.WPos(), buffer.Space());
        buffer.WSeek(Buffer::Current, len);
        return len;
      };
   int  Receive(unsigned char *buffer, int len);
   int  Receive(void *buffer, int len) { return Receive((unsigned char *)buffer, len);};

   int  ReceiveFrom(Host &addr, Buffer &buffer) 
      { int len = ReceiveFrom(addr, buffer.WPos(), buffer.Space());
        buffer.WSeek(Buffer::Current, len);
        return len;
      };
   int  ReceiveFrom(Host &addr, unsigned char *buffer, int len);

   string Peername(void);
};
   

#endif
