// ====================================================================
// Sockettcp.cc -- A class implementing tcp sockets
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.6 $ $Date: 2006/07/04 14:03:32 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#define SSLptr SSL

#include <iostream>
#include <openssl/err.h>
#include <openssl/ssl.h>
#include <openssl/rand.h>

#include "defaults.h"
#include "global.h"
#include "slock.h"
#include "socketssl.h"
#include "syslog.h"

extern int errno;

// ====================================================================
// Global SSL Setup
// ====================================================================

static  SSL_CTX *sslContext; // SSL Context structure
static  SLock   sslSetup;
static  char *CAFile   = NULL;
static  char *CAPath   = NULL;
static  char *CertFile = NULL;
static  char *KeyFile  = NULL;

#define RAND_SIZE 64            // Bytes of random data for seed.
#define RAND_DEV "/dev/urandom" // Where to read random data.
#define RSA_KEYSIZE 512         // Size of an RSA key

void SocketSSL::ThrowSSLErr(char *rtn)
{
   unsigned long int err;
   char errorbuf[256];
   
   err = ERR_get_error();
   throw File::IOErr(rtn, err, filename, ERR_error_string(err, errorbuf));
}

int verify_callback(int ok, X509_STORE_CTX *store)
{
    if (!ok)
    {
        X509 *cert = X509_STORE_CTX_get_current_cert(store);
        int  depth = X509_STORE_CTX_get_error_depth(store);
        int  err = X509_STORE_CTX_get_error(store);
 
        SYSLOG << "Cert error " << err
               << " at depth " << depth
               << ": " << X509_verify_cert_error_string(err)
               << endl;

        // Allow self-signed certs
        if (err == X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN) ok = true; 
    }
 
    return ok;
}

void SocketSSL::SSLSetup(void)
{
   if (sslSetup) return;      // Quick test for already setup.

   sslSetup.Lock();
   
   if (! sslSetup)  // Retest under lock to prevent race
   {
      struct stat statbuf;
      
      if (stat(RAND_DEV, &statbuf) == 0) 
      {
         RAND_load_file(RAND_DEV, RAND_SIZE);
      }
      if (! RAND_status())
      {
         throw File::IOErr("RAND_INIT", RAND_SIZE, filename, "Insufficient entropy");
      }
      
      SSL_load_error_strings();
      SSL_library_init();
      
      sslContext = SSL_CTX_new(SSLv23_method());
      if (! sslContext) ThrowSSLErr("SSL_CTX_new");

      if (CertFile)
      {
         if (SSL_CTX_use_certificate_chain_file(sslContext, CertFile) < 0)
            ThrowSSLErr("SSL_CTX_use_certificate_file");
         if (GLOBAL.Verbose()) 
            SYSLOG << Severity(Syslog::info)
                   << "Loaded certificate from " 
                   << CertFile
                   << endl;
      }
      
      if (KeyFile)
      {
         if (SSL_CTX_use_PrivateKey_file(sslContext, KeyFile, SSL_FILETYPE_PEM) < 0)
            ThrowSSLErr("SSL_CTX_use_PrivateKey_file");
         
         if (! SSL_CTX_check_private_key(sslContext))
            ThrowSSLErr("SSL_CTX_check_private_key");
      }

      if (CAFile || CAPath)
      {
         if (SSL_CTX_load_verify_locations(sslContext, CAFile, CAPath) < 0)
            ThrowSSLErr("SSL_CTX_load_verify_locations");
      
         SSL_CTX_set_verify(sslContext, SSL_VERIFY_PEER, verify_callback);
         SSL_CTX_set_verify_depth(sslContext, 2);
      }
      
      sslSetup = true;
   }

   sslSetup.Unlock();
}

// ====================================================================
// SSL Constructors. 
// ====================================================================

SocketSSL::SocketSSL(void)
{
   ssl = 0;
}

SocketSSL::SocketSSL(int fd)
      : SocketTCP(fd)
{
   ssl = 0;
}

SocketSSL::SocketSSL(SocketTCP &s)
      : SocketTCP(s)
{
   ssl = 0;
}

// ====================================================================
// Set the CA and cert locations
// ====================================================================

void SocketSSL::SetCAFile(const char *file)
{
   if (CAFile) free(CAFile);

   if (!file) { CAFile = NULL; return; }

   CAFile = (char *) malloc(1+strlen(file));
   strcpy(CAFile,file);
}

void SocketSSL::SetCAPath(const char *path)
{
   if (CAPath) free(CAPath);

   if (!path) { CAPath = NULL; return; }

   CAPath = (char *) malloc(1+strlen(path));
   strcpy(CAPath,path);
}

void SocketSSL::SetCertFile(const char *file)
{
   if (CertFile) free(CertFile);

   if (!file) { CertFile = NULL; return; }

   CertFile = (char *) malloc(1+strlen(file));
   strcpy(CertFile,file);

   // Assume a PEM file with key appended.
   if (! KeyFile) SetKeyFile(file);
}

void SocketSSL::SetKeyFile(const char *file)
{
   if (KeyFile) free(KeyFile);

   if (!file) { KeyFile = NULL; return; }

   KeyFile = (char *) malloc(1+strlen(file));
   strcpy(KeyFile,file);
}

// ====================================================================
// Create the SSL structures for the connection.
// ====================================================================

void SocketSSL::Create(void) 
{
   SocketTCP::Create();
   SSLSetup();
   if (!ssl) ssl = SSL_new(sslContext);
   SSL_set_fd(ssl,fd);
}

// ====================================================================
// Destroy a socket
// ====================================================================

SocketSSL::~SocketSSL(void)
{
   if (ssl) SSL_free(ssl);
}

// ====================================================================
// Accept an SSL connection
// ====================================================================

SocketSSL *SocketSSL::Accept(void)
{
   SocketSSL *newsocket = new SocketSSL;
   
   return Accept(newsocket);
}

SocketSSL *SocketSSL::Accept(SocketSSL *newsocket)
{
   // Get the TCP socket from a normal accept().

   SocketTCP::Accept(newsocket);
   
   // Do the SSL Accept.

   newsocket->AcceptSSL();
   
   return newsocket;
}

// Copy the TCP socket info to the SSL socket object, create
// the SSL structures, and accept the SSL handshake.

void SocketSSL::AcceptSSL(void)
{
   if (!ssl) ssl = SSL_new(sslContext);
   SSL_set_fd(ssl,fd);

   if (SSL_accept(ssl) < 0) ThrowSSLErr("SSL_accept");
}


// ====================================================================
// Initialize a socket to connect to a remote service.
// Used for the client side of a connection.
// ====================================================================

void SocketSSL::Connect(Host host)
{
   int rc;
   
   // Do TCP connection first.  This calls Create(), which will set up
   // the SSL structures.

   Socket::Connect(host);
   
   // Do the SSL handshake.

   if (SSL_connect(ssl) < 0) ThrowSSLErr("SSL_connect");

   if (CAFile || CAPath) CheckCert();
}

// ====================================================================
// Check certificate from partner.
// ====================================================================

bool SocketSSL::CheckCert(void)
{
   int rc;
   
   if ((rc = SSL_get_verify_result(ssl)) != X509_V_OK) 
   {
      cerr << "Error: peer certificate: " 
           << X509_verify_cert_error_string(rc)
           << endl;
      
      SYSLOG << Severity(Syslog::warning)
             << "SSL Certificate does not validate, error "
             << rc
             <<endl;
      return false;
   }

   X509 *peer = SSL_get_peer_certificate(ssl);
   char cn[256];

   if (! peer) 
   { 
      if (GLOBAL.Verbose())
      {
         SYSLOG << Severity(Syslog::warning)
                << "No SSL Certificate received from peer"
                <<endl;
      }
      return false;
   }

   X509_NAME_get_text_by_NID(X509_get_subject_name(peer), NID_commonName, cn, 256);
   peerCN = cn;
   
   if (GLOBAL.Verbose())
   {
      SYSLOG << Severity(Syslog::info)
             << "SSL Certificate from "
             << cn
             << " verified"
             <<endl;
   }

   return true;
}

// ====================================================================
// GetSSLCipher() - Get cipher negotiated by SSL.
//                  Pass a buffer for re-entrant code.
// ====================================================================

Buffer &SocketSSL::GetSSLCipher(Buffer &buffer)
{
   buffer.Clear();
   
   if (!ssl)
   {
      buffer += "Insecure connection";
      return buffer;
   }

   buffer += SSL_get_cipher(ssl);
   return buffer;
}

// ====================================================================
// Write the contents of a buffer to a file. 
// ====================================================================

void SocketSSL::Send(const unsigned char *buffer, int len)
{
   int error = 0;

   // Default buffer length is string length.

   if (len == 0) len = strlen((char *)buffer);
   error = SSL_write(ssl, (char *)buffer, len);
   
   if (error == -1) 
   { 
      ThrowSSLErr("SSL_write");
   }
}

// ====================================================================
// Read from a file into a buffer.
// ====================================================================

int SocketSSL::Receive(unsigned char *buffer, int len)
{
   int amt;

   amt = SSL_read(ssl, (char *)buffer, len);
   
   if (amt == -1) 
   {
      ThrowSSLErr("SSL_read");
   }
   
   return amt;
}
