// ====================================================================
// syslog.cc -  A class to encapsulate syslog and present it in a 
//              streamio-oriented manner.  Also make it thread-safe.
// ==================================================================== */

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.6 $ $Date: 2007/04/02 16:08:13 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include "syslog.h"

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <sys/syslog.h>
#include <syslog.h>
#include <unistd.h>


char *sevnames[] = {"emerg", "alert", "critical", "error", "warning", 
                    "notice", "info", "debug"};


// ====================================================================
// Initialize the log.
// ==================================================================== 

   void pthread_key_destructor(void *);

Syslog::Syslog(string prog, int cl, int sev)
{
   echofd = 1;                  // Default echo to STDOUT.

   progname = prog;             // Program name.
   logclass = cl;               // Default class.
   severity = sev;              // Default severity

   pthread_key_create(&datakey, pthread_key_destructor); // Create thread-data key.
}

Syslog::Syslog(char *prog, int cl, int sev)
{
   echofd = 1;                  // Default echo to STDOUT.

   progname = prog;             // Program name.
   logclass = cl;               // Default class.
   severity = sev;              // Default severity

   pthread_key_create(&datakey, pthread_key_destructor); // Create thread-data key.
}

// ====================================================================
// Individually reset the default values.
// ====================================================================

void Syslog::ProgName(string prog)
{
   progname = prog;
}

void Syslog::Class(int cl)
{
   logclass = cl;
}

// ====================================================================
// Destructor
// ====================================================================

Syslog::~Syslog(void)
{   
   perthread *threadinfo = getdata();
   
   if (threadinfo->wpos > threadinfo->buffer + PREFIX)
      WriteBuffer(threadinfo);
}

// ====================================================================
// Function to get the per-thread data.  If not initialized, create
// a new data block for this thread.
// ====================================================================

perthread *Syslog::getdata(void)
{
   perthread *threadinfo;

   threadinfo = (perthread *) pthread_getspecific(datakey);
   
   if (threadinfo == NULL)
   {
      threadinfo = (perthread *) malloc(sizeof(perthread));
      
      threadinfo->severity = severity;
      threadinfo->buffer = (char *) malloc(BSIZE);
      threadinfo->wpos = threadinfo->buffer + PREFIX;
      threadinfo->buffer[0] = '\0';
      
      pthread_setspecific(datakey, threadinfo);
   }
   
   return threadinfo;
}

// Cleanup function to release storage.

void pthread_key_destructor(void *value)
{
   perthread *threadinfo = (perthread *) value;

   free(threadinfo->buffer);
   free(threadinfo);
}

// ====================================================================
// Basic append to buffer functions.
// ====================================================================

void Syslog::Append(const unsigned char val, perthread *threadinfo)
{
   if (threadinfo == NULL) threadinfo = getdata();
   
   *threadinfo->wpos++ = val;
   
   if (val == '\n' || threadinfo->wpos > threadinfo->buffer+BSIZE) WriteBuffer(threadinfo);
}

void Syslog::Append(const unsigned char *val, perthread *threadinfo)
{
   if (threadinfo == NULL) threadinfo = getdata();

   const unsigned char *p = val;
   
   while (*p)
   {
      *threadinfo->wpos++ = *p;
      
      if (*p++ == '\n' || threadinfo->wpos > threadinfo->buffer+BSIZE) WriteBuffer(threadinfo);
   }
}

// ====================================================================
// Overloaded operators for streaming I/O
// ====================================================================

// Add integer to stream.

Syslog& Syslog::operator<< (int val)
{
   char num[11];                // No 32-bit integer can be more than 10 decimal digits.
   char *p = num+10;

   *p = '\0';
   
   do 
   {
      *--p = '0' + (int) (val % 10);
      val = val / 10;
   }  while (val > 0);
   
   Append(p);

   return *this;
}

// Add ostream object to stream.  Only handles putting endl object on stream.

Syslog& Syslog::operator<< (ostream & (val)(ostream &))
{  
//   if (endl == val) 
      Append('\n');
   
   return *this;
}

// Add Severity stream object to stream.

Syslog& Syslog::operator<< (const Severity &val)
{  
   perthread *threadinfo = getdata();
   
   threadinfo->severity = val.severity;
   
   return *this;
}

// ====================================================================
// WriteBuffer(): writes the data contained in the buffer
// to the syslog and possibly echos it to echofd. 
// ====================================================================

void Syslog::WriteBuffer(perthread *threadinfo)
{ 
   // Insure null termination for syslog(); save end location.

   *threadinfo->wpos = '\0';
   char *temp = threadinfo->wpos;

   // Point to beginning of string, and back off for prefix information.

   char *prefix = threadinfo->buffer + PREFIX;
   
   prefix -= (progname.size() + strlen(sevnames[threadinfo->severity]) + 4);

   // Reposition wpos to write the prefix information.  If we did this 
   // correctly, we won't overlay anything.

   threadinfo->wpos = prefix;
   
   Append(progname.c_str(), threadinfo);
   Append(" (", threadinfo);
   Append(sevnames[threadinfo->severity], threadinfo);
   Append(") ", threadinfo);
   
   // Write to syslog now.

   openlog(progname.c_str(), severity, logclass << 3);
   
   ::syslog(severity, "%.500s", prefix + progname.size() + 1);
   
   closelog();
   
   // If requested, also write to an open file.

   if (echofd)
   {
      ::write(echofd, prefix, temp - prefix);
   }

   // If collecting history, put in queue;

   if (histsize > 0)
   {
      histlock.Lock();
      string s = prefix;
      history.push(s);
      while (history.size() > histsize) history.pop();
      histlock.Unlock();
   }

   // Clear buffer.

   threadinfo->wpos = threadinfo->buffer + PREFIX;

}
