// ====================================================================
// syslog.hh - A class to encapsulate syslog and present it in a 
//             streamio-oriented manner.  Also make it thread-safe.
// ==================================================================== 

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.4 $ $Date: 2007/04/02 16:08:13 $ $Author: boeheim $
// ******************************************************************** //

#ifndef syslog_h
#define syslog_h

#include <string> 
#include <pthread.h>
#include <queue>

#include "slock.h"

// ====================================================================
// Structure that keeps per-thread information
// ====================================================================

#define BSIZE 1024
#define PREFIX 32

typedef struct
{
   int severity;
   char *buffer;
   char *wpos;
}  perthread;

// ====================================================================
// The syslog class
// ====================================================================

class Severity
{
public:
   int severity;
   Severity(int s) : severity(s) {};
};

class Syslog
{
private:
   
   int echofd;                  // File Descriptor to echo log to.
   int severity;                // Integer severity defined by system syslog.
   int logclass;                // Syslog class to log to.
   string progname;             // Program name in text.
   int histsize;                // Size of history buffer
   std::queue<string> history;  // History queue
   SLock histlock;              // Lock queue when changing.
   
   pthread_key_t datakey;       // Key to get thread data.
   
   perthread *getdata(void);    // Internal function to get thread data.
   
   void WriteBuffer(perthread *); // Internal function to write & clear buffer.
   
public:
   
   // Severity levels and facility codes, from syslog.h

   enum {emerg=0, alert, critical, error, warning, notice, info, debug};
   enum {kern=0, user, mail, daemon, auth, syslog, lpr, news, uucp, cron,
         authpriv, ftp, local0=16, local1, local2, local3, local4, local5,
         local6, local7};

   // Constructors and destructors

   Syslog(string prog = "", int cl = local0, int sev = debug);
   Syslog(char  *prog = "", int cl = local0, int sev = debug);
   ~Syslog(void);
   
   // Set the program name and logging class.

   void ProgName(string prog);
   void Class(int cl);
   
   void Echo(int fd) { echofd = fd; }; // Echo(false) to turn off.
   void SetHistory(int h) { histsize = h; }
   std::queue<string> History(void) { return history; }

   // Operators to stream various data types.

   void Append(const unsigned char  val, perthread *threadinfo=0);
   void Append(const unsigned char *val, perthread *threadinfo=0);
   void Append(const          char *val, perthread *threadinfo=0) { Append((unsigned char *)val, threadinfo); }; 

   Syslog& operator<< (const unsigned char val) { Append(val); return *this; };
   Syslog& operator<< (const unsigned char *val) { Append(val); return *this; };
   Syslog& operator<< (const string &val) { Append(val.c_str()); return *this; };
   Syslog& operator<< (const int val);
   Syslog& operator<< (ostream & (val)(ostream &));
   Syslog& operator<< (const Severity &val);
};

extern Syslog        SYSLOG;    // Reference to global syslogger.

#endif


