// ====================================================================
// Class for implementing a TTY file device.
// ====================================================================

using namespace std;

#include "ttyfile.h"
#include "syslog.h"
#include "consoled.h"
#include "entry.h"

#include <algorithm>
#include <iostream>
#include <termios.h>
#include <unistd.h>

extern int errno;

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.9 $ $Date: 2006/11/22 14:49:47 $ $Author: ahorvath $
// ******************************************************************** //

// ====================================================================
// Tables of baud rates and parity settings.
// ====================================================================

static Entry baudtab[] =        // Default in first position.
{
   Entry("9600",  B9600),
#ifdef B115200
   Entry("115200", B115200),
#endif
   Entry("57600", B57600),
   Entry("38400", B38400),
   Entry("19200", B19200),
   Entry("4800",  B4800),
   Entry("2400",  B2400),
   Entry("1200",  B1200),
};

#define baudtab_end (baudtab+sizeof(baudtab)/sizeof(Entry))

static Entry paritytab[] =      // Default in first position.
{
#if !defined(PAREXT)
#define PAREXT  0
#endif
   Entry("n", CS8, 0),                      // pass 8 bits, no parity         
   Entry("e", PARENB|CS7, 0),               // even                           
   Entry("m", PARENB|CS7|PARODD|PAREXT, 0), // mark                           
   Entry("o", PARENB|CS7|PARODD, 0),        // odd                            
   Entry("p", CS8, 0),                      // pass 8 bits, no parity         
   Entry("s", PARENB|CS7|PAREXT, 0),        // space                          
};

#define paritytab_end (paritytab+sizeof(paritytab)/sizeof(Entry))   

// ====================================================================
// Constructor for TTYFile object
// ====================================================================

TTYFile::TTYFile(void)
{
   readonly = false;
   autobaud = false;
   baud = baudtab[0].Val1();
   setparity = paritytab[0].Val1();
   clrparity = paritytab[0].Val2();
}

// ====================================================================
// Translate a string like "9600e" to the baud rate and parity 
// settings indicated.
// ==================================================================== 

void TTYFile::SetMode(const string &Mode)
{
   string numPart = "9600";
   string charPart = "p";
   string::size_type pos;
   Entry *found;

   // Set defaults.

   readonly = false;
   baud = baudtab[0].Val1();
   setparity = paritytab[0].Val1();
   clrparity = paritytab[0].Val2();

   // Check that something was passed.

   if (Mode.size() == 0)
   {
      cerr << "Null value passed to SetMode" << endl;
      return;
   }
   
   // If autobaud, figure it out later.

   if (Mode == "auto") { autobaud = true; return; }
   
   // Locate the numeric and character parts of the mode.
   
   pos = Mode.find_first_not_of("0123456789");

   if (pos == 0)
   {
      cerr << "Baud rate not specified in '" << Mode << "'" << endl;
      return;
   }

   if (pos == string::npos) pos = Mode.size();
   else charPart = Mode.substr(pos,1);

   numPart  = Mode.substr(0,pos);
   
   // Look up the baud rate value.

   found = find(baudtab, baudtab_end, numPart);

   if (found != baudtab_end) baud = found->Val1();
   else cerr << "Invalid baud rate '" << numPart 
             << "' in " << Mode << endl;
   
   // Look up the parity value.

   found = find(paritytab, paritytab_end, charPart);

   if (found != paritytab_end) 
   { 
      setparity = found->Val1();
      clrparity = found->Val2();
   }
   else cerr << "Invalid parity specification '" << charPart 
             << "' in " << Mode << endl;
}

// ====================================================================
// Get the current speed of the line as a character string
// ====================================================================

string TTYFile::GetSpeed(void)
{
   struct termios sbuf;
   int binspeed;
   string dispspeed;
   Entry *found;
   TCGetAttr(sbuf);

   binspeed = cfgetospeed(&sbuf);
   
   // Look up the baud rate value.

   found = find(baudtab, baudtab_end, binspeed);

   if (found != baudtab_end) dispspeed = found->Name();
   else dispspeed = "Not in table!";
   
   return dispspeed;
}

// ====================================================================
// Set the speed of the port to a specified value.
// ====================================================================

void TTYFile::SetSpeed(string speed)
{
   struct termios sbuf;
   Entry *found;
   TCGetAttr(sbuf);

   // Look up the baud rate value.

   found = find(baudtab, baudtab_end, speed);

   if (found != baudtab_end) baud = found->Val1();
   else cerr << "Invalid baud rate '" << speed << endl;
   
   if (cfsetospeed(&sbuf,baud) == -1)
      throw IOErr("cfsetospeed", errno, filename);
   
   if (cfsetispeed(&sbuf,baud) == -1)
      throw IOErr("cfsetispeed", errno, filename);

   TCSetAttr(TCSANOW, sbuf);   

   return;
}

// ====================================================================
// Set the speed to the next setting.
// ====================================================================

string TTYFile::NextSpeed(void)
{
   struct termios sbuf;
   Entry *found;
   TCGetAttr(sbuf);

   int binspeed = cfgetospeed(&sbuf);
   
   // Look up the baud rate value.

   found = find(baudtab, baudtab_end, binspeed);

   if (found == baudtab_end) 
   {
      cerr << "Can't determine next baud rate" << endl;
      return "Error";
   }

   found++;
   
   if (found == baudtab_end) found = baudtab;

   int baud = found->Val1();
      
   if (cfsetospeed(&sbuf,baud) == -1)
      throw IOErr("cfsetospeed", errno, filename);
   
   if (cfsetispeed(&sbuf,baud) == -1)
      throw IOErr("cfsetispeed", errno, filename);

   TCSetAttr(TCSANOW, sbuf);   

   return found->Name();
}

// ====================================================================
// Get terminal attributes.
// ====================================================================

void TTYFile::TCGetAttr(struct termios &sbuf)
{
   int error;

   error = tcgetattr(fd, &sbuf);

   if (error) throw IOErr("tcgetattr", errno, filename);
}

// ====================================================================
// Set terminal attributes.
// ====================================================================

void TTYFile::TCSetAttr(int flags, struct termios &sbuf)
{
   int error;

   error = tcsetattr(fd, flags, &sbuf);

   if (error) throw IOErr("tcsetattr", errno, filename);
}


// ====================================================================
// Get tty's program group.
// ====================================================================

int TTYFile::TCGetPgrp(void)
{
   return tcgetpgrp(fd);
}

// ====================================================================
// Set tty's program group.
// ====================================================================

void TTYFile::TCSetPgrp(int grp)
{
   int error;

   error = tcsetpgrp(fd, grp);

   if (error) throw IOErr("tcsetpgrp", errno, filename);
}

// ====================================================================
// Set tty's controlling process.
// ====================================================================

void TTYFile::SetControlTTY(void)
{
   int error;

#ifdef TIOCSCTTY
   error = ioctl(fd, TIOCSCTTY, 1);

   if (error) throw IOErr("ioctl(TIOCSCTTY)", errno, filename);
#endif
}

// ====================================================================
// Set tty's controlling process.
// ====================================================================

void TTYFile::NoControlTTY(void)
{
   int error;

#ifdef TIOCSCTTY
   error = ioctl(fd, TIOCNOTTY, 1);

   if (error) throw IOErr("ioctl(TIOCNOTTY)", errno, filename);
#endif
}

// ====================================================================
// Send a break on a TTY line.
// ====================================================================
 
void TTYFile::SendBreak(void)
{
   if (tcsendbreak(fd, 9) == -1)
      throw IOErr("tcsendbreak", errno, filename);
}

// ====================================================================
// Toggle flow control on TTY.  Return -1 for failure, 0 for
// flow off, 1 for flow on.
// ====================================================================
   
int TTYFile::FlowToggle(void)
{
  int flow = -1;

   struct termios sbuf;

   TCGetAttr(sbuf);
   
   if ((sbuf.c_iflag & IXOFF) != 0)
   {
      flow = 0;
      sbuf.c_iflag &= ~(IXOFF|IXON);
   }
   else
   {
      flow = 1;
      sbuf.c_iflag |= IXOFF|IXON;
   }

   try
   {
      TCSetAttr(TCSANOW, sbuf);
   }
   catch( IOErr e )
   {
      flow = -1;
   }

    return flow;

}

// ====================================================================
// Turn flow control on for TTY.  Return -1 for failure, 1 for
// flow on.
// ==================================================================== 
   
void TTYFile::FlowOn(void)
{
   struct termios sbuf;

   TCGetAttr(sbuf);
   
   sbuf.c_iflag |= IXOFF|IXON;

   TCSetAttr(TCSANOW, sbuf);
}

// ====================================================================
// Turn flow control off for TTY.  Return -1 for failure, 0 for 
// flow off.
// ==================================================================== 
   
void TTYFile::FlowOff(void)
{
   struct termios sbuf;

   TCGetAttr(sbuf);
   
   sbuf.c_iflag &= ~(IXOFF|IXON);

   TCSetAttr(TCSANOW, sbuf);
}

// ====================================================================
// Hangup the connection
// ====================================================================

void TTYFile::Hangup(void)
{
   struct termios sbuf, savebuf;
   
   TCGetAttr(sbuf);

   savebuf = sbuf;
   
   if (cfsetospeed(&sbuf,0) == -1)
      throw IOErr("cfsetospeed", errno, filename);
   
   if (cfsetispeed(&sbuf,0) == -1)
      throw IOErr("cfsetispeed", errno, filename);
   
   TCSetAttr(TCSANOW, sbuf);
   sleep(3);
   TCSetAttr(TCSANOW, savebuf);
}


// ====================================================================
// Toggle tab expansion on TTY.
// ====================================================================
  
#if !defined(XTABS)             // XXX hack 
#define XTABS   TAB3
#endif
   
void TTYFile::TabsToggle(void)
{
//   struct termios sbuf;
      
//   TCGetAttr(sbuf);

//   if (XTABS == (sbuf.c_oflag & TABDLY))
//   {
//      sbuf.c_oflag &= ~TABDLY;
//      sbuf.c_oflag |= TAB0;
//   } else
//   {
//      sbuf.c_oflag &= ~TABDLY;
//      sbuf.c_oflag |= XTABS;
//   }

//   TCSetAttr(TCSANOW, sbuf);

   return;
}

// ====================================================================
// Configure a TTY device to a normal configuration
// ==================================================================== 

int TTYFile::SetSane(void)
{
   struct termios termp;
   struct stat stPerm;
   
   // -----------------------------------------------------------------
   // here we should fstat for `read-only' checks
   // ----------------------------------------------------------------- 

   try
   {
      Stat(stPerm);
   }
   catch( IOErr e )
   {
      SYSLOG << Severity(Syslog::error)
             << e.routine 
             << ": " << e.filename 
             << ": " << e.message 
             << endl;
   }
	 
   if ((stPerm.st_mode & 0222) == 0)
      readonly = true;     // any device that is read-only we won't write to 


   // -----------------------------------------------------------------
   // Get terminal attributes
   // -----------------------------------------------------------------
   
   TCGetAttr(termp);

   // -----------------------------------------------------------------
   // Turn on: echo
   //          icrnl
   //          opost   Post processing
   //          icanon  Line editing
   //          isig    Signal generation
   // ----------------------------------------------------------------- 
   
   termp.c_iflag = ICRNL | IXON | IXANY;
   termp.c_lflag = ICANON | ISIG | ECHO | ECHOE | ECHOK | CLOCAL;
//   termp.c_oflag = OPOST | ONLCR | TAB3;
   termp.c_oflag = OPOST | ONLCR;
   
 
   // -----------------------------------------------------------------
   // Set the control characters.
   // -----------------------------------------------------------------

   termp.c_cc[VKILL ] = '\025';  // ^U
   termp.c_cc[VERASE] = '\010';  // ^H
   termp.c_cc[VEOF  ] = '\004';  // ^D
   termp.c_cc[VINTR ] = '\003';  // ^C
   termp.c_cc[VSTART] = '\021';  // ^Q
   termp.c_cc[VSTOP ] = '\023';  // ^S
   termp.c_cc[VSUSP ] = '\032';  // ^Z

   // -----------------------------------------------------------------
   // Set terminal attributes
   // -----------------------------------------------------------------

   TCSetAttr(TCSANOW, termp);   
   
   return 0;
}

// ====================================================================
// Configure a TTY device for RAW mode (termios version)
// ==================================================================== 

int TTYFile::SetRaw(void)
{
   struct termios termp;
   struct stat stPerm;
   
   // -----------------------------------------------------------------
   // here we should fstat for `read-only' checks
   // ----------------------------------------------------------------- 

   try
   {
      Stat(stPerm);
   }
   catch( IOErr e )
   {
      SYSLOG << Severity(Syslog::error)
             << e.routine 
             << ": " << e.filename 
             << ": " << e.message 
             << endl;
   }
	 
   if ((stPerm.st_mode & 0222) == 0)
      readonly = true;     // any device that is read-only we won't write to 


   // -----------------------------------------------------------------
   // Get terminal attributes
   // -----------------------------------------------------------------
   
   TCGetAttr(termp);

   // -----------------------------------------------------------------
   // Set the baud rates of the connection.
   // -----------------------------------------------------------------

   if (cfsetospeed(&termp,baud) == -1)
      throw IOErr("cfsetospeed", errno, filename);
   
   if (cfsetispeed(&termp,baud) == -1)
      throw IOErr("cfsetispeed", errno, filename);
   
   // -----------------------------------------------------------------
   // Turn off: echo
   //           icrnl
   //           opost   No post processing
   //           icanon  No line editing
   //           isig    No signal generation
   // Turn on:  ixoff
   // ----------------------------------------------------------------- 
   
   termp.c_cflag = (termp.c_cflag & ~CSIZE) | CS8;
   termp.c_cflag |= CLOCAL | CREAD; 
   termp.c_cflag &= ~(PARENB | PARODD | CRTSCTS);
   // this seems to break e.g. some BIOSes
   // termp.c_iflag = IGNBRK | PARMRK;
   termp.c_iflag = IGNBRK;
   termp.c_lflag = 0;
   termp.c_oflag = 0;
   termp.c_iflag &= ~(IXON|IXOFF|IXANY);
   
 
   // -----------------------------------------------------------------
   // Set the VMIN == 1 (min 1 char to satisfy read)
   // Set the VTIME == 5 (x0.1 sec) 
   // Don't bother with the control characters as they are not used
   // -----------------------------------------------------------------

   termp.c_cc[VTIME] = 1;
   termp.c_cc[VMIN] = 5;

   // -----------------------------------------------------------------
   // Set terminal attributes
   // -----------------------------------------------------------------

   TCSetAttr(TCSANOW, termp);   
   
   return 0;
}

// ====================================================================
// Turn RTS on or off
// ====================================================================

void TTYFile::SetRTS(bool turnon)
{
   unsigned int stat = TIOCM_RTS;
   int rc;

   if (turnon) rc = ioctl(fd, TIOCMBIS, &stat);
   else        rc = ioctl(fd, TIOCMBIC, &stat);

   if (rc != 0)
   {
      SYSLOG << Severity(Syslog::error)
             << "TIOCMBI[SC] RTS: " << filename 
             << ": " << strerror(errno)
             << endl;
      return;
   }
}

void TTYFile::SetCTS(bool turnon)
{
   unsigned int stat = TIOCM_CTS;
   int rc;

   if (turnon) rc = ioctl(fd, TIOCMBIS, &stat);
   else        rc = ioctl(fd, TIOCMBIC, &stat);

   if (rc != 0)
   {
      SYSLOG << Severity(Syslog::error)
             << "TIOCMBI[SC] CTS: " << filename 
             << ": " << strerror(errno)
             << endl;
      return;
   }
}

void TTYFile::SetDSR(bool turnon)
{
   unsigned int stat = TIOCM_DSR;
   int rc;

   if (turnon) rc = ioctl(fd, TIOCMBIS, &stat);
   else        rc = ioctl(fd, TIOCMBIC, &stat);

   if (rc != 0)
   {
      SYSLOG << Severity(Syslog::error)
             << "TIOCMBI[SC] DSR: " << filename 
             << ": " << strerror(errno)
             << endl;
      return;
   }
}

void TTYFile::SetCAR(bool turnon)
{
   unsigned int stat = TIOCM_CAR;
   int rc;

   if (turnon) rc = ioctl(fd, TIOCMBIS, &stat);
   else        rc = ioctl(fd, TIOCMBIC, &stat);

   if (rc != 0)
   {
      SYSLOG << Severity(Syslog::error)
             << "TIOCMBI[SC] CAR: " << filename 
             << ": " << strerror(errno)
             << endl;
      return;
   }
}

void TTYFile::SetDTR(bool turnon)
{
   unsigned int stat = TIOCM_DTR;
   int rc;

   if (turnon) rc = ioctl(fd, TIOCMBIS, &stat);
   else        rc = ioctl(fd, TIOCMBIC, &stat);

   if (rc != 0)
   {
      SYSLOG << Severity(Syslog::error)
             << "TIOCMBI[SC] DTR: " << filename 
             << ": " << strerror(errno)
             << endl;
      return;
   }
}

void TTYFile::SetRI (bool turnon)
{
   unsigned int stat = TIOCM_RI;
   int rc;

   if (turnon) rc = ioctl(fd, TIOCMBIS, &stat);
   else        rc = ioctl(fd, TIOCMBIC, &stat);

   if (rc != 0)
   {
      SYSLOG << Severity(Syslog::error)
             << "TIOCMBI[SC] RI: " << filename 
             << ": " << strerror(errno)
             << endl;
      return;
   }
}

// ====================================================================
// Get line signals from device
// ====================================================================

vector<bool> TTYFile::GetSignals(void)
{
   vector<bool> sigState(6, false);
   
   unsigned int stat = 0;
      
   if (ioctl(fd, TIOCMGET, &stat) == 0)
   {
      sigState[SigRTS] = (stat & TIOCM_RTS) ? true : false;
      sigState[SigCTS] = (stat & TIOCM_CTS) ? true : false;
      sigState[SigDSR] = (stat & TIOCM_DSR) ? true : false;
      sigState[SigCAR] = (stat & TIOCM_CAR) ? true : false;
      sigState[SigDTR] = (stat & TIOCM_DTR) ? true : false;
      sigState[SigRI ] = (stat & TIOCM_RNG) ? true : false;
   }
   
   return sigState;
}
