// ====================================================================
// Utility functions not tied to an object.
// ====================================================================

// ******************************************************************** //
//                  SLAC Console Management Tools                       //
//         Chuck Boeheim, Stanford Linear Accelerator Center            //
//                                                                      //
// $Revision: 1.10 $ $Date: 2006/07/18 09:47:50 $ $Author: boeheim $
// ******************************************************************** //

using namespace std;

#include <errno.h>
#include <iostream>
#include <pwd.h>
#include <signal.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string>
#include <ctype.h>
#include <unistd.h>
#include <vector>

#include "autorange.h"
#include "host.h"
#include "psstream.h"
#include "syslog.h"
#include "utility.h"

// ====================================================================
// Lowercase a string
// ====================================================================

string &Lower(string &str)
{
   for (int i = 0; i < (int) str.size(); i++) str[i] = tolower(str[i]);
   
   return str;
}

// ====================================================================
// Uppercase a string
// ====================================================================

string &Upper(string &str)
{
   for (int i = 0; i < (int) str.size(); i++) str[i] = toupper(str[i]);
   
   return str;
}

// ====================================================================
// Test a string for being all integer characters
// ====================================================================

bool IsInteger(const string &str)
{
   for (int i = 0; i < (int) str.size(); i++)
      if (! isdigit(str[i])) return false;
   
   return true;
}

// ====================================================================
// Convert a string to an integer.
// ====================================================================

int StrToInteger(const string &str) 
{
   int result = 0;
   int sign   = 1;
   int first  = 0;
   
   if (str[first] == '-') 
   {
      first++;
      sign = -1;
   }
   
   for (int i = first; i < (int) str.size(); i++)
   {
      if (! isdigit(str[i])) throw string("Not an integer: ") + str;
      result = result * 10 + (int) str[i] - '0';
   }

   result *= sign;
   
   return result;
}

// ====================================================================
// Convert a string of octal digits to an octal number.
// ====================================================================

int StrToOctal(const string &str) 
{
   int result = 0;
   
   for (int i = 0; i < (int) str.size(); i++)
   {
      if (str[i] < '0' || str[i] > '7') 
         throw string("Not an octal integer: ") + str;
      result = result * 8 + (int) str[i] - '0';
   }
   
   return result;
}

// ====================================================================
// Tokenize a line into a vector of strings.  ",=" are
// treated as white space and can be used as punctuation. 
// ====================================================================

char WHITE[] = ",= \t\n\r";

vector<string> Tokenize(string line)
{
   return Tokenize(line, WHITE);
}

vector<string> Tokenize(string line, char *split)
{
   string::size_type spos, epos;
   vector<string> tokens;

   spos = 0;
   epos = 0;

   line += "\n";
   
   while (1)
   {
      spos = line.find_first_not_of(split, epos);
      
      if (spos == string::npos) break; // All whitespace.
      if (line[spos] == '#') break;    // Comment
      
      if (line[spos] == '"')    // Quoted string
      {
         epos = line.find_first_of("\"", spos+1);
         
         if (epos == string::npos) epos = line.size();
         
         tokens.push_back(line.substr(spos+1, epos-spos-1));
         
         epos++;
      }
      else                      // Bare token
      {
         epos = line.find_first_of(split, spos);
      
         if (epos == string::npos) epos = line.size();
         
         tokens.push_back(line.substr(spos,epos-spos));
      }
   }
 
   return tokens;
}

// ====================================================================
// Make names by expanding range notations such as host{001-099}  
// tty{a-b}{a-z}{00-16}
// ====================================================================

vector<string> Makenames(string line, int copies)
{
   vector<string> names;
   vector<string> pieces;
   class autorange *iter = NULL;
   
   pieces = Tokenize(line, "{}\t\n\r ");
   for (unsigned int i = 0; i < pieces.size(); i++)
   {
      if (pieces[i].find_first_of("-") == string::npos) iter = new autoconst(pieces[i], iter);
      else if (isdigit(pieces[i][0])) iter = new autonum(pieces[i], iter); 
      else iter = new autostr(pieces[i], iter);
   }
      
   unsigned int ct = iter->count();
   for (int i = ct; i > 0; i--)
   {
      string s = iter->incr();
      names.push_back(s);
   }

   delete iter;
   
   return names;
}


// ====================================================================
// A Hostname is matched against a reference that should include the
// domain name, and wildcards are allowed.  
// console.slac.stanford.edu matches patterns 
// console.slac.stanford.edu, *.slac.stanford.edu, *.*.stanford.edu, 
// and *.*.*.edu.
// ====================================================================

char HOSTPUNC[] = ". \t\n\r";

bool HostCompare(const char *test, const char *ref)
{
   char testcopy[256];
   char refcopy[256];

   char *t = testcopy;
   char *r = refcopy;

   strncpy(testcopy, test, sizeof(testcopy)-1);
   strncpy(refcopy,  ref,  sizeof(refcopy )-1);
   
   testcopy[255] = '\0';
   refcopy[255]  = '\0';
   
   if (! strchr(testcopy, '.')) 
   {
      strcat(testcopy, ".");
      strcat(testcopy, Host::LocalDomain().c_str());
   }
   if (! strchr(refcopy,  '.')) 
   {
      strcat(refcopy, ".");
      strcat(refcopy,  Host::LocalDomain().c_str());
   }
   
   for (;;)
   {
      if (*t == '\0' && *r == '\0') return true;
      
      if (*t == '\0' || *r == '\0') return false;
      
      if (*t == *r) { t++; r++; continue; };

      if (*t == '*')
      {
         t++;
         while (*r != '.' && *r != '\0') r++;
         continue;
      } 
      if (*r == '*')
      {
         r++;
         while (*t != '.' && *t != '\0') t++;
         continue;
      }

      return false;             // non-match, not wildcard.
   }
}

// ====================================================================
// Simple test for a name matching the local host.
// ====================================================================

bool IsLocalHost(string hostname)
{
	if (hostname == "localhost" ||
		 HostCompare(hostname.c_str(), Host::LocalFQName().c_str()))
		return true;
	return false;
}

// ====================================================================
// Get the userid so it can be sent to the server
// ====================================================================

string FindUsername(void)
{
   char        *userstr = NULL;
   passwd      *pwdMe;

   if (((userstr = getenv("USER")) || (userstr = getenv("LOGNAME"))) &&
          (pwdMe = getpwnam(userstr)) && getuid() == pwdMe->pw_uid) 
   {
      // use the login $USER is set to, if it is our (real) uid 
   } 
   else if (!(pwdMe = getpwuid(getuid()))) 
   {
      throw string("Unable to obtain userid");
   }
   
   string username = pwdMe->pw_name;

   return username;
}

// ====================================================================
// Turn a process into a daemon.  Call before starting any other
// threads in a multi-threaded environment.
// ====================================================================

void Daemonize(void)
{
   // Mask unwanted signals.

#ifndef _AIX
   sigset_t signalset, signalnew;
   sigemptyset(&signalset);
   sigaddset(&signalset, SIGQUIT);
   sigaddset(&signalset, SIGINT);
   pthread_sigmask(SIG_BLOCK, &signalset, &signalnew);
#endif 

   // Disassociate from parent process by forking and exiting the 
   // intermediate process.

   int status = fork();
   
   switch (status)
   {
       case -1:                 // Error
          SYSLOG << Severity(Syslog::error)
                 << "Error forking process fork: "
                 << strerror(errno)
                 << endl;
          exit(1);
       case 0:                  // Child: continue
          break;
       default:                 // Parent: exit
          exit(0);
   }
   
   // Start a new session ID, so signals don't propagate.
   
   setsid();

   // Close file descriptors

   fclose(stdin);
   fclose(stdout);
   fclose(stderr);
}


// ====================================================================
// Return true if file exists, false if it doesn't.
// Produce a warning message if no option supplied,
// no message if "optional" is specified, and
// abort if "required" is specified.
// ====================================================================

bool CheckFile(string filename, string option)
{
   struct stat statbuf;
   int rc;
   
   if (filename != "")
   {
      rc = stat(filename.c_str(), &statbuf);
   
      if (rc == 0) return true;
   }

   if (option != "optional")
      cerr << "File " << filename << " does not exist" << endl;
   
   if (option == "required") exit(1);

   return false;
}
